/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: btctl-util.c,v 1.2 2002/07/19 20:51:14 kds Exp $

   Utility functions to read and parse the BT server config file. 

   Fixes:
   		Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
*/

#include <affix/config.h>

#include <stdio.h>
#include <stdlib.h>
#include <sys/fcntl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <string.h>
//#include <ctype.h>
#include <signal.h>
#include <termios.h>
#include <fcntl.h>

#include <affix/btcore.h>

#include "btctl.h"

char	cachefile[80];
struct inquiry_cache_head	ichead = { (struct inquiry_cache*)&ichead, 0 };


/*
 * Inquiry Cache Stuff
 */
void inquiry_cache_retire(void)
{
	struct inquiry_cache	*entry;
	
	for (entry = ichead.next; entry!=(struct inquiry_cache*)&ichead; entry = entry->next)
		entry->state = ICSTATE_GONE;
}

void inquiry_cache_print(int state)
{
	struct inquiry_cache	*entry;
	char	bda[20], buf[256], *name;
	int	i;
	char	ch;
	
	for (entry = ichead.next, i = 1; entry!=(struct inquiry_cache*)&ichead; entry = entry->next, i++) {
		if (!(entry->state & state))
			continue;
		entry->id = i;
		bda2str(bda, &entry->bda);
		parse_cod(buf, entry->class);
		if (entry->name[0] != '\0')
			name = entry->name;
		else
			name = "Unknown";
		switch (entry->state) {
			case ICSTATE_RANGE:
				ch = '+';
				break;
			case ICSTATE_GONE:
				ch = '-';
				break;
			case ICSTATE_UNKNOWN:
			default:
				ch = ' ';
		}
		printf("%c%d: Address: %s, Class: 0x%06X, Name: \"%s\"\n", 
				ch, entry->id, bda, entry->class, name);
		printf("    %s\n", buf);
	}
}


void inquiry_cache_free(void)
{
	struct inquiry_cache	*entry;
	
	for (;;) {
		entry = ichead.next;
		ichead.next = entry->next; 
		if (entry==(struct inquiry_cache*)&ichead)
			break;
		free(entry);
	}
}

int inquiry_cache_resolve(BD_ADDR *bda, int id)
{
	struct inquiry_cache	*entry;
	
	for (entry = ichead.next; entry!=(struct inquiry_cache*)&ichead; entry = entry->next) {
		if (entry->id == id) {
			*bda = entry->bda;
			return 0;
		}
	}
	return -1;
}

struct inquiry_cache *inquiry_cache_lookup(BD_ADDR bda)
{
	struct inquiry_cache	*entry;
	
	for (entry = ichead.next; entry!=(struct inquiry_cache*)&ichead; entry = entry->next) {
		if (bda_equal(&bda, &entry->bda))
			return entry;
	}
	return NULL;
}

int inquiry_cache_del(struct inquiry_cache *node)
{
	struct inquiry_cache	*entry, *prev;

	for (prev = (struct inquiry_cache*)&ichead, entry = ichead.next;
			entry != (struct inquiry_cache*)&ichead; prev = entry, entry = entry->next) {
		if (node == entry) {
			prev->next = entry->next;
			free(entry);
			return 0;
		}
	}
	return -1;
}

struct inquiry_cache *inquiry_cache_add(BD_ADDR bda, uint32_t class, char *name)
{
	struct inquiry_cache	*entry, *prev, *mount;

	for (prev = (struct inquiry_cache*)&ichead, entry = ichead.next;
			(entry != (struct inquiry_cache*)&ichead) || (entry = NULL);
			prev = entry, entry = entry->next) {
		if (bda_equal(&bda, &entry->bda)) {
			/* unlink */
			prev->next = entry->next;
			break;
		}
	}
	if (!entry) {
		/* create new */
		entry = malloc(sizeof(struct inquiry_cache));
		if (!entry) {
			perror("inquiry_cache allocation failed\n");
			return NULL;
		}
		memset(entry, 0, sizeof(struct inquiry_cache));	/* FIXME: change it */
		entry->bda = bda;
	}
	entry->state = ICSTATE_RANGE;
	entry->class = class;
	if (name)
		strcpy(entry->name, name);
	for (prev = (struct inquiry_cache*)&ichead, mount = ichead.next;
			mount != (struct inquiry_cache*)&ichead; prev = mount, mount = mount->next) {
		if (mount->state == ICSTATE_GONE) {
			break;
		}
	}
	entry->next = mount;
	prev->next = entry;
	return entry;
}

void inquiry_cache_load(void)
{
	char	buf[1024];
	FILE	*cfd;
	size_t	read;
	char	*next = (char*)buf, *elem, *attrs, *attr, *value;
	int	count = 0;
	BD_ADDR		bda;
	char		*name;
	uint32_t	class;
	
	cfd = fopen(cachefile, "r");
	if (!cfd) {
		return;
	}

	read = fread(buf, 1, sizeof(buf), cfd);
	if (!read) {
		fclose(cfd);
		return;
	}

	while ((elem = xml_element(&next, &attrs))) {
		if (strcmp(elem, "device-listing") == 0)
			break;
	}

	count = 0;
	while ((elem = xml_element(&next, &attrs))) {
		if (strcmp(elem, "/device-listing") == 0)
			break;
		name = NULL;
		//printf("element: %s\n", elem);
		//printf("attr left: %s\n", attrs);
		// get attributes
		if (strcmp(elem, "device") == 0) {
			struct inquiry_cache	*entry;
			while ((attr = xml_attribute(&attrs, &value))) {
				//printf("%s = %s\n", attr, value);
				if (strcmp(attr, "bda") == 0) {
					str2bda(&bda, value);
				} else if (strcmp(attr, "class") == 0) {
					sscanf(value, "%x", &class);
				} else if (strcmp(attr, "name") == 0) {
					name = value;
				}
			}
			entry = inquiry_cache_add(bda, class, name);
			entry->state = ICSTATE_UNKNOWN;
			entry->id = ++count;
		}
	}
	fclose(cfd);
}

int inquiry_cache_save(void)
{
	struct inquiry_cache	*entry;
	FILE	*cfd;
	char	bda[20];

	cfd = fopen(cachefile, "w");
	if (!cfd) {
		perror("Unable to open cache file\n");
		return -1;
	}
	fprintf(cfd, "<device-listing>\n");
	for (entry = ichead.next; entry!=(struct inquiry_cache*)&ichead; entry = entry->next) {
		bda2str(bda, &entry->bda);
		fprintf(cfd, "<device bda=\"%s\" class=\"%x\"", bda, entry->class);
		if (entry->name[0] != '\0')
			fprintf(cfd, " name=\"%s\"", entry->name);

		fprintf(cfd, "/>\n");
	}
	fprintf(cfd, "</device-listing>");
	fclose(cfd);
	return 0;
}


int get_bda(BD_ADDR *bda, char *arg)
{
	int	err;
	
	err = str2bda(bda, arg);
	if (!err) {
		/* try resolve */
		int	id;
		id = atoi(arg);
		//printf("id: %d\n", id);
		if (!id) 
			return -1;
		err = inquiry_cache_resolve(bda, id);
		if (err)
			return -1;
	}
	return 0;
}

/*
 * CLASS Of Device stuff
 */
int parse_cod(char *buf, uint32_t  cod)
{
	int	i, count = 0, found = 1;

	switch ( cod & COD_MAJOR ) {
		case COD_COMPUTER:
			for (i = 0; codMinorComputerMnemonic[i].msg || (found=0); i++) {
				if (codMinorComputerMnemonic[i].id == (cod & COD_MINOR)) {
					count += sprintf(buf+count, "Computer (%s)", codMinorComputerMnemonic[i].msg);
					break;
				}
			}
			if (!found)
				count += sprintf(buf+count, "Computer (Unclassified)");
			break;
		case COD_PHONE:
			for (i = 0; codMinorPhoneMnemonic[i].msg || (found=0); i++) {
				if (codMinorPhoneMnemonic[i].id == (cod & COD_MINOR)) {
					count += sprintf(buf+count, "Phone (%s)", codMinorPhoneMnemonic[i].msg);
					break;
				}
			}
			if (!found)
				count += sprintf(buf+count, "Phone (Unclassified)");
			break;
		case COD_MAUDIO:
			for (i = 0; codMinorAudioMnemonic[i].msg || (found=0); i++) {
				if (codMinorAudioMnemonic[i].id == (cod & COD_MINOR)) {
					count += sprintf(buf+count, "Audio (%s)", codMinorAudioMnemonic[i].msg);
					break;
				}
			}
			if (!found)
				count += sprintf(buf+count, "Audio (Unclassified)");
			break;
		default:
			for (i = 0; codMajorClassMnemonic[i].msg || (found=0); i++) {
				if (codMajorClassMnemonic[i].id == (cod & COD_MAJOR)) {
					count += sprintf(buf+count, "%s (Unclassified)", codMajorClassMnemonic[i].msg);
					break;
				}
			}
			if (!found)
				count += sprintf(buf+count, "Unclassified (Unclassified)");
	}
	count += sprintf(buf+count, " [");
	for (i = 0; codServiceClassMnemonic[i].msg; i++) {
		if (codServiceClassMnemonic[i].id & cod) {
			count += sprintf(buf+count, "%s,", codServiceClassMnemonic[i].msg);
		}
	}
	count--;	// remove ,
	count += sprintf(buf+count, "]");

	return 0;
}

int assign_cod(__u32 *cod)
{
	int	i, found = 1;
	char	*arg;

	/* get Major */
	arg = argv[argind++];
	if (arg == NULL)
		return -1;

	*cod = 0;
	for (i = 0; codMajorClassMnemonic[i].match || (found=0); i++) {
		if (strncasecmp(codMajorClassMnemonic[i].match, arg, 3) == 0) {
			*cod |= codMajorClassMnemonic[i].id;
			break;
		}
	}
	if (!found)
		return -1;

	/* get minor */
	arg = argv[argind++];
	if (arg == NULL)
		return -1;

	switch ( *cod & COD_MAJOR ) {
		case COD_COMPUTER:
			for (i = 0; codMinorComputerMnemonic[i].match || (found=0); i++) {
				if (strncasecmp(codMinorComputerMnemonic[i].match, arg, 3) == 0) {
					*cod |= codMinorComputerMnemonic[i].id;
					break;
				}
			}
			break;
		case COD_PHONE:
			for (i = 0; codMinorPhoneMnemonic[i].match || (found=0); i++) {
				if (strncasecmp(codMinorPhoneMnemonic[i].match, arg, 3) == 0) {
					*cod |= codMinorPhoneMnemonic[i].id;
					break;
				}
			}
			break;
		case COD_MAUDIO:
			for (i = 0; codMinorAudioMnemonic[i].match || (found=0); i++) {
				if (strncasecmp(codMinorAudioMnemonic[i].match, arg, 3) == 0) {
					*cod |= codMinorAudioMnemonic[i].id;
					break;
				}
			}
			break;
		default:
			found = 0;
	}

	if (!found)
		return -1;

	/* get services */
	for (arg = argv[argind++]; arg; arg = argv[argind++]) {
		for (i = 0; codServiceClassMnemonic[i].match || (found=0); i++) {
			if (strncasecmp(codServiceClassMnemonic[i].match, arg, 3) == 0) {
				*cod |= codServiceClassMnemonic[i].id;
				break;
			}
		}
		if (!found)
			return -1;
	}

	return 0;
}


