/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   */

/* 
   $Id: btctl.c,v 1.12 2002/08/29 13:14:37 kds Exp $

   btctl - driver control program

	Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
	Imre Deak <ext-imre.deak@nokia.com>
*/

#include <affix/config.h>

#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/errno.h>

#include <fcntl.h>
#include <unistd.h>
#include <syslog.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <getopt.h>
#include <string.h>
#include <termios.h>
#include <setjmp.h>

//#include <readline/readline.h>

#include <affix/bluetooth.h>
#include <affix/btcore.h>
#include <affix/debug.h>

#include "btctl.h"

struct str2id_map {
	char	*str;
	__u32	id;
};

int	argind;
char	btdev[IFNAMSIZ] = "bt0";
//int	sdpmode = 1;
int	ftpmode = 0;
char	*home;
char	*user;
char	confdir[80];

int print_hci_error(int err)
{
	if (err) {
		if (err < 0)
			perror("System error");
		else
			printf("HCI error: %s\n", hci_errormsg(err));
		exit(1);
	}
	return 0;
}

int cmd_inquiry(struct btctl_command *cmd)
{
	int		fd, i;
	__u32		length;
	int		err;
	INQUIRY_ITEM	devs[20];
	__u8		num;
	
	switch (cmd->cmd) {
		case 2:
			inquiry_cache_print(ICSTATE_ALL);
			return 0;
		case 3:
			inquiry_cache_free();
			inquiry_cache_save();
			return 0;
	}

	if (argv[argind] == NULL) {
		printf("length missed\n");
		return -1;
	}
	sscanf(argv[argind], "%x", &length);

	fd = hci_open(btdev);
	if (fd == -1) {
		printf("Unable to open device %s: %s\n", btdev, sys_errlist[errno]);
		return -1;
	}
	inquiry_cache_retire();
	printf("Searching ...\n");
	err = HCI_Inquiry(fd, length, 20, devs, &num);
	print_hci_error(err);
	printf("done.\n");
	if (num == 0) {
		printf("No devices found.\n");
	} else {
		for (i = 0; i < num; i++) {
			//char	str[20];
			//bda2str(str, &devs[i].bda);
			//printf("Adding %s\n", str);
			inquiry_cache_add(devs[i].bda, devs[i].Class_of_Device, NULL);
		}
		inquiry_cache_save();
		inquiry_cache_print(ICSTATE_RANGE);
	}
	return 0;
}

int cmd_discovery(struct btctl_command *cmd)
{
	int		fd, i;
	__u32		length;
	int		err;
	INQUIRY_ITEM	devs[20];
	char		namebuf[248];
	__u8		num;
	char		*name;

	if (argv[argind] == NULL) {
		printf("length missed\n");
		return -1;
	}
	sscanf(argv[argind], "%x", &length);

	fd = hci_open(btdev);
	if (fd == -1) {
		printf("Unable to open device %s: %s\n", btdev, sys_errlist[errno]);
		return -1;
	}
	inquiry_cache_retire();
	printf("Searching ...\n");
	err = HCI_Inquiry(fd, length, 20, devs, &num);
	print_hci_error(err);
	if (num == 0) {
		printf("done.\nNo devices found.\n");
	} else {
		printf("Searching done. Resolving names ...\n");
		for (i = 0; i < num; i++) {
			err = HCI_RemoteNameRequest(fd, &devs[i], namebuf);
			if (!err)
				name = namebuf;
			else 
				name = NULL;
			inquiry_cache_add(devs[i].bda, devs[i].Class_of_Device, name);
		}
		printf("done.\n");
		inquiry_cache_save();
		inquiry_cache_print(ICSTATE_RANGE);
	}
	return 0;
}

int cmd_bdaddr(struct btctl_command *cmd)
{
	int	err, fd;
	BD_ADDR	bda;
	char	buf[20];

	fd = hci_open(btdev);
	if (fd == -1) {
		printf("Unable to open device %s: %s\n", btdev, sys_errlist[errno]);
		return -1;
	}

	err = HCI_ReadBDAddr(fd, &bda);
	print_hci_error(err);

	bda2str(buf, &bda);
	printf("Local Address: %s\n", buf);

	hci_close(fd);

	return 0;
}

int cmd_name(struct btctl_command *cmd)
{
	int	err, fd;
	char	name[248];

	fd = hci_open(btdev);
	if (fd == -1) {
		printf("Unable to open device %s: %s\n", btdev, sys_errlist[errno]);
		return -1;
	}

	if (argv[argind] == NULL) {
		// read name
		err = HCI_ReadLocalName(fd, name);
	} else {
		strncpy(name, argv[argind], 248);
		err = HCI_ChangeLocalName(fd, name);
	}

	print_hci_error(err);

	if (argv[argind] == NULL) {
		// read name
		printf("Name: %s\n", name);
	}

	hci_close(fd);
	return 0;
}

int cmd_scan(struct btctl_command *cmd)
{
	int	err, fd, flag;
	__u8	mode;
	char	*param;

	if (argv[argind] == NULL) {
		printf("scan mode missed\n");
		return -1;
	}

	fd = hci_open(btdev);
	if (fd == -1) {
		printf("Unable to open device %s: %s\n", btdev, sys_errlist[errno]);
		return -1;
	}

	err = HCI_ReadScanEnable(fd, &mode);
	print_hci_error(err);

	while (argv[argind]) {
		param = argv[argind];
		if (param[0] == '+') flag = 1, param++;
		else if (param[0] == '-') flag = 0, param++;
		else flag = 1;

		if (strcmp(param, "disc") == 0)
			flag?(mode|=0x01):(mode&=~0x01);
		else if (strcmp(param, "conn") == 0)
			flag?(mode|=0x02):(mode&=~0x02);

		argind++;
	}

	err = HCI_WriteScanEnable(fd, mode);
	print_hci_error(err);

	hci_close(fd);

	return 0;
}

int cmd_pincode(struct btctl_command *cmd)
{
	int		err;
	BD_ADDR		bda;
	int		Length;
	__u8		Code[16];

	if (cmd->cmd == 1 && (argv[argind] == NULL || argv[argind+1] == NULL)) {
		printf("Parameters missed\n");
		return -1;
	}

	if (argv[argind]) {
		if (strcasecmp(argv[argind], "default") == 0)
			memset(&bda, 0, 6);
		else {
			err = get_bda(&bda, argv[argind]);
			if (err) {
				printf("Incorrect address given\n");
				return 1;
			}
		}
		if (cmd->cmd == 1) {
			Length = strlen(argv[argind+1]);
			strncpy((char*)Code, argv[argind+1], 16);
			err = hci_add_pin(&bda, Length, Code);
		} else
			err = hci_remove_pin(&bda);
	} else
		err = hci_remove_pin(NULL);

	print_hci_error(err);
	return err;
}

int cmd_key(struct btctl_command *cmd)
{
	int		err;
	BD_ADDR		bda;

	if (argv[argind]) {
		err = get_bda(&bda, argv[argind]);
		if (err) {
			printf("Incorrect address given\n");
			return 1;
		}
		err = hci_remove_key(&bda);
	} else
		err = hci_remove_key(NULL);

	return err;
}

int cmd_security(struct btctl_command *cmd)
{
	int			err, mode, fd;
	struct hci_dev_attr	da;

	if (!argv[argind]) {
		print_usage(cmd);
		return 1;
	}
	mode = atoi(argv[argind++]);
	
	fd = hci_open(btdev);
	if (fd == -1) {
		printf("Unable to open device %s: %s\n", btdev, sys_errlist[errno]);
		return -1;
	}

	da.sec_mode = mode;
	da.attr = HCI_ATTR_SECMODE;

	err = hci_set_attr(fd, &da);
	print_hci_error(err);

	return 0;
}



int cmd_remotename(struct btctl_command *cmd)
{
	int		err, fd;
	BD_ADDR		bda;
	INQUIRY_ITEM	dev;
	char		Name[248];

	if (argv[argind] == NULL) {
		printf("Bluetooth address missed\n");
		return -1;
	}

	err = get_bda(&bda, argv[argind]);
	if (err) {
		printf("Wrong address format\n");
		return 1;
	}

	fd = hci_open(btdev);
	if (fd < 0) {
		printf("Unable to open device %s: %s\n", btdev, sys_errlist[errno]);
		return fd;
	}

	dev.bda = bda;
	dev.PS_Repetition_Mode = 0x00;
	dev.PS_Mode = 0x00;
	dev.Clock_Offset = 0x00;
	err = HCI_RemoteNameRequest(fd, &dev, Name);
	print_hci_error(err);

	printf("Remote name: %s\n", Name);
	return 0;
}

int cmd_role(struct btctl_command *cmd)
{
	int	err, fd;
	struct hci_dev_attr	da;
	
	if (argv[argind] == NULL) {
		print_usage(cmd);
		return -1;
	}

	fd = hci_open(btdev);
	if (fd == -1) {
		printf("Unable to open device %s: %s\n", btdev, sys_errlist[errno]);
		return -1;
	}

	if (strcmp(argv[argind++], "deny") == 0)
		da.allow_switch = ROLE_DENY_SWITCH;
	else
		da.allow_switch = ROLE_ALLOW_SWITCH;

	if (strcmp(argv[argind++], "master") == 0)
		da.role = ROLE_BECOME_MASTER;
	else
		da.role = ROLE_REMAIN_SLAVE;

	da.attr = HCI_ATTR_ALLOW_SWITCH | HCI_ATTR_ROLE;
	
	err = hci_set_attr(fd, &da);
	print_hci_error(err);

	hci_close(fd);
	return 0;

}


int cmd_class(struct btctl_command *cmd)
{
	int		err, fd;
	__u32		cod;
	__u32		value;
	char		buf[80];

	fd = hci_open(btdev);
	if (fd == -1) {
		printf("Unable to open device %s: %s\n", btdev, sys_errlist[errno]);
		return -1;
	}

	if (argv[argind] == NULL) {
		// read name
		err = HCI_ReadClassOfDevice(fd, &cod);
		print_hci_error(err);
		parse_cod(buf, cod);
		printf("Class: 0x%06X, %s\n", cod, buf);
	} else {
		if (strstr(argv[argind], "0x")) {
			err = sscanf(argv[argind], "%x ", &value);
			if (err > 0)
				cod = value;
		} else {
			// parse command line
			err = assign_cod(&cod);
			if (err) {
				printf("bad arguments\n");
				return -1;
			}
		}
		err = HCI_WriteClassOfDevice(fd, cod);
		print_hci_error(err);
	}

	hci_close(fd);
	return 0;
}

int cmd_auth(struct btctl_command *cmd)
{
	int	err, fd;
	__u8	val = 0;

	if (argv[argind] == NULL) {
		print_usage(cmd);
		return -1;
	}

	fd = hci_open(btdev);
	if (fd == -1) {
		printf("Unable to open device %s: %s\n", btdev, sys_errlist[errno]);
		return -1;
	}

	if (cmd->cmd == 1) {
		if (strcmp(argv[argind], "on") == 0)
			val = 1;
		err = HCI_WriteAuthenticationEnable(fd, val);
	} else {
		/* authenticate */
		int		handle;
		BD_ADDR		bda;

		err = get_bda(&bda, argv[argind]);
		if (err) {
			printf("Wrong address format\n");
			return 1;
		}
		handle = hci_get_conn(fd, &bda);
		if (handle < 0) {
			printf("Connection not found\n");
			return 1;
		}
		err = HCI_AuthenticationRequested(fd, handle);
	}
	print_hci_error(err);
	hci_close(fd);

	return 0;
}

int cmd_encrypt(struct btctl_command *cmd)
{
	int	err, fd;
	__u8	val = 0;

	if (argv[argind] == NULL) {
		print_usage(cmd);
		return -1;
	}

	fd = hci_open(btdev);
	if (fd == -1) {
		printf("Unable to open device %s: %s\n", btdev, sys_errlist[errno]);
		return -1;
	}

	if (strcmp(argv[argind], "ppp") == 0)
		val = 1;
	else if (strcmp(argv[argind], "bcast") == 0)
		val = 2;

	err = HCI_WriteEncryptionMode(fd, val);
	print_hci_error(err);

	hci_close(fd);

	return 0;
}

/*
 * L2CAP
 */

void print_data(__u8 *p, int len)
{
	int	i,j;
	char 	buf[81];

	if (len > 100)
		return;
	for (i = 0; i < len; i++) {
		if ( ((i % 16) == 0) && i > 0)
			printf("%s\n", buf);
		j = (i % 16) * 3;
		sprintf( &(buf[j]), "%02x ", p[i]);
	}
	if (len)
		printf("%s\n", buf);
}


int cmd_ping(struct btctl_command *cmd)
{
	int		err, fd, i;
	BD_ADDR		bda;
	struct sockaddr_affix	saddr;
	char		str[20];
	int		size;
	__u8		*data;

	if (argv[argind] == NULL || argv[argind+1] == NULL) {
		printf("parameters missing\n");
		print_usage(cmd);
		return 1;
	}

	err = get_bda(&bda, argv[argind++]);
	if (err) {
		printf("Incorrect address given\n");
		return 1;
	}
	bda2str(str, &bda);
	printf("Connecting to host %s ...\n", str);

	size = atoi(argv[argind++]);

	fd = socket(PF_AFFIX, SOCK_STREAM, BTPROTO_L2CAP);
	if (fd < 0) {
		printf("Unable to create RFCOMM socket: %d\n", PF_AFFIX);
		return 1;
	}

	saddr.bda = bda;
	saddr.port = 0;
	saddr.local = BDADDR_ANY;
	err = connect(fd, (struct sockaddr*)&saddr, sizeof(saddr));
	if (err < 0) {
		printf("Unable to connect to remote side\n");
		close(fd);
		return 1;
	}

	data = (__u8*)malloc(size);
	if( data == NULL )
		return 0;

	for(i = 0; i < size; i++)
		data[i] = i;

	printf("Sending %d bytes...\n", size);
	print_data(data, size);
	if (cmd->cmd == 0) {
		err = l2cap_ping(fd, data, size);
		if (err < 0) {
			printf("ping error\n");
			exit(1);
		}
		printf("Received %d bytes back\n", err);
		print_data(data, err);
	} else if (cmd->cmd == 1) {
		struct timeval	tv_start, tv_end;
		long int	sec, rsec;
		long int	usec, rusec;
		double		speed;

		gettimeofday(&tv_start, NULL);
		err = l2cap_singleping(fd, data, size);
		if (err < 0) {
			printf("ping error\n");
			exit(1);
		}
		printf("flushing...\n");
		err = l2cap_flush(fd);
		if (err < 0) {
			printf("flush error\n");
			exit(1);
		}
		gettimeofday(&tv_end, NULL);			

		sec = tv_end.tv_sec - tv_start.tv_sec;
		usec = (1000000 * sec) + tv_end.tv_usec - tv_start.tv_usec;
		rsec = usec/1000000;
		rusec = (usec - (rsec * 1000000))/10000;
		speed = (double)(size)/((double)(rsec) + (double)(rusec)/100);

		printf("%d bytes sent in %ld.%ld secs (%.2f B/s)\n", size, rsec, rusec, speed);
	}

	close(fd);
	return 0;
}


/*
 * RFCOMM stuff
 */
int cmd_connect(struct btctl_command *cmd)
{
	int		err, fd;
	BD_ADDR		bda;
	int		sch;
	int		line;
	struct sockaddr_affix	saddr;
	char		str[20];

	if (sdpmode) {
		if (argv[argind] == NULL) {
			printf("Address missing\n");
			print_usage(cmd);
			return 1;
		}
	} else {
		if (argv[argind] == NULL || argv[argind+1] == NULL) {
			printf("parameters missing\n");
			print_usage(cmd);
			return 1;
		}
	}

	err = get_bda(&bda, argv[argind]);
	if (err) {
		printf("Incorrect address given\n");
		return 1;
	}
	bda2str(str, &bda);
	printf("Connecting to host %s ...\n", str);

#if defined(CONFIG_AFFIX_SDP)
	if (sdpmode) {
		UUID16Bit	ServiceID;

		if (argv[argind+1]) {
			if (strncasecmp(argv[argind+1], "SERial", 3) == 0)
				ServiceID = SerialPortServiceClassID; 
			else if (strncasecmp(argv[argind+1], "DUN", 3) == 0)
				ServiceID = DialupNetworkingServiceClassID;
			else if (strncasecmp(argv[argind+1], "LAN", 3) == 0)
				ServiceID = LANAccessUsingPPPServiceClassID;
			else if (strncasecmp(argv[argind+1], "HEAdset", 3) == 0)
				ServiceID = HeadsetServiceClassID;
			else
				ServiceID = SerialPortServiceClassID;
		} else
			ServiceID = SerialPortServiceClassID;

		sch = SDPServiceSearchPortRequest(bda, ServiceID);
		if (sch > 0)
			printf("Service found on channel %d\n", sch);
		else if ( sch == 0 ) {
			printf("Service is not available\n");
			return -1;
		} else {
			printf("Unable to get service channel\n");
			return 1;
		}
	} else
#endif
		sch = atoi(argv[argind+1]);


	printf("Connecting to channel %d ...\n", sch);

	fd = socket(PF_AFFIX, SOCK_STREAM, BTPROTO_RFCOMM);
	if (fd < 0) {
		printf("Unable to create RFCOMM socket: %d\n", PF_AFFIX);
		return 1;
	}

	err = rfcomm_set_type(fd, RFCOMM_BTY);
	if (err < 0) {
		BTERROR("Unable to set RFCOMM interface type to tty");
		close(fd);
		return 1;
	}

	saddr.bda = bda;
	saddr.port = sch;
	saddr.local = BDADDR_ANY;
	err = connect(fd, (struct sockaddr*)&saddr, sizeof(saddr));
	if (err < 0) {
		printf("Unable to connect to remote side\n");
		close(fd);
		return 1;
	}

	line = rfcomm_bind_port(fd, -1);
	if (line < 0) {
		printf("Unable to bind a port to the socket\n");
		close(fd);
		return 1;
	}

	printf("Connected. Bound to line %d [/dev/bty%1$d].\n", line);

	close(fd);
	return 0;
}

int cmd_disconnect(struct btctl_command *cmd)
{
	int	line = -1, err;

	if (argv[argind] == NULL) {
		printf("disconnect all\n");
	} else {
		line = atoi(argv[argind]);
		printf("Disconnecting line %d [/dev/bty%1$d]\n", line);
	}

	// Disconnect here
	err = rfcomm_disconnect(line);
	if (err < 0) {
		printf("Unable to disconnect line: %d\n", line);
		return 1;
	}

	return 0;
}

void print_status(struct rfcomm_port *port, int pnum)
{
	int	i;
	char	bda[20];

	printf("Connected lines:\n");
	for (i=0; i < pnum; i++, port++) {
		bda2str(bda, &port->bda);
		printf("line: %d [/dev/bty%1$d], bda: %s, channel: %d\n",
				port->line, bda, port->channel);
	}
}

int cmd_status(struct btctl_command *cmd)
{
	struct rfcomm_port	info[10];
	int		count;

	count = rfcomm_get_ports(info, 10);
	if (count < 0) {
		printf("Unable to get portinfo\n");
		return 1;
	}

	if (count == 0)
		printf("No connected lines\n");
	else {
		print_status(info, count);
	}

	return 0;
}

struct str2id_map pkt_list[] = {
	{"DM1", PT_DM1},
	{"DH1", PT_DH1},
	{"DM3", PT_DM3},
	{"DH3", PT_DH3},
	{"DM5", PT_DM5},
	{"DH5", PT_DH5},
	{"HV1", PT_HV1},
	{"HV2", PT_HV2},
	{"HV3", PT_HV3},
	{0, 0}
};


int cmd_pkttype(struct btctl_command *cmd)
{
	int	err, value, fd, i;
	struct hci_dev_attr	da;
	
	fd = hci_open(btdev);
	if (fd == -1) {
		printf("Unable to open device %s: %s\n", btdev, sys_errlist[errno]);
		return -1;
	}

	if (argv[argind] == NULL)
		value = 0; //set default
	else {
		if (argv[argind][0] == '0')
			err = sscanf(argv[argind], "%x", &value);
		else {
			// try symbolic notation
			for (value = 0; argv[argind]; argind++ ) {
				for (i = 0; pkt_list[i].str; i++)
					if (strcasecmp(argv[argind], pkt_list[i].str) == 0)
						value |= pkt_list[i].id;
			}
		}
	}

	da.pkt_type = value;
	da.attr = HCI_ATTR_PKT_TYPE;
	err = hci_set_attr(fd, &da);
	print_hci_error(err);
	hci_close(fd);

	return 0;

}

int cmd_control(struct btctl_command *cmd)
{
	int	err, fd;

	fd = hci_open(btdev);
	if (fd == -1) {
		printf("Unable to open device %s: %s\n", btdev, sys_errlist[errno]);
		return -1;
	}

	switch (cmd->cmd) {
		case 0:
			{
				__u16	timeo;

				if (argv[argind] == NULL) {
					// read name
					err = HCI_ReadPageTimeout(fd, &timeo);
				} else {
					sscanf(argv[argind], "%hu", &timeo);
					printf("Timeo: %u\n", timeo);
					err = HCI_WritePageTimeout(fd, timeo);
				}

				print_hci_error(err);

				if (argv[argind] == NULL) {
					// read name
					printf("Timeo: %u\n", timeo);
				}
			}
			break;

		case 1: {
				int	handle, reason;
				BD_ADDR	bda;

				if (strstr(argv[argind], "0x")) {
					err = sscanf(argv[argind++], "%x", &handle);
					if (err <= 0) {
						printf("format error\n");
						return 1;
					}
				} else {
					err = get_bda(&bda, argv[argind++]);
					if (err) {
						printf("Incorrect address given\n");
						return 1;
					}
					handle = hci_get_conn(fd, &bda);
					if (handle < 0) {
						printf("Connection not found\n");
						return 1;
					}
				}
				if (argv[argind]) {
					err = sscanf(argv[argind++], "%x", &reason);
					if (err <= 0)
						reason = 0x13;
				} else
					reason = 0x13;
					
				printf("Disconnecting handle %#hx ...\n", handle);
				err = HCI_Disconnect(fd, handle, reason);
				print_hci_error(err);
			}
			break;

		case 2: {
				INQUIRY_ITEM	dev;
				
				err = get_bda(&dev.bda, argv[argind++]);
				if (err) {
					printf("Incorrect address given\n");
					return 1;
				}
				dev.PS_Repetition_Mode = 0x00;
				dev.PS_Mode = 0x00;
				dev.Clock_Offset = 0x00;
				err = __HCI_CreateConnection(fd, &dev, 
						PT_DM1|PT_DH1|PT_DM3|PT_DH3|PT_DM5|PT_DH5, 0);
				print_hci_error(err);
			}

		default:
	}

	hci_close(fd);

	return 0;
}


int cmd_reset(struct btctl_command *cmd)
{
	int	fd, err;

	fd = hci_open(btdev);
	if (fd == -1) {
		printf("Unable to open device %s: %s\n", btdev, sys_errlist[errno]);
		return -1;
	}

	err = HCI_Reset(fd);
	print_hci_error(err);
	hci_close(fd);

	return 0;
}


struct str2id_map flag_list[] = {
	{"hcicore", DBHCI},
	{"afhci", DBAFHCI},
	{"hcimgr", DBHCIMGR},
	{"hcisched", DBHCISCHED},
	{"hci", DBHCI|DBAFHCI|DBHCIMGR|DBHCISCHED},
	{"pl2cap", DBL2CAP},
	{"afl2cap", DBAFL2CAP},
	{"l2cap", DBL2CAP|DBAFL2CAP},
	{"prfcomm", DBRFCOMM},
	{"afrfcomm", DBAFRFCOMM},
	{"bty", DBBTY},
	{"rfcomm", DBRFCOMM|DBAFRFCOMM|DBBTY},
	{"uart", DBUART},
	{"usb", DBUSB},
	{"drv", DBUART|DBUSB},
	{"allmod", DBALLMOD},

	{"init", DBINIT},
	{"xmit", DBXMIT},
	{"xmite", DBXMITE},
	{"recv", DBRECV},
	{"recve", DBRECVE},
	{"ctrl", DBCTRL},
	{"ctrle", DBCTRLE},
	{"allerror", DBALLERROR},
	{"alltype", DBALLTYPE},

	{"timer", DBTIMER},
	{"parse", DBPARSE},
	{"chardump", DBCHARDUMP},
	{"dump", DBDUMP},
	{"fname", DBFNAME},
	{"func", DBFUNC},
	{"alldetail", DBALLDETAIL},

	{"all", 0xFFFFFFFF},
	{0, 0}
};

int cmd_debug(struct btctl_command *cmd)
{
	int	err, flag, i;
	__u32	dbmask;
	char	*param;
	int	mod_added = 0;

	err = hci_get_dbmask(&dbmask);
	print_hci_error(err);

	while (argv[argind]) {
		param = argv[argind];
		if (param[0] == '+') flag = 1, param++;
		else if (param[0] == '-') flag = 0, param++;
		else flag = 1;

		for (i = 0; flag_list[i].str; i++)
			if (strcasecmp(param, flag_list[i].str) == 0) {
				flag?(dbmask|=flag_list[i].id):(dbmask&=~flag_list[i].id);
				if (flag && (flag_list[i].id & DBALLMOD))
					mod_added = 1;
			}
		argind++;
	}
	if (mod_added && (!(dbmask & DBALLTYPE) || !(dbmask & DBALLDETAIL)))
		dbmask |= DBALLTYPE | DBALLDETAIL;

	err = hci_set_dbmask(dbmask);
	print_hci_error(err);

	return 0;
}

int status(void)
{
	int		err, fd;
	char		str[20], name[248], buf[80];
	__u32		cod;
	__u8		val;
	__u8		hci_version, lmp_version;
	__u16		hci_revision, manufacture_name, lmp_subversion;
	__u64		lmp_features;
	int		devs[16], i, num;
	__u16		acl_len, acl_num, sco_num;
	__u8		sco_len;
	struct hci_dev_attr	da;

	printf("\n");

	num = hci_get_devs(devs);
	if (num < 0) {
		print_hci_error(num);
		return num;
	}

	if (num == 0) {
		printf("No Bluetooth Adapters found\n");
		return 0;
	}

	for (i=0; i<num; i++) {
		fd = hci_open_id(devs[i]);
		if (fd < 0) {
			printf("Unable to open Bluetooth device: %d\n\n", devs[i]);
			return -1;
		}
		
		err = hci_get_attr(fd, &da);
		if (err < 0) {
			printf("Unable to get Device Info for: %d\n", devs[i]);
			continue;
		}

		bda2str(str, &da.bda);
		printf("%s\t%s\n", da.name, str);
		err = HCI_ReadLocalVersionInformation(fd, &hci_version, &hci_revision, &lmp_version,
				&manufacture_name, &lmp_subversion);
		print_hci_error(err);
		printf("\tBasebend manufacture name: %s, id: %d\n", (manufacture_name<NUM_LMPCOMPID)?lmp_compid[manufacture_name]:"Unknown", manufacture_name);
		printf("\tBaseband is %s compliant\n", lmp_version?"1.1":"1.0");

		err = HCI_ReadLocalName(fd, name);
		print_hci_error(err);
		printf("\tName: \"%s\"\n", name);

		err = HCI_ReadClassOfDevice(fd, &cod);
		print_hci_error(err);
		parse_cod(buf, cod); 
		printf("\tClass: 0x%06X, %s\n", cod, buf);

		err = HCI_ReadScanEnable(fd, &val);
		print_hci_error(err);
		printf("\tScan Mode: %s, %s\n", (val&0x01)?"discoverable":"non-discoverable",
				(val&0x02)?"connectable":"non-connectable");

		err = HCI_ReadAuthenticationEnable(fd, &val);
		print_hci_error(err);
		printf("\tAuthentication: %s\n", (val)?"enabled":"disabled");

		err = HCI_ReadEncryptionMode(fd, &val);
		print_hci_error(err);
		printf("\tEncryption: %s\n", (val==0x00)?"disabled":(val==0x01)?"point-to-point":"broadcast");

		printf("\tSecurity mode: %d\n", da.sec_mode);
		printf("\tPacket types: %x\n", da.pkt_type);

		printf("\tConnect role: %s\n", (da.role==ROLE_ALLOW_SWITCH)?"allow switch":"master only");
		printf("\tAccept role: %s\n", (da.role==ROLE_BECOME_MASTER)?"become master":"remain slave");
		// Read buffer information
		err = HCI_ReadBufferSize(fd, &acl_len, &sco_len, &acl_num, &sco_num);
		print_hci_error(err);
		printf("\tBuffers:\n");
		printf("\t\tACL: %d x %d bytes\n", acl_num, acl_len);
		printf("\t\tSCO: %d x %d bytes\n", sco_num, sco_len);
		// Read Supported Features
		err = HCI_ReadLocalSupportedFeatures(fd, &lmp_features);
		print_hci_error(err);
		printf("\tSuported features:\n");
		if (lmp_features & (LF_3SLOTS|LF_5SLOTS)) {
			printf("\t\tPacket types: ");
			if (lmp_features & LF_3SLOTS)
				printf("3-slots, ");
			if (lmp_features & LF_5SLOTS)
				printf("5-slots");
		}
		if (lmp_features & (LF_PAGING_SCHEME|LF_CQD_DATARATE|LF_RSSI)) {
			printf("\n\t\tRadio features: ");
			if (lmp_features & LF_PAGING_SCHEME)
				printf("paging scheme, ");
			if (lmp_features & LF_RSSI)
				printf("RSSI, ");
			if (lmp_features & LF_CQD_DATARATE)
				printf("Channel quality driven data rate");
		} else
			printf("\n\t\tRadio features: none");

		if (lmp_features & (LF_SWITCH|LF_HOLD_MODE|LF_SNIFF_MODE|LF_PARK_MODE)) {
			printf("\n\t\tPolicy: ");
			if (lmp_features & LF_SWITCH)
				printf("switch, ");
			if (lmp_features & LF_HOLD_MODE)
				printf("hold mode, ");
			if (lmp_features & LF_SNIFF_MODE)
				printf("sniff mode, ");
			if (lmp_features & LF_PARK_MODE)
				printf("park mode ");
		} else
			printf("\n\t\tPolicy: not supported");

		printf("\n\t\tEncryption: %s",
				(lmp_features&LF_ENCRYPTION)?"supported":"not supported");

		if (lmp_features & (LF_SLOT_OFFSET|LF_TIMING_ACCURACY)) {
			printf("\n\t\tClock modes: ");
			if (lmp_features & LF_SLOT_OFFSET)
				printf("slot offset, ");
			if (lmp_features & LF_TIMING_ACCURACY)
				printf("timing accuracy");
		} else
			printf("\n\t\tClock modes: none");

		if (lmp_features & (LF_SCO|LF_HV2|LF_HV3|LF_ULAWLOG|LF_ALAWLOG|LF_CVSD)) {
			printf("\n\t\tAudio: ");
			if (lmp_features & LF_SCO)
				printf("SCO link, ");
			if (lmp_features & LF_HV2)
				printf("HV2, ");
			if (lmp_features & LF_HV3)
				printf("HV3, ");
			if (lmp_features & LF_ULAWLOG)
				printf("u-Law log, ");
			if (lmp_features & LF_ALAWLOG)
				printf("a-Law log, ");
			if (lmp_features & LF_CVSD)
				printf("CVSD, ");
			if (lmp_features & LF_TRANSPARENT_SCO)
				printf("transparent SCO");
		} else 
			printf("\n\t\tAudio: not supported");

		printf("\n\t\tPower Control: %s",
				(lmp_features&LF_POWER_CONTROL)?"supported":"not supported");

		printf("\n\n");
		hci_close(fd);
	}

	return 0;
}

int cmd_test(struct btctl_command *cmd)
{
#if 0
	{
		struct Write_Current_IAC_LAP	cmd;
		struct Inquiry			event;
		INQUIRY_ITEM			Item;
		printf("sizeof(lap) = %d, sizeof(cmd) = %d, sizeof(event) = %d, sizeof(Item) = %d\n",
				sizeof(cmd.IAC_LAP[0]), sizeof(cmd), sizeof(event), sizeof(Item));
		printf("bda: %p, cod: %p\n", &Item.bda, &Item.bda+1);
		return 0;
	
	}
#endif
	return 0;
}

int cmd_disabled(int cmd)
{
	printf("\n-->> Command disabled at compilation time. Enable certain configuration options. <<--\n\n");
	return 0;
}

struct btctl_command cmds[] = {
	{0, 0, 0, 0, "---->>>> General commands <<<<----\n"},
	{"help", cmd_help, 0, "<command name>"},
	{"debug", cmd_debug, 0, "[+|-][<module>|<type>|<detail>|all]",
		"\tmodule:\thcicore|afhci|hcisched|hcimgr|hcilib|hci|\n"
		"\t\tpl2cap|afl2cap|l2cap|\n"
		"\t\tprfcomm|afrfcomm|bty|rfcomm|\n"
		"\t\tusb|uart|drv | allmod\n"
		"\ttype  : init|xmit|xmite|recv|recve|ctrl|ctrle|timer|allerror|alltype\n"
		"\tdetail: dump|chardump|parse|fname|func|alldetail\n"
	},
	{"list", cmd_inquiry, 2,
	},
	{"flush", cmd_inquiry, 3,
	},
	{"prompt", cmd_prompt, 0,
	},
	{"ftp", cmd_prompt, 0, "", "command prompt mode"
	},
	/* security */
	{0, 0, 0, 0, "---->>>> Security commands <<<<----\n"},
	{"pincode", cmd_pincode, 1,"[<address>|default] <pin>"
	},
	{"addpin", cmd_pincode, 1, "[<address>|default] <pin>"
	},
	{"rmpin", cmd_pincode, 0, "[<address>|default]"
	},
	{"unbond", cmd_key, 0, "[<address>]",
		"remove the key\n",
	},
	{"sec_mode", cmd_security, 0, "<mode>",
		"set security mode.\n",
	},
	/* HCI */
	{0, 0, 0, 0, "---->>>> HCI commands <<<<----\n"},
	{"inquiry", cmd_inquiry, 1, "<length>"
	},
	{"bdaddr", cmd_bdaddr, 1,
	},
	{"name", cmd_name, 1, "[<name>]"
	},
	{"scan", cmd_scan, 1, "[+|-][disc|conn]"
	},
	{"remotename", cmd_remotename, 1, "<bda>"
	},
	{"discovery", cmd_discovery, 1, "<length>"
	},
	{"role", cmd_role, 1, "<allow|deny> <master|slave>"
	},
	{"class", cmd_class, 1, "[<class> | <major> <minor> <service1> ... <servicen>]",
		"\tmajor:\tcomputer|phone|audio\n"
		"\tminor:\tdesktop|laptop|server|lap\n"
		"\tservice:\tnetworking,information,rendring,capturing,audio,transfer,telephony\n"
	},
	{"auth", cmd_auth, 1, "[on|off]"
	},
	{"authreq", cmd_auth, 2, "<address>",
		"authenticate remote device\n"
	},
	{"encrypt", cmd_encrypt, 1, "[off|ppp|bcast]"
	},
	{"pkt_type", cmd_pkttype, 0, "[<0xXXX> | <mnemonic>]",
		"\tmnemonic: DM1 DH1 DM3 DH3 DM5 DH5 HV1 HV2 HV3>]\n"
	},
	{"page_to", cmd_control, 0, "<to>",
		"set page timeout\n"
	},
	{"hdisc", cmd_control, 1,
	},
	{"hconnect", cmd_control, 2,
	},
	{"reset", cmd_reset, 0,
	},
	{"ping", cmd_ping, 0, "<bda> <size>",
		"Send ping packet to remote device\n"
	},
	{"speed", cmd_ping, 1,
	},
	/* UART */
#if defined(CONFIG_AFFIX_UART)
	{0, 0, 0, 0, "---->>>> UART commands <<<<----\n"},
	{"open_uart", cmd_uart, 1, "<name> <vendor> [speed] [flags]",
		"\tvendor: any|csr|ericsson|...]\n"
	},
	{"close_uart", cmd_uart, 0, "<name>"
	},
#else
#endif
	/* AUDIO */
#if defined(CONFIG_AFFIX_AUDIO)
	{0, 0, 0, 0, "---->>>> AUDIO commands <<<<----\n"},
	{"path", cmd_audio, 0,
	},
	{"audio", cmd_audio, 1,
	},
	{"scoflow", cmd_audio, 2,
	},
	{"addsco", cmd_addsco, 0,
	},
	{"listen", cmd_listen, 0,
	},
	{"play", cmd_play, 0,
	},
	{"record", cmd_record, 0,
	},
#else
#endif
	/* RFCOMM */
#if defined(CONFIG_AFFIX_RFCOMM)
	{0, 0, 0, 0, "---->>>> RFCOMM commands <<<<----\n"},
	{"connect", cmd_connect, 1, "<address> [<channel> | [service_type]]",
		"\tservice_type: SERial | DUN | LAN | HEAdset. (Serial if omitted)\n"
	},
	{"disconnect", cmd_disconnect, 1, "[line]"
	},
	{"status", cmd_status, 1,
	},
#else
#endif
	/* SDP */
#if defined(CONFIG_AFFIX_SDP)
	{0, 0, 0, 0, "---->>>> SDP commands <<<<----\n"},
	{"browse", cmd_browse, 0, "<address>"
	},
	{"search", cmd_search, 0, "<address>"
	},
	{"port", cmd_search, 1,
	},
#else
#endif
	/* OBEX */
#if defined(CONFIG_AFFIX_OBEX)
	{0, 0, 0, 0, "---->>>> OBEX commands <<<<----\n"},
	{"open", cmd_open, 0, "<address> [<channel>]"
	},
	{"close", cmd_close, 0,
	},
	{"ls", cmd_ls, 0, "[<address> [<channel>]]"
	},
	{"put", cmd_put, 0, "[<address> [<channel>]] <file name>"
	},
	{"get", cmd_get, 0, "[<address> [<channel>]] <file name>"
	},
	{"push", cmd_push, 0, "<address> [<channel>] <file name>"
	},
	{"rm", cmd_rm, 0, "[<address> [<channel>]] <file name>"
	},
	{"cd", cmd_cd, 0, "<dir name>"
	},
	{"mkdir", cmd_mkdir, 0, "<dir name>"
	},
#else
#endif
	/* PAN */
#if defined(CONFIG_AFFIX_PAN)
	{0, 0, 0, 0, "---->>>> PAN commands <<<<----\n"},
	{"pan", cmd_pan, 0, "<interface> [m (start_addr stop_addr)* ] [p (range_start range_stop)* ]",
		"Examples: lists all filters of PAN device eth0:\n"
		"\t-> panctl eth0\n"
		"\tset multicast filter to 03:00:00:20:00:00 only:\n"
		"\t-> panctl eth0 m 03:00:00:20:00:00 03:00:00:20:00:00\n"
		"\tset protocol filter to range 0x800 - 0x801 and 0x806\n"
		"\t-> panctl eth0 p 0x800 0x801 0x806 0x806\n"
		"\treset protocol filter\n"
		"\t-> panctl eth0 p\n"
	},
#else
#endif
	{"test", cmd_test, 0,
	},
	{0, 0, 0, NULL}
};

int cmdnum = sizeof(cmds)/sizeof(cmds[0]);

void print_help(struct btctl_command *cmd)
{
	if (cmd->name)
		printf("%s %s\n", cmd->name, cmd->arg?cmd->arg:"");
	if (cmd->msg)
		printf("%s", cmd->msg);

}

void print_usage(struct btctl_command *cmd)
{
	printf("usage: %s %s\n", cmd->name, cmd->arg?cmd->arg:"");
}

int cmd_help(struct btctl_command *cmd)
{
	int	i;

	if (!argv[argind]) {
		print_usage(cmd);
		exit(0);
	}

	for (i = 0; i < cmdnum; i++) {
		if (cmds[i].name)
			if (strcmp(cmds[i].name, argv[argind]) == 0) {
				print_help(&cmds[i]);
				return 0;
			}
	}
	printf("invalid command: %s\n", argv[argind]);
	return 	0;
}

void usage(void)
{
	int	i;
	
	printf("Usage: btctl [-i <name>] [--nosdp | -s] [<command> [parameters..]>\n");

	for (i = 0; i < cmdnum; i++) {
		if (cmds[i].name)
			printf("%s %s\n", cmds[i].name, cmds[i].arg?cmds[i].arg:"");
		else if (cmds[i].msg)
			printf("%s", cmds[i].msg);
	}

	printf("\n"
		"Notes:\n"
		"\t1. *--nosdp* option required to manually define server channel\n"
		"\t2. use *search* command if *browse* returns nothing\n"
		"\t3. argument *channel* is not used in SDP mode\n"
		"\t4. arguments *address* and *channel* are not used with\n"
		"\t   commands: *ls*, *get*, *put*, *rm* in FTP (prompt) mode\n"
		"\n"
		);
}

/* Command line mode */
static inline int white_space(char c)
{
	return (c == ' ' || c == '\t');
}
static inline int end_line(char c)
{
	return (c == '\n' || c == '\0');
}

static inline char *find_non_white(char *str)
{
	while (white_space(*str))
		str++;
	return (!end_line(*str)?str:NULL);
}

static inline char *find_white(char *str)
{
	while (!end_line(*str) && !white_space(*str))
		str++;
	return str;
}

static inline char *find_char(char *str, char c)
{
	while (*str && *str != c)
		str++;
	return (*str?str:NULL);
}

char *get_string(char **str)
{
	char	*first, *next;
	int	quotes = 0;

	if (*str == NULL)
		return NULL;
	first = find_non_white(*str);
	if (first == NULL)
		return NULL;

	if (*first == '"') {
		quotes = 1;
		first++;
	}
	next = first;
	if (quotes) {
		next = find_char(next, '"');
		if (!next) {
			printf("parse error\n");
			return NULL;
		}
		*next = '\0';
		next++;
	} else {
		next = find_white(next);
		if (*next) {
			*next = '\0';
			next++;
		}
	}

	if (*next)
		next = find_non_white(next);

	*str = next;
	return first;
}


char **parse_command_line(char *str)
{
	static char	*_argv[32];	// for 32 arguments
	char	*arg;
	int	i;

	_argv[0] = argv[0];
	for (i = 1; (arg = get_string(&str)) ; i++) {
		_argv[i] = arg;
		//printf("argv[%d]: %s\n", i, arg);
	}
	_argv[i] = NULL;

	return (i>1)?_argv:NULL;
}

sigjmp_buf	sigj;

void signal_handler(int sig)
{
	if (sig == SIGINT) {
		printf("\n");
		siglongjmp(sigj, 0);
	}
	BTINFO("Sig handler : %d", sig);
#if defined(CONFIG_AFFIX_OBEX)
	cmd_close(0);
#endif
	printf("Terminating on signal (%d) ...\n", sig);
	exit(0);
}

int cmd_prompt(struct btctl_command *_cmd)
{
	char	cmd_line[80], *str, **cmd;
	int	err;

	btlog_mask |= BTDEBUG_MODULE_OBEX;
	btdaemon = 1;
	ftpmode = 1;

	/* install signal handlers to unregister services from SDP */
	signal(SIGCHLD, SIG_IGN);
	signal(SIGINT, signal_handler);		// CTRL-C
	signal(SIGTERM, signal_handler);
	signal(SIGABRT, signal_handler);
	signal(SIGQUIT, signal_handler);

	printf("Affix version: %s\n", affix_version);
	printf("Wellcome to OBEX ftp. Type ? for help.\n");
	printf("Mode: %s\n", linkmode?"TCP/IRDA":"Bluetooth");
	printf("SDP: %s\n", sdpmode?"yes":"no");
	for (;;) {
		sigsetjmp(sigj, 1);
		printf("ftp> ");
		// get a line
		str = fgets(cmd_line, 80, stdin);
		//str = readline("ftp> ");
		if (str == NULL) {
			//printf("\n");
			break;
		}
		//printf("got: %s, len: %d\n", str, strlen(str));
		cmd = parse_command_line(str);
		if (cmd == NULL)
			continue;
		argv = cmd++;
		argind = 2;
		if (strcasecmp(*cmd, "quit") == 0) {
			printf("Bye!!!\n");
			break;
		} else if (strcasecmp(*cmd, "mode") == 0) {
			if (argv[argind]) {
				if (strcasecmp(argv[argind], "tcp") == 0)
					linkmode = LINK_TCP;
				else if (strcasecmp(argv[argind], "irda") == 0)
					linkmode = LINK_IRDA;
				else
					linkmode = LINK_BLUETOOTH;
			}
			printf("Mode: %s\n", linkmode?"TCP/IRDA":"Bluetooth");
		} else if (strcasecmp(*cmd, "lls") == 0) {
			system("ls");
		} else if (strcasecmp(*cmd, "lcd") == 0) {
			char	*name;

			if (argv[argind] == NULL) {
				printf("No dir name.\n");
				continue;
			}
			name = argv[argind];
			chdir(name);
		} else if (strcasecmp(*cmd, "debug") == 0) {
			btlog_mask |= BTDEBUG_MODULE_OBEX;
			btdaemon = 1;
		} else if (strcasecmp(*cmd, "?") == 0 || strcasecmp(*cmd, "help") == 0) {
			usage();
		} else if (strcasecmp(*cmd, "sdp") == 0) {
			sdpmode = !sdpmode;
			printf("sdp: %s\n", sdpmode?"yes":"no");
		} else {
			int	i;
			for (i = 0; i < cmdnum || (printf("Invalid command.\n") && 0); i++)
				if (cmds[i].name)
					if (strcmp(cmds[i].name, *cmd) == 0) {
						err = cmds[i].func(&cmds[i]);
						break;
					}
		}
	}

#if defined(CONFIG_AFFIX_OBEX)
	cmd_close(0);
#endif
	printf("Terminating ...\n");

	ftpmode = 0;
	btlog_mask &= ~BTDEBUG_MODULE_OBEX;

	return 0;
}


int main(int _argc, char **_argv)
{
	int	c, lind, i, err = 0;
	struct option	opts[] = {
		{"help", 0, 0, 'h'},
		{"tcp", 0, 0, 't'},
		{"nosdp", 0, 0, 's'},
		{0, 0, 0, 0}
	};

	/* set environment */
	argc = _argc;
	argv = _argv;
	home = getenv("HOME");
	user = getenv("USER");
	sprintf(confdir, "%s/affix", home);
	i = mkdir(confdir, 0755);
	if (i<0 && errno != EEXIST) {
		perror("Unable to create dir\n");
		return 1;
	}
	sprintf(cachefile, "%s/cache", confdir);
	inquiry_cache_load();

#if defined(CONFIG_AFFIX_SDP)
	sdpmode = 1;
#else
	sdpmode = 0;
#endif

	for (;;) {
		c = getopt_long(argc, argv, "+htrsi:", opts, &lind);
		if (c == -1)
			break;
		switch (c) {
			case 'h':
				usage();
				return 0;
				break;

			case 'i':
				strncpy(btdev, optarg, IFNAMSIZ);
				break;

			case 't':
				linkmode = LINK_TCP;
				break;

			case 'r':
				linkmode = LINK_IRDA;
				break;

			case 's':
				sdpmode = 0;
				break;

			case ':':
				printf("Missing parameters for option: %c\n", optopt);
				return 1;
				break;

			case '?':
				printf("Unknown option: %c\n", optopt);
				return 1;
				break;
		}
	}

	argind = optind;

	if (argv[argind] == NULL) {
		printf("Affix version: %s\n", affix_version);
		printf("---->>>> Affix Bluetooth Driver control program <<<<----\n"
				"For help use \"btctl -h\"\n"
		      );
		//usage();
		return status();
	}
	for (i = 0; i < cmdnum || (printf("Invalid command: %s\n", argv[argind]) && 0); i++)
		if (cmds[i].name)
			if (strcmp(cmds[i].name, argv[argind]) == 0) {
				argind++;
				err = cmds[i].func(&cmds[i]);
				break;
			}

	inquiry_cache_free();
			
	return err;
}


