%{
/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   */

/* 
   $Id: btsrv-parser.y,v 1.8 2002/07/24 15:46:41 kds Exp $

   parser for btsrv configuration file

	Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
*/

#include <affix/config.h>

#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/errno.h>

#include <fcntl.h>
#include <unistd.h>
#include <syslog.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <getopt.h>
#include <string.h>
#include <termios.h>

#include <affix/bluetooth.h>
#include <affix/btcore.h>
#include <affix/debug.h>

#include "btsrv.h"


struct
{
	char		name[MAX_PROFILE_NAME_LEN];
	sdpreg_func_t	reg_func;
} profiles[] =
#ifdef CONFIG_AFFIX_SDP
{
        {"SerialPort", sdpreg_serialport},
        {"DialupNetworking", sdpreg_dialupnetworking},
	{"LANAccess", sdpreg_lanaccess},
        {"OBEXFiletransfer", sdpreg_obex_filetransfer},
	{"OBEXObjectPush", sdpreg_obex_objectpush}
};
#else
{
        {"SerialPort", NULL},
        {"DialupNetworking", NULL},
	{"LANAccess", NULL},
        {"OBEXFiletransfer", NULL},
	{"OBEXObjectPush", NULL}
};
#endif

#define NUM_PROFILES ((int)(sizeof(profiles) / sizeof(profiles[0])))


//#define YYPARSE_PARAM	param
//#define YYLEX_PARAM	param

int yylex(void);
void yyerror(char *s);
void myyyerror(char *format, ...);

extern int yylineno;

char	*config_file;
int 	svcnum = 0;

int	waitsemi = 0;

%}

/* bison declaration */
//%defines
//%name-prefix="btsrv_"
//%output="btsrv-config.c"

%union {
	int	num;
	char	*str;
}

/* Configuration attribute tokens */
%token	DEVICE SERVICE OPT_PROFILE OPT_FLAGS OPT_EXEC


%token <num>	NUM
%token <str>	STRING WORDLIST WORD

%type <str>	name
%type <str>	string params

%start config

%%

config:	/* empty */
	| list
	;

list:	entry
	| list entry
	;

entry:	device
	| service
	;


device:
	'<' DEVICE name '>'
		devopt_list
	'<' '/' DEVICE '>'
	;


string:	WORD | STRING ;

params:	WORD | WORDLIST ;
	
name:	'*' { ; }
	| string
	;
	
optend:	{ waitsemi = 1; } ';';

devopt_list:	/*empty*/
		| devopt_list1
		;

devopt_list1:	devopt optend
		| devopt_list1 devopt optend
		;

devopt:	STRING {}
	;

service:
	'<' SERVICE string 
			{ 
				strcpy(services[svcnum].name, $3);
				BTDEBUG("found service: [%s]\n", $3);
			} 
		'>'
		svcopt_list
	'<' '/' SERVICE '>'	{ svcnum++; }
	;

svcopt_list:	/*empty*/
		| svcopt_list1
		;

svcopt_list1:	svcopt optend
		| svcopt_list1 svcopt optend
		;

svcopt:	
	OPT_PROFILE string
			{ 
				int	i;
				BTDEBUG("found profile: %s\n", $2); 
				for (i = 0; i < NUM_PROFILES; i++)
				if (strcasecmp(profiles[i].name, $2) == 0)
					break;
				if (i == NUM_PROFILES) {
					BTERROR("Invalid profile name %s in config file %s line %d",
						$2, config_file, yylineno);
					return -1;
				}
				services[svcnum].profile_reg_func = profiles[i].reg_func;
				// process flags
			}
	| OPT_FLAGS params
			{ 
				char	*fp, *ss = $2;
				BTDEBUG("found flag list: [%s]\n", $2);
				services[svcnum].flags = 0;
				for (fp = strtok(ss, ", "); fp; fp = strtok(NULL, ", ") ) {
					if (strcmp(fp, "tty") == 0) {
						services[svcnum].flags |= SRV_FLAG_RFCOMM_TTY;
					} else if (strcmp(fp, "std") == 0) {
						services[svcnum].flags |= SRV_FLAG_STD;
					} else if (strcmp(fp, "socket") == 0) {
					} else {
						BTERROR("Invalid flag [%s] in config file %s line %d",
						fp, config_file, yylineno);
						return -1;
					}
				}
			}
	| OPT_EXEC STRING 
			{ 
				strcpy(services[svcnum].cmd, $2);
				BTDEBUG("found exec: %s\n", services[svcnum].cmd);
			}
	;

		
%%

void yyerror(char *s)
{
	BTERROR("reading config file - %s at line %d\n", s, yylineno);
}

void myyyerror(char *format, ...)
{
	int	n;
	char	str[255];
	va_list	ap;

	va_start(ap, format);
	n = vsnprintf(str, sizeof(str), format, ap);
	va_end(ap);

	yyerror(str);
}
	
	

int btsrv_read_config(char *file, int *service_num)
{
	int			err;
	extern FILE		*yyin;
	
	config_file = file;
	yyin = fopen(config_file, "r");
	if (!yyin) {
		BTERROR("Unable to open config file %s", config_file);
                return -1;
	}
	svcnum = 0;
	err = yyparse();	/* parse config file */
	*service_num = svcnum;
	fclose(yyin);
	yyin = NULL;

	return err;
}

