/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2002 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: bthfp.c,v 1.4 2002/07/19 20:51:14 kds Exp $

   Hands Free profile client library

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

*/

#include <affix/config.h>

#include <glib.h>
#include <string.h>
#include <stdlib.h>
#include <libgen.h>

#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>

#include <stdio.h>
#include <unistd.h>
#include <errno.h>

#include <affix/bluetooth.h>
#include <affix/btcore.h>
#include <affix/debug.h>

// SDP stuff
#include <affix/sdp.h>
#include <affix/sdpsrv.h>
#include <affix/sdpclt.h>

// OBEX stuff
//#include <obex.h>
//#include "obex_server.h"
//#include "obex_client.h"

// Local include
#include <affix/bthfp.h>


/* internal declarations */
int hfp_service_register(hfp_t *handle);
int hfp_service_unregister(hfp_t *handle);


hfp_t *hfp_init(hfp_event_t eventcb, int flags)
{
	hfp_t		*handle = NULL;

	handle = malloc(sizeof(hfp_t));
	if (handle == NULL)
		return NULL;

	handle->eventcb = eventcb;

	return handle;
}

void hfp_cleanup(hfp_t *handle)
{
	hfp_service_unregister(handle);
	free(handle);
}

int getSupportedFeatures(ServiceRecord *svcRec, uint16_t *SupportedFeatures)
{
	GetBasicAttribute(AttrSupportedFeatures, SupportedFeatures, uint16);
}

int hfp_service_search(hfp_t *handle)
{
	int		status = -1, i, s, p;
	uint16_t	count = 0;
	GSList		*pSeq = NULL;
	GSList		*attrList = NULL;
	uint32_t	attr1 = AttrServiceRecordHandle;
	uint32_t	attr2 = AttrProtocolDescriptorList;
	uint32_t	attr3 = AttrSupportedFeatures;
	GSList		*searchList = NULL;
	UUID		uuid[2];
	AccessProtocols	*pAccessProtocols = NULL;
	SDPServerHandle srvHandle;
	ServiceRecord	*svcRec;


	status = SDPInit(0);
	if (status != E_OK)
		return status;

	srvHandle = SDPOpenServerConnection(handle->peer.bda);
	if (srvHandle < 0)
		return srvHandle;

	attrList = g_slist_append(attrList, &attr1);
	attrList = g_slist_append(attrList, &attr2);
	attrList = g_slist_append(attrList, &attr3);

	makeUUIDFrom16Bits(&uuid[0], AGHandsFreeServiceClassID);
	makeUUIDFrom16Bits(&uuid[1], SerialPortServiceClassID);

	for (s = 0; s < 2; s++ ) {
		searchList = g_slist_append(searchList, &uuid[s]);
		status = __SDPServiceSearchRequest(srvHandle, searchList, 1, &pSeq, &count);
		BTDEBUG("Status: %d, count: %d, len: %d", status, count, (int)g_slist_length(pSeq));
		if (searchList) {
			g_slist_free(searchList);
			searchList = NULL;
		}
		if ((status == 0) && (count > 0))
			break;
	}
	if ((status == 0) && (count > 0)) {
		for (i = 0; i < (int)g_slist_length(pSeq); i++) {
			ServiceRecordHandle svcRecHandle = *(ServiceRecordHandle*)g_slist_nth_data(pSeq, i);
			BTDEBUG("Geting attributes, handle: %#x", svcRecHandle);
			status = __SDPServiceAttributeRequest(srvHandle, svcRecHandle,
					IndividualAttributes, attrList, 0xFFFF, &svcRec, &count);
			if (status != 0)
				goto exit;

			if (getAccessProtocols(svcRec, &pAccessProtocols) == E_OK) {
				for (s = 0; s < (int)g_slist_length(pAccessProtocols->pSequence); s++ ) {
					GSList *protDescSeq = (GSList*)g_slist_nth_data(pAccessProtocols->pSequence, s);
					for (p = 0; p < (int)g_slist_length(protDescSeq); p++ ) {
						ProtocolDescriptor *pPD = (ProtocolDescriptor*)g_slist_nth_data(protDescSeq, p);
						if (UUID2value(&pPD->protocolId) == RFCOMM_UUID) {
							if (pPD->portNumber) {
								handle->peer.port = pPD->portNumber;
							}
							goto next;
						}
					}
				}
			}
next:
			if (getSupportedFeatures(svcRec, &handle->SupportedFeatures)) {
				BTDEBUG("Features found: %hx", handle->peerSupportedFeatures);
			}

		}
	}
exit:
	SDPCloseServerConnection(srvHandle);
	
	return status?status:handle->peer.port;

}

/*
 * Register HandsFree service
 */
int hfp_service_register(hfp_t *handle)
{
	int		status = -1;
	ServiceRecord	*handleRec;
	GSList		*browseGroupList = NULL;
	UUID		*pBrowseRootUUID;
	UUID		uuidHandsFree, uuidSerialPort, uuidAudio;
	GSList		*SvcClassIdList = NULL;
	GSList		*profDescList = NULL;
	ProfileDescriptor *profDesc;
	AccessProtocols ap;
	GSList		*pSeq = NULL;
	ProtocolDescriptor *pDesc, *pDescNext;
	
	/*
	 ** First create a service record handle
	 */
	handleRec = createServiceRecord();
	if (handleRec == NULL) {
		BTERROR("createServiceRecord failed");
		return -1;
	}
	/* set Service Class ID List */ 
	makeUUIDFrom16Bits(&uuidHandsFree, HandsFreeServiceClassID);
	SvcClassIdList = g_slist_append(SvcClassIdList,&uuidHandsFree);
#if 1
	// set additional Class ID
	makeUUIDFrom16Bits(&uuidSerialPort, SerialPortServiceClassID);
	SvcClassIdList = g_slist_append(SvcClassIdList,&uuidSerialPort);
	makeUUIDFrom16Bits(&uuidAudio, HeadsetServiceClassID);
	SvcClassIdList = g_slist_append(SvcClassIdList,&uuidAudio);
//	GenericAudioServiceClassID
#endif
	status = setServiceClassID(handleRec, SvcClassIdList);
	if (status != 0) {
		BTERROR("setServiceClassID failed");
		return -1;
	}
	/* set Profile Desc List */
	profDesc = (ProfileDescriptor *)malloc(sizeof(ProfileDescriptor));
	profDesc->vnum = 0x0100;
	makeUUIDFrom16Bits(&profDesc->profileUUID, HandsFreeServiceClassID);
	profDescList = g_slist_append(profDescList, profDesc);
	if (setBTProfileDescriptorList(handleRec, profDescList) != 0) {
		BTERROR("setBTProfileDescriptorList failed");
		return -1;
	}
	/* set Browse Group List */ 
	pBrowseRootUUID = (UUID *)malloc(sizeof(UUID));
	memset(pBrowseRootUUID, 0, sizeof(UUID));
	makeUUIDFrom16Bits(pBrowseRootUUID, PublicBrowseGroup);
	browseGroupList = g_slist_append(browseGroupList, pBrowseRootUUID);
	if (setBrowseGroupList(handleRec, browseGroupList) != 0) {
		BTERROR("setBrowseGroupList failed");
		return -1;
	}

	/* set Protocol Desc List */
	pDesc = (ProtocolDescriptor *)malloc(sizeof(ProtocolDescriptor));
	makeUUIDFrom16Bits(&pDesc->protocolId, L2CAP_UUID);
	pDesc->portNumber = 0xffff;
	pDesc->versionNumber = 0xffff;

	pDescNext = (ProtocolDescriptor *)malloc(sizeof(ProtocolDescriptor));
	makeUUIDFrom16Bits(&pDescNext->protocolId, RFCOMM_UUID);
	pDescNext->portNumber = handle->local.port;
	pDescNext->portNumberSize = 1;
	pDescNext->versionNumber = 0xffff;

	ap.pSequence =NULL;
	pSeq = g_slist_append(pSeq, pDesc);
	pSeq = g_slist_append(pSeq, pDescNext);
	ap.pSequence = g_slist_append(ap.pSequence,pSeq);
	if (setAccessProtocols(handleRec, &ap) != 0) {
		BTERROR("setAccessProtocols failed");
		return -1;
	}


	if (setInformationalAttributes(handleRec, handle->name, NULL, NULL) != 0) {
		BTERROR("setInformationalAttributes failed");
		return -1;
	}

	/* set Supported Features */
	addBasicAttribute(handleRec, UnsignedInteger_16, &handle->SupportedFeatures, 
			AttrSupportedFeatures);

	if (registerServiceRecord(handleRec) != 0) {
		BTERROR("registerServiceRecord failed");
		return -1;
	}

	handle->svcRec = handleRec;

	return 0;
}


int hfp_service_unregister(hfp_t *handle)
{
	int	status;
	
	if (handle->svcRec) {
		status = deleteServiceRecord(handle->svcRec);
		switch (status) {
			case E_OK:
				BTINFO("service %s deleted", handle->name);
				break;
			case E_INVALID_ARGS:
				BTERROR("Error: You cannot delete service %s", handle->name);
				break;
			default:
				BTERROR("Something went wrong. deleteServiceRecord() returned: %d", status);
		}
		handle->svcRec = NULL;
	}
 
	return 0;
}


/*
 * Create 
 */
int hfp_server_register(hfp_t *handle)
{
	int			err, fd;
	struct	sockaddr_affix	sa;
	socklen_t		len = sizeof(sa);

	fd = socket(PF_AFFIX, SOCK_STREAM, BTPROTO_RFCOMM);
	if (fd < 0) {
		BTERROR("Unable to create RFCOMM socket");
		return -1;
	}
	
	if (fcntl(fd, F_SETFL, O_NONBLOCK) != 0) {
		BTERROR("Unable to set non blocking mode for RFCOMM socket");
		return -1;
	}

	sa.family = AF_AFFIX;
	sa.port = 0;		/* dynamically allocated */
	sa.bda = BDADDR_ANY;
	err = bind(fd, (struct sockaddr*)&sa, sizeof(sa));
	if (err < 0) {
		BTERROR("Unable to bind address");
		return -1;
	}
	
	err = getsockname(fd, (struct sockaddr*)&sa, &len);
	if (err < 0) {
		BTERROR("Unable to get address info");
		return -1;
	}

	BTINFO("Bound service \"%s\" to server channel %d", handle->name, sa.port);

	err = listen(fd, 5);
	if (err < 0) {
		BTERROR("Unable to listen for connection requests");
		return -1;
	}

	handle->sfd = fd;
	handle->local = sa;

#if defined(CONFIG_AFFIX_SDP)
	err = SDPInit(SDP_SVC_PROVIDER);
	if (err) {
		BTERROR("Unable to initialize SDP");
		return -1;
	}
	if (sdpmode && hfp_service_register(handle) < 0 )
		return -1;
#endif	
	return 0;

	
}

int hfp_connect(hfp_t *handle, void *addr)
{
	int		err, fd;
	char		str[20];
	BD_ADDR		bda;
	int		sch;
	struct sockaddr_affix	*sa = addr;

	bda = sa->bda;
	handle->peer = *sa;

	bda2str(str, &bda);
	BTDEBUG("Connecting to host %s ...", str);

#if defined(CONFIG_AFFIX_SDP)
	if (sa->port == 0) {
		sch = hfp_service_search(handle);
		if (sch > 0)
			BTDEBUG("Service found on channel %d", sch);
		else if ( sch == 0 ) {
			BTDEBUG("Service is not available");
			return -1;
		} else {
			BTDEBUG("Unable to get service channel: %d", sch);
			return -1;
		}
	} else
#endif	
		sch = sa->port;

	sleep(1);
	fd = socket(PF_AFFIX, SOCK_STREAM, BTPROTO_RFCOMM);
	if (fd < 0) {
		BTDEBUG("Unable to create RFCOMM socket: %d", PF_AFFIX);
		return -1;
	}

	handle->peer.port = sch;
	handle->peer.local = BDADDR_ANY;
	err = connect(fd, (struct sockaddr*)&handle->peer, sizeof(*sa));
	if (err < 0) {
		BTDEBUG("Unable to connect to remote side");
		close(fd);
		return -1;
	}
	handle->fd = fd;
	
	return 0;
}

int hfp_handle_input(hfp_t *handle)
{
	int	status, timeout = -1;
	
	struct pollfd	fds[2];

	fds[SERVERFD].fd = handle->sfd;
	fds[SERVERFD].events = POLLIN;

	fds[CLIENTFD].fd = handle->fd;
	fds[CLIENTFD].events = POLLIN;
	

	status = poll(fds, 2, timeout);
	if (status > 0) {
		if ((fds[CLIENTFD].revents | fds[SERVERFD].revents) & (POLLERR | POLLHUP | POLLNVAL))
			;
		else {
			if (fds[SERVERFD].revents) {
				if (fds[SERVERFD].revents & POLLIN) { /* incomming connection */
				}
			}
			if (fds[CLIENTFD].revents) {
				if (fds[CLIENTFD].revents & POLLIN) {/* data from peer */
				}
			}
		}
	} else if (status < 0)
		return status;

	return 0;
}

