/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: btcore.c,v 1.7 2002/07/25 15:35:09 kds Exp $

   HCI Command Library

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
*/

#include <asm/byteorder.h>

#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/resource.h>
#include <sys/errno.h>
#include <sys/uio.h>

#include <fcntl.h>
#include <unistd.h>
#include <syslog.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#include <linux/types.h>


/* affix specific */
#include <affix/config.h>
#include <affix/bluetooth.h>
#include <affix/btcore.h>

/* Library variables */
int 		argc;
char		**argv;
int 		btdaemon = 0;
char		*affix_version = PACKAGE_STRING;
unsigned long	btlog_mask;		
int		btlog_options = 0;

char		btdev[IFNAMSIZ];
int		linkmode = LINK_BLUETOOTH;
int		sdpmode = 1;



/* ERROR messages */
char *hci_errlist[] = {
	"NO ERROR",
	"Unknown HCI command",
	"No connection",
	"Hardware failure",
	"Page timeout",
	"Authentication failure",
	"Key missing",
	"Memory full",
	"Connection timeout",
	"Max number of connections",
	"Max number of SCO connections to a device",
	"ACL connection already exists",
	"Command disallowed",
	"Host rejected due to limited resources",
	"Host rejected due to security reason",
	"Host rejected due to remote device is only a personal device",
	"Host timeout",
	"Unsupported feature or parameter value",
	"Invalid HCI command parameters",
	"Other end terminated connection: user ended connection",
	"Other end terminated connection: low resources",
	"Other end terminated connection: about to power off",
	"Connection terminated by local host",
	"Repeated attempts",
	"Pairing not allowed",
	"Unknow LMP PDU",
	"Unsupported remote feature",
	"SCO offset rejected",
	"SCO interval rejected",
	"SCO air mode reejected",
	"Invalid LMP parameters",
	"Unspecified error",
	"unsupported LMP parameters value",
	"Role change not allowed",
	"LMP response timeout",
	"LMP error transaction collision",
	"LMP PDU not allowed",
	NULL
};

char *lmp_compid[] = {
	"Ericsson Mobile Communications",
	"Nokia Mobile Phones",
	"Intel Corp",
	"IBM Corp",
	"Toshiba Corp",
	"3Com",
	"Microsoft",
	"Lucent",
	"Motorola",
	"Infineon Technologies AG",
	"Cambridge Silicon Radio",
	"Silicon Wave",
	"Digianswer",
	"Texas Instruments Inc.",
	"Parthus Technologies Inc.",
	"Broadcom Corporation",
	"Mitel Semiconductor",
	"Widcomm Inc.",
	"Telencomm Inc.",
	"Atmel Corporation",
	"Mitsubishi Electric Corporation",
	"RTX Telecom A/S",
	"KC Technology Inc.",
	"Newlogic",
	"Transilica Inc.",
	"Rohde & Schwartz GmbH &Co. KG",
	"TTPCom Limited",
	"Signia Technologies Inc.",
	"Conexant Systems Inc.",
	"Qualcomm",
	"Inventel",
	"AVM Berlin",
	"BrandSpeed Inc.",
	"Mansella Ltd",
	"NEC Corporation",
	"WavePlus Technology Co., Ltd.",
	"Alcatel",
	NULL
};

char *lmp_features[] = {

	NULL
};


/* devices classes */
struct affix_tupla codServiceClassMnemonic[] =  {
	{COD_NETWORKING, "Networking", "netw"},
	{COD_RENDERING, "Rendering", "rend"},
	{COD_CAPTURING, "Capturing", "capt"},
	{COD_TRANSFER, "Object Transfer", "tran"},
	{COD_AUDIO, "Audio", "audi"},
	{COD_TELEPHONY, "Telephony", "tele"},
	{COD_INFORMATION, "Information", "info"},
	{0, NULL, NULL}
};

struct affix_tupla codMajorClassMnemonic[] =  {
	{COD_MISC, "Miscellaneous", "misc"},
	{COD_COMPUTER, "Computer", "comp"},
	{COD_PHONE, "Phone", "phon"},
	{COD_LAP, "LAN Access Point", "lap"},
	{COD_MAUDIO, "Audio", "audi"},
	{COD_PERIPHERAL, "Peripheral", "peri"},
	{0, NULL, NULL}
};

struct affix_tupla codMinorComputerMnemonic[] =  {
	{COD_DESKTOP, "Desktop", "desk"},
	{COD_COMPUTER, "Server", "serv"},
	{COD_LAPTOP, "Laptop", "lapt"},
	{COD_HANDPC, "Handheld PC/PDA", "hand"},
	{COD_PALMPC, "Palm PC/PDA", "palm"},
	{0, NULL, NULL}
};

struct affix_tupla codMinorPhoneMnemonic[] =  {
	{COD_CELLULAR, "Cellular", "cell"},
	{COD_CORDLESS, "Cordless", "cord"},
	{COD_SMART, "Smart phone", "smart"},
	{COD_MODEM, "Wired Modem/VoiceGW", "modem"},
	{0, NULL, NULL}
};

struct affix_tupla codMinorAudioMnemonic[] =  {
	{COD_HEADSET, "Headset", "head"},
	{0, NULL, NULL}
};

void hcierror(int err)
{
	if (err) {
		if (err < 0)
			perror("System error");
		else
			printf("HCI error: %s\n", hci_errormsg(err));
	}
}


/* general purpose */
int str2bda(BD_ADDR *p, char *str)
{
	int	i, val, err;
	char	*res=(char*)p;

	for (i = 5; i >= 0; i--) {
		err = sscanf(str, "%2x", &val);
		if (err == 0)
			return 0;
		res[i] = val;
		if (i == 0)
			break;
		str = strchr(str,':');
		if (str == 0)
			return 0;
		str++;
	}
	return 1;
}

void bda2str(char *str, BD_ADDR *p)
{
	__u8	*ma=(__u8*)p;

	sprintf(str, "%02x:%02x:%02x:%02x:%02x:%02x", 
		ma[5], ma[4], ma[3], ma[2], ma[1], ma[0]);

}

void btlog_printf(int priority, const char *format, ...)
{
	static char buf[1024];
	va_list list;
	
	va_start(list, format);
	vsprintf(buf, format, list);

	if (btdaemon) {
		openlog(argv[0], btlog_options, LOG_DAEMON);
		syslog(priority, "%s", buf);
		closelog();
	} else {
		fprintf(stderr, "%s", buf);
	}
	va_end(list);
}

void btlog_hexdump(int priority, const char *fname, const unsigned char *data, int len)
{
	int 	i,j;
	char	buf[81];

	for (i = 0; i < len; i++) {
		if (((i % 16) == 0) && i > 0)
			btlog_printf(priority, "%s: %s\n", fname, buf);
		j = (i % 16) * 3;
		sprintf(&(buf[j]), "%02x ", data[i]);
	}
	if (len)
		btlog_printf(priority, "%s: %s\n", fname, buf);
}

int daemonize(void)
{
	int 		fd;
	struct rlimit	flim;
	if (getppid() != 1) {
		signal(SIGTTOU, SIG_IGN);
		signal(SIGTTIN, SIG_IGN);
		signal(SIGTSTP, SIG_IGN);
		if (fork() != 0) return 0;	/* exit if parent */
		setsid();
	}
	getrlimit(RLIMIT_NOFILE, &flim);
	for (fd = 0; fd < (int)flim.rlim_max; fd++)
		close(fd);
	chdir("/");

	return 1;
}

#include <affix/hci_cmds.h>	//XXX:


/* **********   Control Commands  ************** */

int hci_add_pin(BD_ADDR *bda, int Length, __u8 *Code)
{
	struct PIN_Code	pin;
	int		fd, err;

	pin.bda = *bda;
	pin.Length = Length;
	memcpy(pin.Code, Code, 16);

	fd = btsys_socket(PF_AFFIX, SOCK_RAW, BTPROTO_HCI);
	if (fd < 0) {
		return fd;
	}
	err = btsys_ioctl(fd, BTIOC_REMOVELINKKEY, bda);
	if (!err || errno == ENOENT )
		err = btsys_ioctl(fd, BTIOC_ADDPINCODE, &pin);
	btsys_close(fd);
	return err;
}

int hci_remove_pin(BD_ADDR *bda)
{
	int	fd, err;

	fd = btsys_socket(PF_AFFIX, SOCK_RAW, BTPROTO_HCI);
	if (fd < 0) {
		return fd;
	}
	err = btsys_ioctl(fd, BTIOC_REMOVEPINCODE, bda);
	btsys_close(fd);
	return err;
}

int hci_remove_key(BD_ADDR *bda)
{
	int	fd, err;

	fd = btsys_socket(PF_AFFIX, SOCK_RAW, BTPROTO_HCI);
	if (fd < 0) {
		return fd;
	}
	err = btsys_ioctl(fd, BTIOC_REMOVELINKKEY, bda);
	btsys_close(fd);
	return err;
}


int hci_set_dbmask(__u32 dbmask)
{
	int	fd, err;

	fd = btsys_socket(PF_AFFIX, SOCK_RAW, BTPROTO_HCI);
	if (fd < 0) {
		return fd;
	}
	err = btsys_ioctl(fd, BTIOC_DBMSET, &dbmask);
	btsys_close(fd);
	return err;
}

int hci_get_dbmask(__u32 *dbmask)
{
	int	fd, err;

	fd = btsys_socket(PF_AFFIX, SOCK_RAW, BTPROTO_HCI);
	if (fd < 0) {
		return fd;
	}
	err = btsys_ioctl(fd, BTIOC_DBMGET, dbmask);
	btsys_close(fd);
	return err;
}

int hci_open_uart(char *name, int type, int speed, int flags, int proto)
{
	int			fd, err;
	struct open_uart	line;
	struct stat		st;

	err = stat(name, &st);
	if (err)
		return err;

	if (!S_ISCHR(st.st_mode))
		return -1;

	line.dev = st.st_rdev;
	line.type = type;
	line.speed = speed;
	line.flags = flags;
	line.proto = proto;

	fd = btsys_socket(PF_AFFIX, SOCK_RAW, BTPROTO_HCI);
	if (fd < 0) {
		return fd;
	}
	err = btsys_ioctl(fd, BTIOC_OPEN_UART, &line);
	btsys_close(fd);
	return err;
}

int hci_close_uart(char *name)
{
	int			fd, err;
	struct open_uart	line;
	struct stat		st;

	err = stat(name, &st);
	if (err)
		return err;

	if (!S_ISCHR(st.st_mode))
		return -1;

	line.dev = st.st_rdev;
	line.flags = 0;

	fd = btsys_socket(PF_AFFIX, SOCK_RAW, BTPROTO_HCI);
	if (fd < 0) {
		return fd;
	}
	err = btsys_ioctl(fd, BTIOC_CLOSE_UART, &line);
	btsys_close(fd);
	return err;
}

/* L2CAP - user mode stuff */

/*
 * returns L2CAP channel (socket) MTU
 */
int l2cap_sendping(int fd, char *data, int size, int type)
{
	int		err = 0;
	struct msghdr	msg;
	struct iovec	iov;
	struct cmsghdr	*cmsg;
	char		buf[CMSG_SPACE(0)];
	int		len = size;

	while (len) {
		iov.iov_base = data;
		iov.iov_len = len;
		msg.msg_name = NULL;
		msg.msg_namelen = 0;
		msg.msg_iov = &iov;
		msg.msg_iovlen = 1;
		msg.msg_control = buf;
		cmsg = CMSG_FIRSTHDR(&msg);
		cmsg->cmsg_level = SOL_AFFIX;
		cmsg->cmsg_type = type;
		cmsg->cmsg_len = CMSG_LEN(0);
		msg.msg_controllen = cmsg->cmsg_len;	

		err = sendmsg(fd, &msg, 0);
		if (err < 0)
			return err;
		if (err == 0)
			return (size - len);
		len -= err;
	}
	return size;
}

int l2cap_ping(int fd, char *data, int size)
{
	int	err = 0;
	int	len = 0;

	err = l2cap_sendping(fd, data, size, L2CAP_PING);
	if (err < 0)
		return err;
	while (len < size) {
		err = recv(fd, data+len, size - len, 0);
		if (err < 0)
			return err;
		if (err == 0)
			break;
		len += err;
	}
	return len;
}



/* RFCOMM - user space stuff */
int rfcomm_disconnect(int line)
{
	int	fd;

	fd = socket(PF_AFFIX, SOCK_STREAM, BTPROTO_RFCOMM);
	if (fd < 0)
		return fd;

	return ioctl(fd, SIOCRFCOMM_DISCONNECTPORT, &line);
}

int rfcomm_get_ports(struct rfcomm_port *pi, int size)
{
	int	fd, err;
	struct rfcomm_ports	cp;
	
	fd = socket(PF_AFFIX, SOCK_STREAM, BTPROTO_RFCOMM);
	if (fd < 0)
		return fd;

	cp.info = pi;
	cp.size = size;
	err = ioctl(fd, SIOCRFCOMM_CHECKPORTS, &cp);
	if (err < 0)
		return err;

	return cp.count;
}


/* **************************  Control Commands ******************* */

int hci_get_devs(int *devs)
{
	int	fd, err;

	fd = btsys_socket(PF_AFFIX, SOCK_RAW, BTPROTO_HCI);
	if (fd < 0) {
		return fd;
	}

	err = btsys_ioctl(fd, BTIOC_GETDEVS, devs);
	btsys_close(fd);

	return err;
}

int hci_set_audio(int fd, int mode, __u16 setting)
{
	int	err;
	struct affix_audio	audio;

	if (setting) {
		err = HCI_WriteVoiceSetting(fd, setting);
		if (err)
			return err;
	}
	audio.mode = mode;
	audio.setting = setting;
	return btsys_ioctl(fd, BTIOC_SET_AUDIO, &audio);
}

int hci_get_conn(int fd, BD_ADDR *bda)
{
	struct hci_conn_info	ci;
	int	err;

	ci.bda = *bda;
	err = btsys_ioctl(fd, BTIOC_GET_CONN, &ci);
	if (err)
		return err;
	return ci.handle;
}

