/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   Based on the from Pontus Fuchs <pontus.fuchs@tactel.se>
   Copyright (c) 1999, 2000 Pontus Fuchs, All Rights Reserved.
   
   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/*
   $Id: obex_server.c,v 1.4 2002/08/01 12:43:15 kds Exp $

   OBEX server application for Affix

   Adapted for Affix, Fixes: 
   				Imre Deak <ext-imre.deak@nokia.com>
   				Dmitry Kasatkin
*/

#include <affix/config.h>

#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <getopt.h>

#include <dirent.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>

#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>

#include <glib.h>
#include <openobex/obex.h>

#include <affix/btcore.h>
#include "obex_server.h"

#define IR_SERVICE "OBEX"
#define OBEX_ROOT_PATH "/var/spool/affix"

char	*obex_root_path = NULL;
char	*home;
char	*user;
char	confdir[80];
uid_t	uid;



char *file_base_name(char *path)
{
	char *pos = strrchr(path, '/');
	return pos ? pos + 1 : path;
}

int change_dir(char *new_dir, int mode)
{
	char	new_path[PATH_MAX + 1], root_path[PATH_MAX + 1];
	char	last_comp[NAME_MAX + 1] = {0, };
	int	create_dir = 0;
	char	*p;

	if (new_dir == NULL || new_dir[0] == '\0')
		return -1;

	BTDEBUG("Changing directory to: [%s]\n", new_dir);

	if (realpath(new_dir, new_path) == NULL) {
		if ((mode & CHDIR_CREATE_IF_NOT_EXIST) && errno == ENOENT) {
			/*
			 * in case we want to create a non existent directory
			 * we should check the validity of the path without the last component
			 * so save the last component and strip it off
			 * don't strip the last component if it's ".." because this should be
			 * taken into account when checking against obex_root_path..
			 */
			char	buf[PATH_MAX + 1];
			strcpy(buf, new_dir);
			if ((p = strrchr(buf, '/')) == NULL)
				p = buf;
			else
				p++;
			if (strcmp(p, "..") != 0) {
				strcpy(last_comp, p);
				if (p == buf)
					strcpy(buf, "./");
				else
					*p = '\0';
			}
			if (realpath(buf, new_path)) {
				strcat(new_path, "/");
				strcat(new_path, last_comp);
			} else {
				if (!(mode & CHDIR_CREATE_RECURSIVE))
					return -1;
				strcpy(new_path, new_dir);
			}
			create_dir = 1;
		} else
			return -1;
	}
	
	BTINFO("Creating: %s, create: %s\n", new_path, create_dir?"yes":"no");
	if (!realpath(obex_root_path, root_path) || strncmp(new_path, root_path, strlen(root_path)) != 0)
		return -1;

	if (mode & CHDIR_CHECK_ONLY)
		return 0;
			
	if (create_dir)	{
		int	err;
		if (mode & CHDIR_CREATE_RECURSIVE)
			err = rmkdir(new_path, 0700);
		else
			err = mkdir(new_path, 0700);
		if (err && errno != EEXIST)
			return -1;
	}
	if (chdir(new_path) != 0)
		return -1;
		
	return 0;
}


//
// Read data from stream.
//
static gint readstream(obex_t *handle, obex_object_t *object)
{
	gint 		actual;
	obexsrv_t	*srv = OBEX_GetUserData(handle);
	const guint8	*buf;
	gint		len;
		
	actual = OBEX_ObjectReadStream(handle, object, &buf);
	if (actual > 0) {
		len = write(srv->sfd, buf, actual);
	} else if (actual == 0) {
		close(srv->sfd); srv->sfd = -1;
	}

	//BTDEBUG("Got %d bytes\n", actual);
	return actual;
}


void put_server(obex_t *handle, obex_object_t *object)
{
	obex_headerdata_t hv;
	guint8		hi;
	gint		hlen;
	const guint8	*body = NULL;
	gint 		body_len = 0;
	gchar		*name = NULL;
	gchar		*namebuf = NULL;
	char		*p;
	int		endofbody = 0;
	int		delete = 0;
	int		empty_obj = 0;
	gint		len, con_id = -1;
	gchar		*obj_type = NULL;
	obexsrv_t	*srv = OBEX_GetUserData(handle);
	

	BTDEBUG("");

	while(OBEX_ObjectGetNextHeader(handle, object, &hi, &hv, &hlen)) {
		switch (hi) {
		case OBEX_HDR_CONNECTION:
			con_id = hv.bq4;
			BTDEBUG("Found connection id: %08x", con_id);
			break;
		case OBEX_HDR_LENGTH:
			len = hv.bq4;
			BTDEBUG("Found Length: %d, hlen: %d", len, hlen);
			break;
		case OBEX_HDR_NAME:
			if( (namebuf = g_malloc(hlen / 2))) {
				OBEX_UnicodeToChar(namebuf, hv.bs, hlen);
				name = namebuf;
			}
			BTDEBUG("Found name: %s", name);
			break;
		case OBEX_HDR_TYPE:
			obj_type = (gchar *)hv.bs;
			BTDEBUG("Found type:%*s, hlen: %d", hlen, obj_type, hlen);
			break;
		case OBEX_HDR_BODY:
			BTDEBUG("Found body");
			body = hv.bs;
			body_len = hlen;
			break;
		case OBEX_HDR_BODY_END:
			BTDEBUG("Found end of body");
			endofbody = 1;
			break;

		default:
			BTDEBUG("Skipped header %02x", hi);
		}
	}
	if (!body) {
		if (endofbody) {
			BTDEBUG("Got a PUT with only end of body header->create empty object.");
			empty_obj = 1;
		} else {
			if (!srv->streamming) {
				BTDEBUG("Got a PUT without a body->delete");
				delete = 1;
			}
		}
	}
	if(!name) {
		if (delete || empty_obj) {
			BTERROR("name is missing.");
			OBEX_ObjectSetRsp(object, OBEX_RSP_BAD_REQUEST, OBEX_RSP_BAD_REQUEST);
			goto error;
		}
		name = "OBEX_PUT_Unknown_object";
		BTDEBUG("Got a put without a name. Setting name to %s", name);

	}
	
	// if path contains folders check if it is valid
	if ((p = strrchr(name, '/')) != NULL) {
		char	dir[PATH_MAX + 1];
		
		strncpy(dir, name, p - name);
		dir[p - name] = '\0';
		if (change_dir(dir, CHDIR_CHECK_ONLY) != 0) {
			BTERROR(" Invalid path.");
			OBEX_ObjectSetRsp(object, OBEX_RSP_INTERNAL_SERVER_ERROR, OBEX_RSP_INTERNAL_SERVER_ERROR);
			goto error;
		}
	}
	if (delete) {
		if (unlink(name) != 0 ) {
			if (errno == ENOENT) {
				OBEX_ObjectSetRsp(object, OBEX_RSP_NOT_FOUND, OBEX_RSP_NOT_FOUND);
				goto error;
			} else if( errno == EISDIR) {
				if (change_dir(name, CHDIR_CHECK_ONLY) != 0) {
					BTERROR(" Invalid dir.");
					OBEX_ObjectSetRsp(object, OBEX_RSP_INTERNAL_SERVER_ERROR, OBEX_RSP_INTERNAL_SERVER_ERROR);
					goto error;
				}
				rmdir(name);
			} else {
				OBEX_ObjectSetRsp(object, OBEX_RSP_INTERNAL_SERVER_ERROR, OBEX_RSP_INTERNAL_SERVER_ERROR);
				goto error;
			}
		}
	} else {
		if (!srv->streamming) {
			safe_save_file(name, body, body_len);
		} else {
			rename(srv->name, name);
		}
	}

	OBEX_ObjectSetRsp(object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
error:
	if (namebuf)
		g_free(namebuf);
	if (srv->streamming) {
		unlink(srv->name);
		free(srv->name); srv->name = NULL;
		srv->streamming = FALSE;
	}
}



#define CHUNK_SIZE 32		// power of 2!

// copy src to *buf increasing *buf if necessary
// *buf_space is the current size of *buf currently
// src_length = strlen(src), buf_length = strlen(*buf)
int copy_to_dyna_buf(char *src, int src_length, char **buf, int *buf_space, int *buf_length)
{
	if (src_length + *buf_length + 1 > *buf_space) {
		int new_size = (*buf_length + src_length + 1 + CHUNK_SIZE - 1) & (~(CHUNK_SIZE - 1));
		*buf = realloc(*buf, new_size);
		if (*buf == NULL)
			return -1;
		*buf_space = new_size;
	}
	memcpy((*buf) + *buf_length, src, src_length);
	*buf_length += src_length;
	(*buf)[*buf_length] = '\0';
	return 0;
}

void get_server_folderlisting(obex_t *handle, obex_object_t *object, gchar *name)
{
	static char header[] = 
		"<?xml version=\"1.0\"?>\n"
		"<!DOCTYPE folder-listing SYSTEM \"obex-folder-listing.dtd\">\n"
		"<folder-listing version=\"1.0\">\n";
	static char trailer[] =
		"</folder-listing>\r";

	obex_headerdata_t hv;
	
	DIR	*dir;
	char	line_buf[256 + PATH_MAX];
	char	*buf;
	int	buf_space, buf_len;
	int	pdir_writeable;
	struct stat	f_stat;
	
	buf = NULL;
	buf_space = 0;
	buf_len = 0;
	
	if (name[0] != '\0') {
		strcpy(line_buf, name);
		if (change_dir(name, CHDIR_CHECK_ONLY) != 0)
			goto error;
	} else {
		if (getcwd(line_buf, sizeof(line_buf)) == NULL)
			goto error;
	}

	dir = opendir(line_buf);
	if (dir == NULL)
		goto error;
		
	if (copy_to_dyna_buf(header, sizeof(header) - 1, &buf, &buf_space, &buf_len) != 0)
		goto error;
	
	if (stat(".", &f_stat) != 0)
		goto error;
	pdir_writeable = (geteuid() == f_stat.st_uid && (f_stat.st_mode & S_IWUSR)) ||
		     (getegid() == f_stat.st_gid && (f_stat.st_mode & S_IWGRP)) ||
		     (f_stat.st_mode & S_IWOTH);
		     
	for (;;) {
		int line_len;
		struct dirent	*dent;
		struct tm	*created, *modified, *accessed;
		char		user_perm[4];
		
		dent = readdir(dir);
		if (dent == NULL)
			break;
		if (stat(dent->d_name, &f_stat) != 0)
			goto error;
			
		created = localtime(&f_stat.st_ctime);
		modified = localtime(&f_stat.st_mtime);
		accessed = localtime(&f_stat.st_atime);
		
		user_perm[0] = '\0';
		if ((geteuid() == f_stat.st_uid && f_stat.st_mode & S_IRUSR) ||
		    (getegid() == f_stat.st_gid && f_stat.st_mode & S_IRGRP) ||
		    (f_stat.st_mode & S_IROTH))
			strcat(user_perm, "R");
			
		if ((geteuid() == f_stat.st_uid && f_stat.st_mode & S_IWUSR) ||
		    (getegid() == f_stat.st_gid && f_stat.st_mode & S_IWGRP) ||
		    (f_stat.st_mode & S_IWOTH))
			strcat(user_perm, "W");
			
		if (pdir_writeable)
			strcat(user_perm, "D");
		
		if (S_ISDIR(f_stat.st_mode) &&
		    (strcmp(dent->d_name, ".") == 0 ||
		     strcmp(dent->d_name, "..") == 0))
			continue;
		line_len = sprintf(line_buf, "<%s name=\"%s\" size=\"%lld\" user-perm=\"%s\" "
						"created=\"%04d%02d%02dT%02d%02d%02dZ\" "
						"modified=\"%04d%02d%02dT%02d%02d%02dZ\" "
						"accessed=\"%04d%02d%02dT%02d%02d%02dZ\"/>\n",
				(S_ISDIR(f_stat.st_mode) ? "folder" : "file"),
				file_base_name(dent->d_name),
				(long long unsigned)f_stat.st_size,
				user_perm,
				created->tm_year + 1900, created->tm_mon + 1, created->tm_mday,
					created->tm_hour, created->tm_min, created->tm_sec,
				modified->tm_year + 1900, modified->tm_mon + 1, modified->tm_mday,
					modified->tm_hour, modified->tm_min, modified->tm_sec,
				accessed->tm_year + 1900, accessed->tm_mon + 1, accessed->tm_mday,
					accessed->tm_hour, accessed->tm_min, accessed->tm_sec
				);
		if (copy_to_dyna_buf(line_buf, line_len, &buf, &buf_space, &buf_len) != 0)
			goto error;
	}
	if (copy_to_dyna_buf(trailer, sizeof(trailer) - 1, &buf, &buf_space, &buf_len) != 0)
		goto error;
	OBEX_ObjectSetRsp(object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
	hv.bq4 = buf_len + 1;
	OBEX_ObjectAddHeader(handle, object, OBEX_HDR_LENGTH, hv, sizeof(guint32), 0);
	hv.bs = buf;
	OBEX_ObjectAddHeader(handle, object, OBEX_HDR_BODY, hv, buf_len + 1, 0);
	free(buf);
	return;
error:
	OBEX_ObjectSetRsp(object, OBEX_RSP_NO_CONTENT, OBEX_RSP_NO_CONTENT);
	if (buf != NULL)
		free(buf);
	return;
}	

//
// Add more data to stream.
//
static gint fillstream(obex_t *handle, obex_object_t *object)
{
	gint 			actual;
	obexsrv_t		*srv = OBEX_GetUserData(handle);
	obex_headerdata_t	hv;
		
	actual = read(srv->sfd, srv->buf, OBEX_STREAM_CHUNK);
	//BTDEBUG("send %d bytes\n", actual);
	if(actual > 0) {
		/* Read was ok! */
		hv.bs = srv->buf;
		OBEX_ObjectAddHeader(handle, object, OBEX_HDR_BODY,
				hv, actual, OBEX_FL_STREAM_DATA);
	}
	else if(actual == 0) {
		/* EOF */
		hv.bs = srv->buf;
		close(srv->sfd); srv->sfd = -1;
		OBEX_ObjectAddHeader(handle, object, OBEX_HDR_BODY,
				hv, 0, OBEX_FL_STREAM_DATAEND);
	}
        else {
		/* Error */
		hv.bs = NULL;
		close(srv->sfd); srv->sfd = -1;
		OBEX_ObjectAddHeader(handle, object, OBEX_HDR_BODY,
				hv, 0, OBEX_FL_STREAM_DATA);
	}

	return actual;
}

void get_server_file(obex_t *handle, obex_object_t *object, char *name)
{
	int			file_size;
	obex_headerdata_t	hv;
	char			*p;
	obexsrv_t		*srv = OBEX_GetUserData(handle);
	
	if(name[0] == '\0') {
		BTDEBUG("Got a GET without a name-header! Get def object.");
		name = OBEX_ROOT_PATH"/nokia.vcf";
		//strcpy(name, "/var/spool/affix/nokia.vcf");
		//OBEX_ObjectSetRsp(object, OBEX_RSP_NOT_FOUND, OBEX_RSP_NOT_FOUND);
		//return;
	}

	BTDEBUG("Got a request for %s", name);
	
	if ((p = strrchr(name, '/')) != NULL) {
		char	dir[PATH_MAX + 1];
		
		strncpy(dir, name, p - name);
		dir[p - name] = '\0';
		if (change_dir(dir, CHDIR_CHECK_ONLY) != 0) {
			BTERROR(" Invalid path.");
			OBEX_ObjectSetRsp(object, OBEX_RSP_NOT_FOUND, OBEX_RSP_NOT_FOUND);
			return;
		}
	}

	srv->sfd = open(name, O_RDONLY, 0);
	if (srv->sfd < 0) {
		BTDEBUG("File not found: %s\n", name);
		OBEX_ObjectSetRsp(object, OBEX_RSP_NOT_FOUND, OBEX_RSP_NOT_FOUND);
		return;
	}

	file_size = get_filesize(name);
	
	srv->buf = g_malloc(OBEX_STREAM_CHUNK);
	if (!srv->buf) {
		close(srv->sfd); srv->sfd = -1;
		OBEX_ObjectSetRsp(object, OBEX_RSP_INTERNAL_SERVER_ERROR, OBEX_RSP_INTERNAL_SERVER_ERROR);
		return;	
	}

	OBEX_ObjectSetRsp(object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
	hv.bq4 = file_size;
	OBEX_ObjectAddHeader(handle, object, OBEX_HDR_LENGTH, hv, sizeof(guint32), 0);
	hv.bs = NULL;
	OBEX_ObjectAddHeader(handle, object, OBEX_HDR_BODY, hv, 0, OBEX_FL_STREAM_START);
}


void get_server(obex_t *handle, obex_object_t *object)
{
	obex_headerdata_t hv;
	guint8 hi;
	gint hlen;
	gchar *obj_type = NULL;

	gchar *name = NULL;
	gchar *namebuf = NULL;
	gint  con_id = -1;

	BTDEBUG("");

	while(OBEX_ObjectGetNextHeader(handle, object, &hi, &hv, &hlen)) {
		switch(hi) {
		case OBEX_HDR_NAME:
			
			if( (namebuf = g_malloc(hlen / 2))) {
				OBEX_UnicodeToChar(namebuf, hv.bs, hlen);
				name = namebuf;
			}
			BTDEBUG(" Found name:  %s", name);
			break;

		case OBEX_HDR_TYPE:
			obj_type = (gchar *)hv.bs;
			BTDEBUG(" Found type:%-30s", obj_type);
			break;
		case OBEX_HDR_CONNECTION:
			con_id = hv.bq4;
			BTDEBUG(" Found connection id: %08x", con_id);
			break;
		
		default:
			BTDEBUG(" Skipped header %02x", hi);
		}
	}
/*
	if (con_id != -1) {
		hv.bq4 = con_id;
		OBEX_ObjectAddHeader(handle, object, OBEX_HDR_CONNECTION, hv, 4, 0);
	} 
*/
	if (obj_type && strcasecmp(obj_type, "x-obex/folder-listing") == 0) {
		get_server_folderlisting(handle, object, name ? name : "");
	} else {
		get_server_file(handle, object, name ? name : "");	
	}

	if (namebuf)
		g_free(namebuf);

	return;
}


void setpath_server(obex_t *handle, obex_object_t *object)
{
	obex_headerdata_t hv;
	guint8 hi;
	gint hlen;

	gint	is_name = -1;
	gchar	*name = NULL;
	gchar	*namebuf = NULL;
	gint	con_id = -1;
	
	guint8	*nonhdrdata;
	int	flags;
	int	len;

	BTDEBUG("");
	
	if((len = OBEX_ObjectGetNonHdrData(object, &nonhdrdata)) == 2) {
		obex_setpath_hdr_t	*hdr = (obex_setpath_hdr_t*)nonhdrdata;
		flags = hdr->flags;
		BTDEBUG("Flags= 0x%02x  Constants=%d", flags, hdr->constants);
		
	} else {
		BTERROR("Invalid packet content. len=%d", len);
		flags = 0;
	}

	while(OBEX_ObjectGetNextHeader(handle, object, &hi, &hv, &hlen)) {
		switch(hi) {
		case OBEX_HDR_NAME:
			if (hlen) {
				if( (namebuf = g_malloc(hlen / 2)))	{
					OBEX_UnicodeToChar(namebuf, hv.bs, hlen);
					name = namebuf;
				}
				is_name = 1;
			} else
				is_name = 0;
			BTDEBUG("Found name:  %s", name);
			break;

		case OBEX_HDR_CONNECTION:
			con_id = hv.bq4;
			BTDEBUG("Found connection id: %08x", con_id);
			break;
		
		default:
			BTDEBUG("Skipped header %02x", hi);
		}
	}
/*
	if (con_id != -1) {
		hv.bq4 = con_id;
		OBEX_ObjectAddHeader(handle, object, OBEX_HDR_CONNECTION, hv, 4, 0);
	}
*/	
	if (is_name == -1) {
		if (flags == 0x03 /*& 1*/) {	// backup a level before applying the name
			if (change_dir("..", 0) != 0)
				goto error;
		}
	} else {
		if (is_name == 0) {
			if (change_dir(obex_root_path, 0) != 0)
				goto error;
		} else {
			if (change_dir(name, (flags&0x02)?0:CHDIR_CREATE_IF_NOT_EXIST) != 0)
				goto error;
		}
	}	
	OBEX_ObjectSetRsp(object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
	if (namebuf)
		g_free(namebuf);
	return;
error:
	OBEX_ObjectSetRsp(object, OBEX_RSP_NOT_FOUND, OBEX_RSP_NOT_FOUND);
	if (namebuf)
		g_free(namebuf);
	return;
}


void connect_server(obex_t *handle, obex_object_t *object)
{
	obex_headerdata_t hv;
	guint8 hi;
	gint hlen;
	guint8 *nonhdrdata;

	const guint8 *who = NULL;
	gint who_len = 0;
	const guint8 *target = NULL;
	gint target_len = 0;
	BTDEBUG("");

	if(OBEX_ObjectGetNonHdrData(object, &nonhdrdata) == 4) {
		obex_connect_hdr_t	*hdr = (obex_connect_hdr_t*)nonhdrdata;
		BTDEBUG("Version: 0x%02x. Flags: 0x%02x  OBEX packet length:%d",
			hdr->version, hdr->flags, ntohs(hdr->mtu));
	} else {
		BTERROR("Invalid packet content.");
	}
	while(OBEX_ObjectGetNextHeader(handle, object, &hi, &hv, &hlen)) {
		switch (hi)
		{
		case OBEX_HDR_WHO:
			who = hv.bs;
			who_len = hlen;
			BTDEBUG("got WHO: %*s", who_len, who);
			break;
		case OBEX_HDR_TARGET:
			target = hv.bs;
			target_len = hlen;
			if (hlen == 16)
				BTDEBUG("got TARGET. UUID:"
							"%08X-%04X-%04X-%04X-%08X%04X",
					*(guint32 *)&target[0], *(guint16 *)&target[4],
					*(guint16 *)&target[6], *(guint16 *)&target[8],
					*(guint32 *)&target[10], *(guint16 *)&target[14]);
			else
				BTDEBUG("got TARGET. unknown fmt");
			break;
		default:	
			BTDEBUG(" Skipped header %02x", hi);
		}
	}
	if (who_len == 6)	{
		if(strncmp("Linux", who, 6) == 0)	{
			BTDEBUG("Weeeha. I'm talking to the coolest OS ever!");
		}
	}
	who = target;
	who_len = target_len;

	OBEX_ObjectSetRsp(object, OBEX_RSP_SUCCESS, OBEX_RSP_SUCCESS);
	if (target) {
		hv.bq4 = 1;
		OBEX_ObjectAddHeader(handle, object, OBEX_HDR_CONNECTION, hv, 4, 0);
		hv.bs = who;
		OBEX_ObjectAddHeader(handle, object, OBEX_HDR_WHO, hv, who_len, 0);
	}
}


void server_request(obex_t *handle, obex_object_t *object, gint cmd)
{
	switch(cmd) {
	case OBEX_CMD_CONNECT:
		connect_server(handle, object);
		break;
	case OBEX_CMD_DISCONNECT:
		BTDEBUG("We got a disconnect-request");
		OBEX_ObjectSetRsp(object, OBEX_RSP_SUCCESS, OBEX_RSP_SUCCESS);
		break;
	case OBEX_CMD_GET:
		/* A Get always fits one package */
		get_server(handle, object);
		break;
	case OBEX_CMD_PUT:
		put_server(handle, object);
		break;
	case OBEX_CMD_SETPATH:
		setpath_server(handle, object);
		break;
	default:
		BTERROR(" Denied %02x request", cmd);
		OBEX_ObjectSetRsp(object, OBEX_RSP_NOT_IMPLEMENTED, OBEX_RSP_NOT_IMPLEMENTED);
		break;
	}
	return;
}

void server_reqhint(obex_t *handle, obex_object_t *object, gint cmd)
{
	switch(cmd) {
	case OBEX_CMD_PUT: {
		obexsrv_t	*srv = OBEX_GetUserData(handle);

		srv->name = strdup("obex_tmp_XXXXXX");
		srv->sfd = mkstemp(srv->name);
		if (srv->sfd < 0) {
			BTDEBUG("unable to create tmp file: %s\n", srv->name);
			return;
		}
		BTDEBUG("created tmp file: %s\n", srv->name);
		OBEX_ObjectReadStream(handle, object, NULL);	//XXX: Enable streamming
		srv->streamming = TRUE;
#if 1
		{
			obex_headerdata_t hv;
			guint8		hi;
			gint		hlen, con_id, len;
			char		name[80];

			while(OBEX_ObjectGetNextHeader(handle, object, &hi, &hv, &hlen)) {
				switch (hi) {
					case OBEX_HDR_CONNECTION:
						con_id = hv.bq4;
						BTDEBUG("Found connection id: %08x", con_id);
						break;
					case OBEX_HDR_LENGTH:
						len = hv.bq4;
						BTDEBUG("Found Length: %d, hlen: %d", len, hlen);
						break;
					case OBEX_HDR_NAME:
						OBEX_UnicodeToChar(name, hv.bs, hlen);
						BTDEBUG("Found name: %s", name);
						break;
				}
			}
		}
#endif
	}
	break;

	default:
	}

	OBEX_ObjectSetRsp(object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
}

void server_done(obex_t *handle, obex_object_t *object, gint obex_cmd, gint obex_rsp)
{
	obexsrv_t	*srv = OBEX_GetUserData(handle);

	BTDEBUG("Server request finished!");

	switch (obex_cmd) {
	case OBEX_CMD_DISCONNECT:
		BTDEBUG("Disconnect done!");
		break;

	default:
		BTDEBUG(" Command (%02x) has now finished", obex_cmd);
		srv->serverdone = TRUE;
		break;
	}
}


//
// Called by the obex-layer when some event occurs.
//
void obex_event(obex_t *handle, obex_object_t *object, gint mode, gint event, gint obex_cmd, gint obex_rsp)
{
	obexsrv_t	*srv = OBEX_GetUserData(handle);

	//BTDEBUG("event: %d, cmd: %x, rsp: %x\n", event, obex_cmd, obex_rsp);

	switch (event) {
		case OBEX_EV_PROGRESS:
			//BTDEBUG("Made some progress...");
			break;

		case OBEX_EV_ABORT:
			BTERROR("Request aborted!");
			break;

		case OBEX_EV_REQDONE:
			server_done(handle, object, obex_cmd, obex_rsp);
			break;
		
		case OBEX_EV_REQHINT:
			/* Accept any command. Not rellay good, but this is a test-program :) */
			server_reqhint(handle, object, obex_cmd);
			//OBEX_ObjectSetRsp(object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
			break;

		case OBEX_EV_REQ:
			server_request(handle, object, obex_cmd);
			break;

		case OBEX_EV_LINKERR:
			BTERROR("Link broken!");
			srv->state = SRVSTATE_CLOSED;
			srv->serverdone = TRUE;
			break;

		case OBEX_EV_PARSEERR:
			BTERROR("Parse error!");
			break;

		case OBEX_EV_STREAMEMPTY:
			fillstream(handle, object);
			break;

		case OBEX_EV_STREAMAVAIL:
			readstream(handle, object);
			break;

		default:
			BTERROR("Unknown event %02x!", event);
			break;
	}
}

void server_start(obexsrv_t *srv, char *service)
{
	int	err = 0;
	
	OBEX_ServerRegister(srv->handle, service);

	for (;;) {
		/* request processing loop */
		BTDEBUG("Processing request...\n");
		srv->serverdone = FALSE;
		while (!srv->serverdone) {
			if ((err = OBEX_HandleInput(srv->handle, 5)) < 0) {
				BTERROR("Error while doing OBEX_HandleInput()");
				break;
			}
		}
		if (srv->state == SRVSTATE_CLOSED)
			break;
		if (err < 0)
			break;
	}
}

void usage(void)
{
	fprintf(stderr, "obex_server -i [<host>]	use TCP socket on host\n"
			"obex_server <device>		use device as an RFCOMM port\n");
	BTERROR("Invalid options");
}

//
//
//
int main (int _argc, char *_argv[])
{
	gboolean	tcpobex = FALSE;
	obexsrv_t	srv;
	int		c, lind, err;

	struct option	opts[] = {
		{"help", 0, 0, 'h'},
		{"inet", 0, 0, 'i'},
		{"rootdir", 1, 0, 'd'},
		{0, 0, 0, 0}
	};

	argc = _argc;
	argv = _argv;
	home = getenv("HOME");
	user = getenv("USER");
	sprintf(confdir, "%s/affix", home);
	err = mkdir(confdir, 0755);
	if (err < 0 && errno != EEXIST) {
		perror("Unable to create dir\n");
		return 1;
	}
	uid = getuid();

	for (;;) {
		c = getopt_long(argc, argv, "+hd:i", opts, &lind);
		if (c == -1)
			break;
		switch (c) {
			case 'h':
				usage();
				return 0;
				break;

			case 'd':
				obex_root_path = strdup(optarg);
				break;

			case 'i':
				tcpobex = TRUE;
				break;

			case ':':
				printf("Missing parameters for option: %c\n", optopt);
				return 1;
				break;

			case '?':
				printf("Unknown option: %c\n", optopt);
				return 1;
				break;
		}
	}

	btlog_mask = BTDEBUG_MODULE_OBEX;
	btdaemon = 1;

	memset(&srv, 0, sizeof(obexsrv_t));

	if (!obex_root_path) {
		if (uid == 0) {
			/* root */
			obex_root_path = strdup(OBEX_ROOT_PATH);
		} else {
			/* set to home dir */
			char	buf[80];
			sprintf(buf, "%s/spool", confdir);
			obex_root_path = strdup(buf);
		}
	}
	
	BTDEBUG("started");
	if (!tcpobex) {
		if (!(srv.handle = btobex_init(&srv.link, obex_event, 0))) {
			BTERROR( "OBEX_Init failed:%s", strerror(errno));
			exit(0);
		}
	} else { 
		// tcpobex
		if (!(srv.handle = OBEX_Init(OBEX_TRANS_INET, obex_event, 0))) {
			BTERROR( "OBEX_Init failed:%s", strerror(errno));
			exit(0);
		}
	}

	OBEX_SetUserData(srv.handle, &srv);

	if (rmkdir(obex_root_path, 0700) != 0 || chdir(obex_root_path) != 0) {
		BTERROR("failed to set the root directory.");
		exit(0);
	}

	server_start(&srv, IR_SERVICE);

	if (!tcpobex)
		btobex_cleanup(&srv.link, 1);

	return 0;
}
