/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author:Guruprasad Krishnamurthy <guruprasad.krishnamurthy@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   */

/*
   $Id: cstate.c,v 1.3 2002/08/29 12:01:15 kds Exp $

Fixes:     Guruprasad Krishnamurthy <guruprasad.krishnamurthy@nokia.com>
*/

#include <affix/config.h>

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <glib.h>

#include "cstate.h"
#include "utils.h"

static GHashTable *cacheTable;

void printContinuationState(ContinuationState *pCState)
{
	BTDEBUG("[%s-%d] : ", __FUNCTION__, __LINE__);
	BTDEBUG("CState TS : 0x%lx", pCState->timestamp);
	BTDEBUG("CState max bytes : %d", pCState->cStateValue.maxBytesSent);
	BTDEBUG("CState last index : %d", pCState->cStateValue.lastIndexSent);
}

static gint orderingFunction(gconstpointer cstate1, gconstpointer cstate2)
{
	gint order = -1;

	const ContinuationStateBuffer *pcstate1 = (const ContinuationStateBuffer *)cstate1;
	const ContinuationStateBuffer *pcstate2 = (const ContinuationStateBuffer *)cstate2;

	if ((pcstate1 != NULL) && (pcstate2 != NULL)) 
		order = pcstate1->timestamp - pcstate2->timestamp;

	return order;
}


void cacheListCleanup(gpointer key, gpointer value, gpointer data)
{
	GSList *cacheList;

	cacheList = (GSList *)value;
	if (cacheList != NULL) {
		int i = 0;
		ContinuationStateBuffer *pBuffer;
		for (;;) {
			gpointer data = g_slist_nth_data(cacheList, i);
			if (data == NULL)
				break;
			pBuffer = (ContinuationStateBuffer *)data;
			free(pBuffer->responseBuffer.data);
			free(pBuffer);
			i++;
		}
		g_slist_free(cacheList);
	}
	
	if(key)
		free(key);
}


void cacheListPrint(gpointer key, gpointer value, gpointer data)
{
	BTDEBUG("Key : %s", (char *)key);
	BTDEBUG("Value : 0x%x", (char *)value);
}


void initContinuationStateCache(void)
{
	cacheTable = g_hash_table_new(g_int_hash, g_int_equal);
}

ContinuationState *getContinuationState(char *buffer)
{
	char *pdata = buffer;
	uint8_t cStateSize = 0;
	ContinuationState *pCState = NULL;

	/*
	 ** Check if continuation state exists, if yes attempt
	 ** to get response remainder from cache, else send error
	 */
	cStateSize = __get_u8(pdata);
	BTDEBUG("Continuation State size : %d", cStateSize);
	pdata += sizeof(uint8_t);
	if (cStateSize != 0) {
		pCState = (ContinuationState *)pdata;
		BTDEBUG("Cstate TS : 0x%lx", pCState->timestamp);
		BTDEBUG("Bytes sent : %d", pCState->cStateValue.maxBytesSent);
	}                 

	return pCState;
}

int setContinuationState(char *pdata, ContinuationState *pCState)
{
	int length = 0;

	if (pCState == NULL) {
		/*
		 ** Set "null" continuationState
		 */
		__put_u8(pdata, 0);
		pdata += sizeof(uint8_t);
		length += sizeof(uint8_t);
	}
	else {
		BTDEBUG("Non null CState id : 0x%lx", pCState->timestamp);
		__put_u8(pdata, sizeof(ContinuationState));
		pdata += sizeof(uint8_t);
		length += sizeof(uint8_t);
		memcpy(pdata, pCState, sizeof(ContinuationState));
		length += sizeof(ContinuationState);
	}

	return length;
}

int setContinuationStatePDU(PDUForm *pduForm, ContinuationState *pCState)
{
	int length = 0;
	char *pdata = NULL;

	pdata = pduForm->data + pduForm->dataSize;
	length = setContinuationState(pdata, pCState);
	pduForm->dataSize += length;
	
	return length;
}

long createContinuationStateBuffer(int fd, PDUForm *response)
{
	ContinuationStateBuffer *pCStateBuffer;
	char *pdata;
	GSList *cacheList = NULL;
	gboolean cacheListExists = FALSE;

	pdata = (char *)malloc(response->dataSize);
	memcpy(pdata, response->data, response->dataSize);
	pCStateBuffer = (ContinuationStateBuffer *)malloc(sizeof(ContinuationStateBuffer));
	memset((char *)pCStateBuffer, 0, sizeof(ContinuationStateBuffer));
	pCStateBuffer->responseBuffer.data = pdata;
	pCStateBuffer->responseBuffer.dataSize = response->dataSize;
	pCStateBuffer->responseBuffer.bufferSize = response->dataSize;
	pCStateBuffer->timestamp = getTimeInSeconds();

	cacheList = (GSList *)g_hash_table_lookup(cacheTable, &fd);
	if (cacheList != NULL)
		cacheListExists = TRUE;
	BTDEBUG("CacheList : 0x%lx", (long)cacheList);
	cacheList = g_slist_insert_sorted(cacheList, pCStateBuffer, orderingFunction);
	if (cacheList != NULL) {
		int	*ptr;
		if (cacheListExists == TRUE) {
			ptr = &fd;
		} else {
			// allocate mem for the key
			ptr = (int*)malloc(sizeof(int));
			*ptr = fd;
		}
		g_hash_table_insert(cacheTable, ptr, cacheList);
	}

#ifdef CONFIG_AFFIX_DEBUG 
	g_hash_table_foreach(cacheTable, cacheListPrint, NULL);
#endif

	return pCStateBuffer->timestamp;
}

GSList *locateCacheList(int fd, ContinuationState *pCState, GSList **listStart)
{
	GSList *cacheList = NULL;
	GSList *tmpList = NULL;

	cacheList = (GSList *)g_hash_table_lookup(cacheTable, &fd);
	if (cacheList != NULL) {
		ContinuationStateBuffer tmpRec;

		*listStart = cacheList;
		tmpRec.timestamp = pCState->timestamp;
		tmpList = g_slist_find_custom(cacheList, &tmpRec, orderingFunction);
	}

	return tmpList;
}


PDUForm *getCachedResponse(int fd, ContinuationState *pCState)
{
	PDUForm *pduForm = NULL;
	GSList *tmpList = NULL;
	GSList *listStart = NULL;

#ifdef CONFIG_AFFIX_DEBUG
	g_hash_table_foreach(cacheTable, cacheListPrint, NULL);
#endif

	tmpList = locateCacheList(fd, pCState, &listStart);
	if (tmpList != NULL) {
		ContinuationStateBuffer *realRec = NULL;

		//BTINFO("Found cached response ...\n");
		realRec = (ContinuationStateBuffer *)tmpList->data;
		if (realRec != NULL)
			pduForm = &realRec->responseBuffer;
	}

	BTDEBUG("Cached response : 0x%lx", (long)pduForm);

	return pduForm;
}

int removeCachedResponse(int fd, ContinuationState *pCState)
{
	int status = -1;
	GSList *cacheList = NULL;
	GSList *tmpList = NULL;
	GSList *listStart = NULL;

	//BTINFO("Removing response...\n");
	tmpList = locateCacheList(fd, pCState, &listStart);
	if (tmpList != NULL) {
		ContinuationStateBuffer *toRemove;

		//BTINFO("found to remove\n");
		toRemove = (ContinuationStateBuffer *)tmpList->data;
		if (toRemove != NULL) {
			cacheList = g_slist_remove(listStart, toRemove);
			g_hash_table_insert(cacheTable, &fd, cacheList);
			status = 0;
		}
	}

	return status;
}

int cleanupCachedResponses(int fd)
{
	GSList *cacheList = NULL;

	cacheList = (GSList *)g_hash_table_lookup(cacheTable, &fd);
	//BTINFO("cleaning up...");
	if (cacheList != NULL) {
		//BTINFO("yes...");
		cacheListCleanup(NULL, cacheList, NULL);
		g_hash_table_remove(cacheTable, &fd);
	}

	return 0;
}


void cleanupContinuationStateBuffer(void)
{
	BTDEBUG("Cache table cleanup");
	if (cacheTable != NULL) {
		g_hash_table_foreach(cacheTable, cacheListCleanup, NULL);
		g_hash_table_destroy(cacheTable);
	}
}

int copyContinuationState(char *pdata, CState *pCState)
{
	int length;

	if (pCState == NULL) {
		__put_u8(pdata, 0);
		length = 1;
	} else {
		__put_u8(pdata++, pCState->length);
		memcpy(pdata, pCState->data, pCState->length);
		length = pCState->length+1;
	}
	return length;
}

