/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author:Guruprasad Krishnamurthy <guruprasad.krishnamurthy@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   */

/*
   $Id: match.c,v 1.1.1.1 2002/07/11 11:21:57 kds Exp $

   Matching function based on UUID(s) of service or it's
   protocol identifier.

	Fixes:     Guruprasad Krishnamurthy <guruprasad.krishnamurthy@nokia.com>

		Dmitry Kasatkin		: matchAnyUUID
*/

#include <affix/config.h>

#include <stdio.h>
#include <glib.h>
#include <stdint.h>
#include <stdlib.h>

#include <affix/sdp.h>
#include <affix/sdpclt.h>

/*
 ** The matching process is defined as "each and every UUID
 ** specified in the "search pattern" must be present in the
 ** "target pattern". Here "search pattern" is the set of UUIDs
 ** specified by the service discovery client and "target pattern"
 ** is the set of UUIDs present in a service record. 
 ** 
 ** Return 1 if each and every UUID in the search
 ** pattern exists in the target pattern
 */
gboolean matchUUID(GSList *searchPattern, GSList *targetPattern)
{
	/*
	 ** The target is a sorted list, so we need not look
	 ** at all elements to confirm existence of an element
	 ** from the search pattern
	 */
	int sIndex = 0;
	gboolean exists = FALSE;
	GSList *pList;

	if ((searchPattern == NULL) || (targetPattern == NULL))
		BTERROR("Fails to get patterns right");

	if (((g_slist_length(searchPattern) > 0) && 
				(g_slist_length(targetPattern) > 0)) &&
			(g_slist_length(targetPattern) >= g_slist_length(searchPattern))) {
		for (;;) {
			UUID *sData128Bit;
			gpointer sData = g_slist_nth_data(searchPattern, sIndex);
			if (sData == NULL)
				break;
			exists = FALSE;
			/*
			 ** Create a 128 bit form of the search UUID
			 */
			sData128Bit= convertTo128Bits((UUID *)sData);
			pList = g_slist_find_custom(targetPattern, sData128Bit, compareUUID);
			free(sData128Bit);
			if (pList != NULL) {
				sIndex++; exists = TRUE;
			} else
				break;
		}
	}

	return exists;
}

/*
 ** The matching process is defined as "any UUID"
 ** specified in the "search pattern" can be present in the
 ** "target pattern".
 */
gboolean matchAnyUUID(GSList *searchPattern, GSList *targetPattern)
{
	/*
	 ** The target is a sorted list, so we need not look
	 ** at all elements to confirm existence of an element
	 ** from the search pattern
	 */
	int sIndex = 0;
	gboolean exists = FALSE;
	GSList *pList;

	if ((searchPattern == NULL) || (targetPattern == NULL))
		BTERROR("Fails to get patterns right\n");

	if (((g_slist_length(searchPattern) > 0) && 
				(g_slist_length(targetPattern) > 0)) &&
			(g_slist_length(targetPattern) >= g_slist_length(searchPattern))) {
		for (;;) {
			UUID *sData128Bit;
			gpointer sData = g_slist_nth_data(searchPattern, sIndex);
			if (sData == NULL)
				break;
			exists = FALSE;
			/*
			 ** Create a 128 bit form of the search UUID
			 */
			sData128Bit= convertTo128Bits((UUID *)sData);
			pList = g_slist_find_custom(targetPattern, sData128Bit, compareUUID);
			free(sData128Bit);
			if (pList) {
				exists = TRUE;
				break;
			}
			sIndex++;
		}
	}

	return exists;
}
