/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author:Guruprasad Krishnamurthy <guruprasad.krishnamurthy@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   */

/*
   $Id: pdu.c,v 1.1.1.1 2002/07/11 11:21:57 kds Exp $

   Contains method to generate the PDU form for a given
   service record. All attributes(by default) are included
   or only a subset (if specified)

Fixes:     Guruprasad Krishnamurthy <guruprasad.krishnamurthy@nokia.com>
*/

#include <affix/config.h>

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <affix/sdp.h>
#include "pdu.h"
#include "utils.h"
#include "match.h"


/*
 ** Sets the length of a data element sequence
 ** The size is either an 8 bit quantity or a 16 bit
 ** quantity. This method is called after the actual 
 ** data is appended to the data element sequence and
 ** PDU length is computed
 */
void setSequenceLengthPDU(PDUForm *pduForm)
{
	/*
	 ** Set the sequence size
	 */
	char *pdata;
	DataTypeDescriptor dataType;

	pdata = pduForm->data;
	dataType = __get_u8(pdata);
	pdata += sizeof(DataTypeDescriptor);

	switch (dataType) {
		case SequenceSize8 :
			__put_u8(pdata, pduForm->dataSize - 
				sizeof(DataTypeDescriptor) - sizeof(DataSizeDescriptor8));
			break;

		case SequenceSize16 :
			__put_u16(pdata, htons(pduForm->dataSize - 
					sizeof(DataTypeDescriptor) - sizeof(DataSizeDescriptor16)));
			break;

		case SequenceSize32 :
			__put_u32(pdata, htonl(pduForm->dataSize - 
					sizeof(DataTypeDescriptor) - sizeof(DataSizeDescriptor32)));
			break;
		default :
			break;
	}
}


/*
 ** This function appends data to the PDU form in the buffer "dst"
 ** from source "src". The data length is also computed and set.
 ** Should the PDU form length exceed 2^8, then sequence type is
 ** set accordingly and the data is memmove()'d.
 */
int appendToPduForm(PDUForm *dst, PDUForm *src)
{
	DataTypeDescriptor dataType;
	char *pdata;

	pdata = dst->data;
	dataType = __get_u8(pdata);

	BTDEBUG("Append src size : %d", src->dataSize);
	BTDEBUG("Append dst size : %d", dst->dataSize);
	BTDEBUG("Dst buffer size : %d", dst->bufferSize);

	if ((dst != NULL) && (dst->data != NULL)) {
		if ((dst->dataSize + src->dataSize) > dst->bufferSize) {
			/*
			 ** Need to realloc !
			 */
			int neededSize = SDP_PDU_CHUNK_SIZE * ((src->dataSize/SDP_PDU_CHUNK_SIZE) + 1);
			dst->data = (char *)realloc(dst->data, (dst->bufferSize + neededSize));
			BTDEBUG("Realloc'ing : %d", neededSize);
			if (dst->data == NULL) {
				BTERROR("Realloc fails \n");
			}
			dst->bufferSize += neededSize;
		}
		if ((dst->dataSize == 0) && (dataType == 0)) {
			//create initial sequence
			__put_u8(pdata, SequenceSize8);
			pdata += sizeof(DataTypeDescriptor);
			dst->dataSize += sizeof(DataTypeDescriptor);
			/*
			 ** Space for sequence size
			 */
			pdata += sizeof(DataSizeDescriptor8);
			dst->dataSize += sizeof(DataSizeDescriptor8);
		}

		memcpy((dst->data + dst->dataSize), src->data, src->dataSize);
		dst->dataSize += src->dataSize;

		dataType = *(DataTypeDescriptor *)dst->data;
		if ((dst->dataSize > UCHAR_MAX) && (dataType == SequenceSize8)) {
			/*
			 ** Need to memmove
			 */
			short offset = sizeof(DataTypeDescriptor) + sizeof(DataSizeDescriptor8);
			memmove((dst->data + offset + 1), (dst->data + offset),
					(dst->dataSize - offset));
			pdata = dst->data;
			__put_u8(pdata, SequenceSize16);
			pdata += sizeof(DataTypeDescriptor);
			dst->dataSize += 1;
		}
	}

	//setSequenceLengthPDU(dst);
	pdata = dst->data; 
	dataType = __get_u8(pdata);
	pdata += sizeof(DataTypeDescriptor);
	switch (dataType) {
		case SequenceSize8 :
			__put_u8(pdata, dst->dataSize - 
				sizeof(DataTypeDescriptor) - sizeof(DataSizeDescriptor8));
			break;

		case SequenceSize16 :
			__put_u16(pdata, htons(dst->dataSize - 
					sizeof(DataTypeDescriptor) - sizeof(DataSizeDescriptor16)));
			break;

		case SequenceSize32 :
			__put_u32(pdata, htonl(dst->dataSize - 
					sizeof(DataTypeDescriptor) - sizeof(DataSizeDescriptor32)));
			break;
		default :
			break;
	}

	return src->dataSize;
}


/*
 ** This function generates the PDU form of a service record for
 ** a specified set of attributes and not the entire set. This is
 ** driven by a service discovery client which asks for attributes
 ** of a service record by speicifying a subset of attribute 
 ** identifiers
 */
void generateServiceRecordPduFormById(
		ServiceRecord *svcRec,
		ServiceAttributeIdentifier attrId,
		PDUForm *attrSubsetPduForm
		)
{
	SDPData *pData = NULL;

	pData = (SDPData *)getSDPData(svcRec, attrId);

	/*
	 ** If a subset pdu form of the service record is created
	 ** it is always in a local buffer "attrSubsetPduForm" defined
	 ** static to this file only
	 */
	if ((pData != NULL) && (pData->pduForm.data != NULL) && 
			(pData->pduForm.dataSize != 0))
		appendToPduForm(attrSubsetPduForm, &pData->pduForm);
}

