/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author:Guruprasad Krishnamurthy <guruprasad.krishnamurthy@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   */

/*
   $Id: sdp_utils.c,v 1.4 2002/08/01 17:52:08 kds Exp $

   SDP client utilities to perform GIAC, connect to an SDP server,
   and to initialize the SDP machinery

Fixes:     	Guruprasad Krishnamurthy <guruprasad.krishnamurthy@nokia.com>

	    Dmitry Kasatkin		: SDPServerConnection changed to complex type
*/

#include <affix/config.h>

#include <stdio.h>

#include <sys/types.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include <affix/sdp.h>
#include <affix/sdpclt.h>
#include <affix/sdpsrv.h>
#include "utils.h"

static int	running = 0;
static int	sdp_flags = 0;
static pid_t	sdp_pid = 0;


/*
 ** SDP Initialization procedure. Sets up SDP internal
 ** data structures. Must be called before any other
 ** SDP function can be invoked.
 **
 ** INPUT :
 **
 **   NONE
 **
 ** OUTPUT :
 ** 
 **   E_OK : Initialization success
 **
 **   E_FAILURE : Initialization failure
 **
 */
int SDPInit(int flags)
{
	int	err = 0;
	
	if (flags&SDP_SVC_PROVIDER) {
		pid_t	pid = affix_pidof("btsdp", PIDOF_SINGLE, 0);
		if (pid == 0) {
			/* start server */
			pid = fork();
			if (pid < 0) {
				BTERROR("Unable to fork process\n");
				return -1;
			}
			if (!pid) {
				/* child */
				err = execlp("btsdp", "btsdp", "-D", NULL);
				if (err)
					BTERROR("Unable to run btsdp\n");
			} else {
				struct sockaddr_in	in;
				int			fd;
				socklen_t		len;
				time_t			stm, tm;

				in.sin_family = AF_INET;
				in.sin_port = SDP_TCP_PORT;
				in.sin_addr.s_addr = inet_addr("127.0.0.1");
				len = sizeof(in);

				fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
				if (fd<0)
					return -1;
				stm = time(NULL);
				for (;;) {
					err = connect(fd, (struct sockaddr *)&in, len);
					if (!err)
						break;
					tm = time(NULL);
					if (difftime(tm, stm) > 2) {
						close(fd);
						return -1;
					}	
				}
				close(fd);
			}
		}
		sdp_pid = pid;
	}
	if (!running) {
		err = sdp_init(client);
		if (!err) {
			running = 1;
			sdp_flags = flags;
		}
	}
	return err;
}

void SDPCleanup(void)
{
	if (sdp_flags & SDP_SVC_PROVIDER) {
		/* btsdp exit if no services available */
		kill(sdp_pid, SIGUSR2);
	}
	
	if (running) {
		sdp_shutdown();
		running = 0;
	}
}
	
/*
 ** Create a connection (L2CAP) to the specified
 ** Bluetooth device (BD_ADDR). 
 **
 ** Note : The identification of the device on which
 **        the SDP server is known to run is outside 
 **        the scope of this function. For a utility
 **        function to do this, see sdp_utils.h
 **
 ** INPUT :
 **  
 **  char * bdAddr :
 **    Address of the device where the SDP server is known
 **    to run. In the format xx:yy:zz:ll:mm
 **
 **  SDPServerHandle :
 **    Contains a handle to the SDP server on the remote
 **    device. All subsequent transactions must be carried
 **    out using this
 **
 */
SDPServerHandle SDPOpenServerConnection(BD_ADDR bdAddr)
{
	int	status = 0;
	int	clientSockFd;
	char	bdAddrStr[32];

	socklen_t	sockLength;

	struct sockaddr_affix	addr;
	struct sockaddr_in	inAddr;

	bda2str(bdAddrStr, &bdAddr);
	BTDEBUG("BT addr str : %s\n", bdAddrStr);
	/*
	 ** Is this to our own (local device) SDP server ?
	 ** If yes, short circuit to a TCP connection
	 */
	if (bda_equal(&bdAddr, &localBTDeviceAddr)) {
		/*
		 ** Code for a local SDP server connection
		 */
		BTDEBUG("Attempt to open a TCP socket, local client\n");
		clientSockFd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
		if (clientSockFd < 0)
			return clientSockFd;
		sockLength = sizeof(inAddr);
		inAddr.sin_family = AF_INET;
		inAddr.sin_port = SDP_TCP_PORT;
		inAddr.sin_addr.s_addr = inet_addr("127.0.0.1");
		BTDEBUG("TCP fd : %d\n", clientSockFd);
		BTDEBUG("Local addr : %d\n", inAddr.sin_addr.s_addr);
		status = connect(clientSockFd, (struct sockaddr *)&inAddr, sockLength);
		if (status == -1) {
			BTERROR("TCP connect error:%s", strerror(errno));
			close(clientSockFd);
		}
	} else {                                             
		/*
		 ** Create a BT socket for ourselves
		 */
		BTDEBUG("Attempt to open BT connection\n");
		clientSockFd = socket(PF_AFFIX, SOCK_STREAM, BTPROTO_L2CAP);
		if (clientSockFd < 0)
			return clientSockFd;
		sockLength = sizeof(addr);
		addr.family = AF_AFFIX;
		addr.port = SDP_PSM;
		addr.bda = bdAddr;
		addr.local = BDADDR_ANY;
		BTDEBUG("Connecting\n");
		status = connect(clientSockFd, (struct sockaddr*)&addr, sockLength);          
		if (status == -1 ) {
			close(clientSockFd);
			BTDEBUG("connect error:%s", strerror(errno));
		} else {
			BTDEBUG("Connect success\n");
		}
	}
	if (status == 0)
		return clientSockFd;
	else {
		switch (errno) {
			case ETIMEDOUT:
				return E_CONNTIMEDOUT;
			case ECONNREFUSED:
				return E_CONNREFUSED;

			default:
				return status;
		}
	}
}                                 


void SDPCloseServerConnection(SDPServerHandle srvHandle)
{
	BTDEBUG("[%s-%d] : \n", __FUNCTION__, __LINE__);
	closeConnection(srvHandle);
}


/*
 ** Connect to local SDP server on this device
 **
 ** INPUT :
 **
 **   NONE
 **
 ** OUTPUT :
 **
 **   SDPServerHandle** points to a list
 **   of SDP servers found in the piconet
 **
 */
SDPServerHandle SDPOpenLocalServerConnection(void)
{
	SDPServerHandle srvHandle = -1;
	BD_ADDR		bdAddr;

	getLocalBDAddr(&bdAddr);
	srvHandle = SDPOpenServerConnection(bdAddr);

	return srvHandle;
}


BD_ADDR SDPGetLocalBDAddr(void)
{
	return localBTDeviceAddr;
}


/*
 ** Find devices in the piconet.
 ** First arg is a pointer to locations where
 ** discovered devices can be added. "numberOfDevices"
 ** is the max count that pItems can hold
 **
 */
int SDPGeneralInquiry(INQUIRY_ITEM *pItems, int numberOfDevices, 
		int duration, uint8_t *devicesFound)
{
	int	fd, i;
	__u32	bdLap;
	int	status;
	
	INQUIRY_ITEM localItems[32];

	fd = hci_open("bt0");
	if (fd == -1) {
		BTERROR("Cannot open for general inquiry:%s", strerror(errno));
		status = -1;
	}
	else {
		bdLap = GIAC;
		status = HCI_Inquiry(fd, duration, 
				32, localItems, devicesFound);
		if (status == 0) {
			INQUIRY_ITEM *ptr;

			ptr = pItems;
			for (i = 0;
					i < ((numberOfDevices > *devicesFound) ?
						*devicesFound : numberOfDevices); i++) {
				*ptr = localItems[i];
				ptr++;
			}
		}
	}

	close(fd);

	return status;
}

