/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author:Guruprasad Krishnamurthy <guruprasad.krishnamurthy@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/*
   $Id: sdpclt.c,v 1.3 2002/07/19 20:51:14 kds Exp $

   Contains the implementation of the SDP service discovery API
 
   Fixes:	Guruprasad Krishnamurthy

   		Dmitry Kasatkin		- Continuation Mechanism
		Dmitry Kasatkin		- Port Request

*/
	 
#include <affix/config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <affix/sdp.h>
#include <affix/sdpclt.h>
#include "servicedb.h"
#include "utils.h"
#include "cstate.h"

/*
** This is a service search request. 
**
** INPUT :
**
**   GSList *svcSearchList
**     Singly linked list containing elements of the search
**     pattern. Each entry in the list is a UUID (DataTypeUUID_16)
**     of the service to be searched
**
**   uint16_t maxSvcRecordCount
**      A 16 bit integer which tells the service, the maximum
**      entries that the client can handle in the response. The
**      server is obliged not to return > maxSvcRecordCount entries
**
** OUTPUT :
**
**   int return value
**     E_OK 
**       The request completed successfully. This does not
**       mean the requested services were found
**     E_FAILURE
**       On any failure
**     E_TIMEOUT 
**       The request completed unsuccessfully due to a timeout
**
**   GSList **svcHandleList
**     This variable is set on a successful return if there are
**     non-zero service handles. It is a singly linked list of
**     service record handles (ServiceRecordHandle)
**
**   uint16_t *handleCount
**     This is a pointer to a 16 bit integer, which is set to 
**     indicate the number of service record handles present in
**     svcHandleList
*/
int __SDPServiceSearchRequest(
		SDPServerHandle srvHandle,
		GSList *svcSearchList, 
		uint16_t maxSvcRecordCount,
		GSList **svcResponseList,
		uint16_t *handleCountInResponse)
{
	int status = 0;
	int requestSize = 0, _requestSize;
	int responseSize = 0;
	int seqLength = 0;
	char *pdata, *_pdata;
	char *requestBuffer =  NULL;
	char *responseBuffer =  NULL;
	SdpPduHeader *pduRequestHeader;
	SdpPduHeader *pduResponseHeader;
	CState *pCState = NULL;
	//PDUForm concatenatedResponseBuffer;
	int responseLength = 0;

	requestBuffer = (char *)malloc(RequestBufferSize);
	if (!requestBuffer)
		return E_FAILURE;
	responseBuffer = (char *)malloc(ResponseBufferSize);
	if (!responseBuffer) {
		free(requestBuffer);
		return E_FAILURE;
	}

	pduRequestHeader = (SdpPduHeader *)requestBuffer;
	pduRequestHeader->pduId = SDP_ServiceSearchRequest;
	pdata = (char *)(requestBuffer + sizeof(SdpPduHeader));
	requestSize = sizeof(SdpPduHeader);

	/*
	 ** Add service class IDs for search
	 */
	seqLength = generateSearchSequencePduForm(pdata, svcSearchList);

	BTDEBUG("Data seq added : %d\n", seqLength);

	/*
	 ** Now set the length and increment the pointer
	 */
	requestSize += seqLength;
	pdata += seqLength;

	/*
	 ** Specify the maximum svc rec count that
	 ** client expects
	 */
	__put_u16(pdata, htons(maxSvcRecordCount));
	requestSize += sizeof(uint16_t);
	pdata += sizeof(uint16_t);

	_requestSize = requestSize;
	_pdata = pdata;
	*svcResponseList = NULL;	//important

	do {
		int bytesScanned = 0;
		int totalServiceRecordCount, currentServiceRecordCount;

		/*
		 ** Add continuation state or NULL (first time)
		 */
		requestSize = _requestSize + copyContinuationState(_pdata, pCState);

		/*
		 ** Set the request header's param length
		 */
		pduRequestHeader->paramLength = htons(requestSize - sizeof(SdpPduHeader));
		pduRequestHeader->transactionId = htons(generateTransactionId());
		/*
		 ** Send the request, wait for response and if
		 ** not error, set the appropriate values
		 ** and return
		 */
		status = sendRequestWaitForResponse(srvHandle, requestBuffer, 
				responseBuffer, requestSize, &responseSize);

		if (status != E_OK)
			break;

		pduResponseHeader = (SdpPduHeader *)responseBuffer;
		if (pduResponseHeader->pduId == SDP_ErrorResponse) {
			status = E_FAILURE;
			break;
		}

		responseLength = ntohs(pduResponseHeader->paramLength);
		pdata = responseBuffer + sizeof(SdpPduHeader);
		/*
		 ** Net service record match count
		 */
		totalServiceRecordCount = 0;
		currentServiceRecordCount = 0;
		totalServiceRecordCount = ntohs(__get_u16(pdata));
		pdata += sizeof(uint16_t);
		bytesScanned += sizeof(uint16_t);
		currentServiceRecordCount = ntohs(__get_u16(pdata));
		pdata += sizeof(uint16_t);
		bytesScanned += sizeof(uint16_t);

		BTDEBUG("Total svc count : %d\n", totalServiceRecordCount);
		BTDEBUG("Current svc count : %d\n", currentServiceRecordCount);
		BTDEBUG("ResponseLength : %d\n", responseLength);
		BTDEBUG("Total svc count : %d\n", totalServiceRecordCount);
		BTDEBUG("Current svc count : %d\n", currentServiceRecordCount);
		BTDEBUG("ResponseLength : %d\n", responseLength);

		if (currentServiceRecordCount == 0) {
			status = E_OK;
			break;
		}
		status = __extractServiceRecordHandles(pdata, svcResponseList, 
				currentServiceRecordCount, &bytesScanned);
		if (status)
			break;
		
		*handleCountInResponse = currentServiceRecordCount;
		/*
		 ** Check if we have ContinuationState set, if yes,
		 ** set the pCState pointer
		 */
		BTDEBUG("BytesScanned : %d\n", bytesScanned);
		if (responseLength > bytesScanned) {
			uint8_t cStateLength = 0;

			pdata = responseBuffer + sizeof(SdpPduHeader) + bytesScanned;
			cStateLength = __get_u8(pdata);
			if (cStateLength > 0) {
				pCState = (CState *)pdata;
				BTDEBUG("Continuation state present\n");
				BTDEBUG("CState length : %d\n", cStateLength);
			} else
				pCState = NULL;
		}
	} while (pCState);
	
	free(requestBuffer);
	free(responseBuffer);
	//if (concatenatedResponseBuffer.data != NULL)
	//free(concatenatedResponseBuffer.data);

	return status;
}

int SDPServiceSearchRequest(
		BD_ADDR bdAddr,
		GSList *svcSearchList, 
		uint16_t maxSvcRecordCount,
		GSList **svcResponseList,
		uint16_t *handleCountInResponse)
{
	int status = 0;
	SDPServerHandle srvHandle;

	srvHandle = SDPOpenServerConnection(bdAddr);
	if (srvHandle < 0)
		return srvHandle;

	status  = __SDPServiceSearchRequest(srvHandle, svcSearchList, 
			maxSvcRecordCount, svcResponseList, handleCountInResponse);
	SDPCloseServerConnection(srvHandle);

	return status;
}



/*
** This is a service attribute request. 
**
** INPUT :
**
**   ServiceRecordHandle svcHandle
**     The handle of the service for which the attribute(s) are
**     requested
**
**   AttributeRequestType attrReqType
**     Attribute identifiers are 16 bit unsigned integers specified
**     in one of 2 ways described below :
**     IndividualAttributes - 16bit individual identifiers
**        They are the actual attribute identifiers in ascending order
**
**     RangeOfAttributes - 32bit identifier range
**        The high-order 16bits is the start of range
**        the low-order 16bits are the end of range
**        0x0000 to 0xFFFF gets all attributes
**
**   GSList *attrIDList
**     Singly linked list containing attribute identifiers desired.
**     Every element is either a uint16_t(attrSpec = IndividualAttributes)  
**     or a uint32_t(attrSpec=RangeOfAttributes)
**
**   uint16_t maxAttrIDByteCount
**     The byte count of the number of attribute IDs specified
**     in the request list
**
** OUTPUT :
**   int return value
**     E_OK 
**       The request completed successfully. This does not
**       mean the requested services were found
**     E_TIMEOUT 
**       The request completed unsuccessfully due to a timeout
**
**   uint16_t *maxAttrResponseByteCount
**     This is a pointer to a 16 bit integer, which is set to 
**     indicate the number of bytes of attributes returned. 
**     This pointer is set on successful return
**
*/


int __SDPServiceAttributeRequest(
		SDPServerHandle srvHandle,
		ServiceRecordHandle svcHandle,
		AttributeRequestType attrReqType,
		GSList *attrIDList,
		uint16_t maxAttrIDByteCount,
		ServiceRecord	**_svcRec,
		uint16_t *maxAttrResponseByteCount)
{
	int status = 0;
	int requestSize = 0, _requestSize;
	int responseSize = 0;
	int attrListByteCount = 0;
	int seqLength = 0;
	char *pdata = NULL, *_pdata;
	char *requestBuffer =  NULL;
	char *responseBuffer =  NULL;
	SdpPduHeader *pduRequestHeader;
	SdpPduHeader *pduResponseHeader;
	// DataTypeDescriptor dataType = 0;
	CState *pCState = NULL;
	uint8_t cStateLength = 0;
	PDUForm concatenatedResponseBuffer;
	ServiceRecord *svcRec = NULL;

	*_svcRec = NULL;

	if ((attrReqType != IndividualAttributes) && (attrReqType != RangeOfAttributes))
		return E_FAILURE;

	requestBuffer = (char *)malloc(RequestBufferSize);
	if (!requestBuffer)
		return E_FAILURE;
	responseBuffer = (char *)malloc(ResponseBufferSize);
	if (!responseBuffer) {
		free(requestBuffer);
		return E_FAILURE;
	}

	memset((char *)&concatenatedResponseBuffer, 0, sizeof(PDUForm));

	pduRequestHeader = (SdpPduHeader *)requestBuffer;
	pduRequestHeader->pduId = SDP_ServiceAttributeRequest;

	pdata = (char *)(requestBuffer + sizeof(SdpPduHeader));
	requestSize = sizeof(SdpPduHeader);

	/*
	 ** Add the service record handle
	 */
	__put_u32(pdata, htonl(svcHandle));
	requestSize += sizeof(ServiceRecordHandle);
	pdata += sizeof(ServiceRecordHandle);

	/*
	 ** Add the maxAttrIDByteCount specifying response limit
	 */
	__put_u16(pdata, htons(maxAttrIDByteCount));
	requestSize += sizeof(uint16_t);
	pdata += sizeof(uint16_t);

	/*
	 ** Get attr seq PDU form 
	 */
	seqLength = generateAttrIDSequencePduForm(pdata, attrIDList, 
			((attrReqType == IndividualAttributes) ? UnsignedInteger_16 : UnsignedInteger_32));
	if (seqLength == -1) {
		status = E_INVALID_ARGS;
		goto exit;
	} 

	pdata += seqLength;
	requestSize += seqLength;
	BTDEBUG("Attr list length : %d\n", seqLength);

	// save before Continuation State
	_pdata = pdata;
	_requestSize = requestSize;

	do {
		int responseCount = 0;
		/*
		 ** Add NULL continuation state
		 */
		requestSize = _requestSize + copyContinuationState(_pdata, pCState);

		/*
		 ** Set the request header's param length
		 */
		pduRequestHeader->transactionId = htons(generateTransactionId());
		pduRequestHeader->paramLength = htons(requestSize - sizeof(SdpPduHeader));

		status = sendRequestWaitForResponse(srvHandle, requestBuffer, 
				responseBuffer, requestSize, &responseSize);

		if (status != E_OK)
			goto exit;

		pduResponseHeader = (SdpPduHeader *)responseBuffer;
		if (pduResponseHeader->pduId == SDP_ErrorResponse) {
			status = E_FAILURE;
			goto exit;
		}

		pdata = responseBuffer + sizeof(SdpPduHeader);
		responseCount  = ntohs(__get_u16(pdata));
		attrListByteCount += responseCount;
		pdata += sizeof(uint16_t);

		/*
		 ** Check if we have continuation state set, if yes, we
		 ** need to re-issue request before we parse ..
		 */
		cStateLength = __get_u8(pdata + responseCount);
		BTDEBUG("Response id : %d\n", pduResponseHeader->pduId);
		BTDEBUG("Attrlist byte count : %d\n", responseCount);
		BTDEBUG("CState length : %d\n", cStateLength);
		/*
		 ** This is a split response, need to concatenate the intermediate
		 ** responses as well as the last one which will have cStateLength == 0
		 */
		if ((cStateLength > 0) || (concatenatedResponseBuffer.dataSize != 0)) {
			char *targetPtr = NULL;

			if (cStateLength > 0)
				pCState = (CState *)(pdata + responseCount);
			else
				pCState = NULL;

			/*
			 ** Build concatenated response buffer
			 */
			concatenatedResponseBuffer.data = (char *)realloc(
					concatenatedResponseBuffer.data, 
					concatenatedResponseBuffer.dataSize + responseCount);
			concatenatedResponseBuffer.bufferSize = 
				concatenatedResponseBuffer.dataSize + responseCount;
			targetPtr = concatenatedResponseBuffer.data + 
				concatenatedResponseBuffer.dataSize;
			memcpy(targetPtr, pdata, responseCount);
			concatenatedResponseBuffer.dataSize += responseCount;
		}
	} while (pCState);

	if (attrListByteCount > 0) {
		int bytesScanned = 0;

		if (concatenatedResponseBuffer.dataSize != 0) {
			pdata = concatenatedResponseBuffer.data;
		}
		svcRec = clientExtractPDU(pdata, svcHandle, &bytesScanned);
		BTDEBUG("In SDPServiceAttributeRequest function \n");
		BTDEBUG("The status of extractServiceRecord is %d\n", status);
		BTDEBUG("The svcHandle is 0x%x\n", svcHandle);
		if (svcRec != NULL) {
			*_svcRec = svcRec;
			*maxAttrResponseByteCount = bytesScanned;
		} else
			status = E_FAILURE;
	}
exit:
	free(requestBuffer);
	free(responseBuffer);
	if (concatenatedResponseBuffer.data != NULL)
		free(concatenatedResponseBuffer.data);
	return status;
}

int SDPServiceAttributeRequest(
		BD_ADDR bdAddr,
		ServiceRecordHandle svcHandle,
		AttributeRequestType attrReqType,
		GSList *attrIDList,
		uint16_t maxAttrIDByteCount,
		ServiceRecord	**svcRec,
		uint16_t *maxAttrResponseByteCount)
{
	int status = 0;
	SDPServerHandle srvHandle;

	srvHandle = SDPOpenServerConnection(bdAddr);
	if (srvHandle < 0)
		return srvHandle;

	status = __SDPServiceAttributeRequest(srvHandle, svcHandle, attrReqType,
			attrIDList, maxAttrIDByteCount, svcRec, maxAttrResponseByteCount);
	SDPCloseServerConnection(srvHandle);

	return status;
}


/*
** This is a service search request combined with the service
** attribute request. First a service class match is done and
** for matching service, requested attributes are extracted
**
** INPUT :
**
**   GSList *svcSearchList
**     Singly linked list containing elements of the search
**     pattern. Each entry in the list is a UUID(DataTypeUUID_16)
**     of the service to be searched
**
**   AttributeSpecification attrSpec
**     Attribute identifiers are 16 bit unsigned integers specified
**     in one of 2 ways described below :
**     IndividualAttributes - 16bit individual identifiers
**        They are the actual attribute identifiers in ascending order
**
**     RangeOfAttributes - 32bit identifier range
**        The high-order 16bits is the start of range
**        the low-order 16bits are the end of range
**        0x0000 to 0xFFFF gets all attributes
**
**   GSList *attrIDList
**     Singly linked list containing attribute identifiers desired.
**     Every element is either a uint16_t(attrSpec = IndividualAttributes)  
**     or a uint32_t(attrSpec=RangeOfAttributes)
**
**   uint16_t maxAttrIDByteCount
**     The byte count of the number of attribute IDs specified
**     in the request list
**
** OUTPUT :
**   int return value
**     E_OK 
**       The request completed successfully. This does not
**       mean the requested services were found
**     E_TIMEOUT 
**       The request completed unsuccessfully due to a timeout
**
**   GSList **svcResponseList
**     This variable is set on a successful return to point to
**     service(s) found. Each element of this list is of type
**     ServiceRecordHandle (of the services which matched the search list)
**
**   uint16_t *maxAttrResponseByteCount
**     This is a pointer to a 16 bit integer, which is set to 
**     indicate the number of bytes of attributes returned. 
**     This pointer is set on successful return
**
*/
int __SDPServiceSearchAttributeRequest(
		SDPServerHandle srvHandle,
		GSList *svcSearchList, 
		AttributeRequestType attrReqType,
		GSList *attrIDList,
		uint16_t maxAttrByteCount,
		GSList **svcResponseList,
		uint16_t *maxAttrResponseByteCount
		)
{
	int status = 0;
	int requestSize = 0, _requestSize;
	int responseSize = 0;
	int seqLength = 0;
	int attrListByteCount = 0;
	char *pdata = NULL, *_pdata;
	char *requestBuffer =  NULL;
	char *responseBuffer =  NULL;
	SdpPduHeader *pduRequestHeader;
	SdpPduHeader *pduResponseHeader;
	DataTypeDescriptor dataType;
	GSList *svcRecHandleList = NULL;
	PDUForm concatenatedResponseBuffer;
	CState	*pCState = NULL;
	int bytesScanned = 0;

	if ((attrReqType != IndividualAttributes) && (attrReqType != RangeOfAttributes))
		return E_FAILURE;
	
	requestBuffer = (char *)malloc(RequestBufferSize);
	if (!requestBuffer)
		return E_FAILURE;
	responseBuffer = (char *)malloc(ResponseBufferSize);
	if (!responseBuffer) {
		free(requestBuffer);
		return E_FAILURE;
	}

	memset((char *)&concatenatedResponseBuffer, 0, sizeof(PDUForm));

	pduRequestHeader = (SdpPduHeader *)requestBuffer;
	pduRequestHeader->pduId = SDP_ServiceSearchAttributeRequest;

	// generate PDU
	pdata = (char *)(requestBuffer + sizeof(SdpPduHeader));
	requestSize = sizeof(SdpPduHeader);

	/*
	 ** Add service class IDs for search
	 */
	seqLength = generateSearchSequencePduForm(pdata, svcSearchList);

	BTDEBUG("Data seq added : %d\n", seqLength);

	/*
	 ** Now set the length and increment the pointer
	 */
	requestSize += seqLength;
	pdata += seqLength;

	/*
	 ** Specify the maximum attr count that client expects
	 */
	__put_u16(pdata, htons(maxAttrByteCount));
	requestSize += sizeof(uint16_t);
	pdata += sizeof(uint16_t);

	BTDEBUG("Max attr byte count : %d\n", maxAttrByteCount);

	/*
	 ** Get attr seq PDU form 
	 */
	seqLength = generateAttrIDSequencePduForm(pdata, attrIDList, 
			((attrReqType == IndividualAttributes) ? UnsignedInteger_16 : UnsignedInteger_32));
	if (seqLength == -1) {
		status = E_INVALID_ARGS;
		goto exit;
	}
	pdata += seqLength;
	BTDEBUG("Attr list length : %d\n", seqLength);
	requestSize += seqLength;

	// save before Continuation State
	_pdata = pdata;
	_requestSize = requestSize;

	do {
		int responseCount = 0;
		int cStateLength = 0;

		pduRequestHeader->transactionId = htons(generateTransactionId());
		/*
		 * Add continuation state
		 * pCState can be null
		 */
		requestSize = _requestSize + copyContinuationState(_pdata, pCState);

		/*
		 ** Set the request header's param length
		 */
		pduRequestHeader->paramLength = htons(requestSize - sizeof(SdpPduHeader));

		status = sendRequestWaitForResponse(srvHandle, requestBuffer, 
				responseBuffer, requestSize, &responseSize);

		if (status != 0)
			goto exit;

		pduResponseHeader = (SdpPduHeader *)responseBuffer;
		if (pduResponseHeader->pduId == SDP_ErrorResponse) {
			status = E_FAILURE;
			goto exit;
		}

		pdata = responseBuffer + sizeof(SdpPduHeader);
		responseCount = ntohs(__get_u16(pdata));
		attrListByteCount += responseCount;
		pdata += sizeof(uint16_t);	// pdata points to attribute list

		/*
		 ** Check if we have continuation state set, if yes, we
		 ** need to re-issue request before we parse ..
		 */
		cStateLength = __get_u8(pdata + responseCount);
		BTDEBUG("Attrlist byte count : %d\n", attrListByteCount);
		BTDEBUG("Response byte count : %d\n", responseCount);
		BTDEBUG("Cstate length : %d\n", cStateLength);
		/*
		 ** This is a split response, need to concatenate the intermediate
		 ** responses as well as the last one which will have cStateLength == 0
		 */
		if ((cStateLength > 0) || (concatenatedResponseBuffer.dataSize != 0)) {
			char *targetPtr = NULL;

			if (cStateLength > 0) 
				pCState = (CState *)(pdata + responseCount);
			else
				pCState = NULL;
			/*
			 ** Build concatenated response buffer
			 */
			concatenatedResponseBuffer.data = (char *)realloc(
					concatenatedResponseBuffer.data, 
					concatenatedResponseBuffer.dataSize + responseCount);
			targetPtr = concatenatedResponseBuffer.data + 
				concatenatedResponseBuffer.dataSize;
			concatenatedResponseBuffer.bufferSize = 
				concatenatedResponseBuffer.dataSize + responseCount;
			memcpy(targetPtr, pdata, responseCount);
			concatenatedResponseBuffer.dataSize += responseCount;
		}
	} while (pCState);

	if (attrListByteCount <= 0)
		goto exit;

	if (concatenatedResponseBuffer.dataSize != 0)
		pdata = concatenatedResponseBuffer.data;
	/*
	 ** Response is a sequence of sequence(s) for one or
	 ** more data element sequence(s) representing services
	 ** for which attributes are returned
	 */
	pdata = extractSequenceType(pdata, &dataType, &seqLength, &bytesScanned);

	BTDEBUG("Bytes scanned : %d\n", bytesScanned);
	BTDEBUG("Seq length : %d\n", seqLength);
	if ((pdata == NULL) || (seqLength == 0)) {
		*maxAttrResponseByteCount = 0;
		goto exit;
	}


	for (;;) {
		int localSeqSize;
		ServiceRecord *svcRec = NULL;

		if (pdata != NULL) {
			localSeqSize = 0;
			svcRec = clientExtractPDU(pdata, 0xffffffff, &localSeqSize);
			if (svcRec == NULL) {
				BTERROR("SVC REC is null\n");
				status = E_FAILURE;
				break;
			}

			bytesScanned += localSeqSize;
			*maxAttrResponseByteCount += localSeqSize;
			pdata += localSeqSize;

			BTDEBUG("Loc seq length : %d\n", localSeqSize);
			BTDEBUG("Svc Rec Handle : 0x%x\n", svcRec->serviceRecordHandle);
			BTDEBUG("Bytes scanned : %d\n", bytesScanned);
			BTDEBUG("Attrlist byte count : %d\n", attrListByteCount);
			svcRecHandleList = g_slist_append(svcRecHandleList, svcRec);

			if (attrListByteCount == bytesScanned) {
				BTDEBUG("Successful scan of service attr lists\n");
				*svcResponseList = svcRecHandleList;
				break;
			}
		}
		else {
			BTDEBUG("This may be a spec interpretation anomaly !");
			BTDEBUG("Expect sequence of sequence(s)\n");
			status = E_FAILURE;
			break;
		}
	}
exit:
	free(requestBuffer);
	free(responseBuffer);
	if (concatenatedResponseBuffer.data != NULL)
		free(concatenatedResponseBuffer.data);
	return status;
}

int SDPServiceSearchAttributeRequest(
		BD_ADDR bdAddr,
		GSList *svcSearchList, 
		AttributeRequestType attrReqType,
		GSList *attrIDList,
		uint16_t maxAttrByteCount,
		GSList **svcResponseList,
		uint16_t *maxAttrResponseByteCount
		)
{
	int status = 0;
	SDPServerHandle srvHandle;

	srvHandle = SDPOpenServerConnection(bdAddr);
	if (srvHandle < 0)
		return srvHandle;

	status = __SDPServiceSearchAttributeRequest(srvHandle, svcSearchList, attrReqType, attrIDList,
			maxAttrByteCount, svcResponseList, maxAttrResponseByteCount);
	SDPCloseServerConnection(srvHandle);

	return status;
}


int SDPServiceSearchPortRequest(BD_ADDR bda, UUID16Bit uuid16Bit)
{
	int		status = -1, i, s, p;
	uint16_t	count = 0;
	GSList		*pSeq = NULL;
	GSList		*attrList = NULL;
	uint16_t	attr1 = AttrServiceRecordHandle;
	uint16_t	attr2 = AttrProtocolDescriptorList;
	GSList		*searchList = NULL;
	UUID		uuid;
	AccessProtocols	*pAccessProtocols = NULL;
	SDPServerHandle srvHandle;
	ServiceRecord	*svcRec;

	status = SDPInit(0);
	if (status != E_OK)
		return status;

	srvHandle = SDPOpenServerConnection(bda);
	if (srvHandle < 0)
		return srvHandle;

	attrList = g_slist_append(attrList, &attr1);
	attrList = g_slist_append(attrList, &attr2);

	makeUUIDFrom16Bits(&uuid, uuid16Bit);
	searchList = g_slist_append(searchList, &uuid);

	status = __SDPServiceSearchRequest(srvHandle, searchList, 1, &pSeq, &count);
	//printf("Status: %d, count: %d\n", status, count);

	if (status  || (count == 0))
			goto exit;

	for (i = 0; i < (int)g_slist_length(pSeq); i++) {
		ServiceRecordHandle svcRecHandle = *(ServiceRecordHandle *)g_slist_nth_data(pSeq, i);
		BTDEBUG("Record Handle: %#x\n", svcRecHandle);
		status = __SDPServiceAttributeRequest(srvHandle, svcRecHandle,
				IndividualAttributes, attrList, 0xFFFF, &svcRec, &count);
		if (status != 0)
			goto exit;

		if (getAccessProtocols(svcRec, &pAccessProtocols) == E_OK) {
			for (s = 0; s < (int)g_slist_length(pAccessProtocols->pSequence); s++ ) {
				GSList *protDescSeq = (GSList*)g_slist_nth_data(pAccessProtocols->pSequence, s);
				for (p = 0; p < (int)g_slist_length(protDescSeq); p++ ) {
					ProtocolDescriptor *pPD = (ProtocolDescriptor*)g_slist_nth_data(protDescSeq, p);
					if (UUID2value(&pPD->protocolId) == RFCOMM_UUID) {
						if (pPD->portNumber) {
							status = pPD->portNumber;
						}
						goto exit;
					}
				}
			}
		}
	}
exit:
	SDPCloseServerConnection(srvHandle);
	return status;
}


