/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author:Guruprasad Krishnamurthy <guruprasad.krishnamurthy@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   */

/*
   $Id: sdpclt_utils.c,v 1.1.1.1 2002/07/11 11:21:57 kds Exp $

   Contains utilities for use by SDP service discovery clients
   to extract individual attributes from a service record
   All utility functions below have a signature 
   int getXXX(ServiceRecordHandle svcRec, XXX *pXXX)
   The pointer pXXX is set to point at the attribute value if
   it is present in the service record and 0 is returned, else
   E_NOT_EXIST is returned

Fixes:     Guruprasad Krishnamurthy <guruprasad.krishnamurthy@nokia.com>
Manel Guerrero Zapata <manel.guerrero-zapata@nokia.com>
*/

#include <affix/config.h>

#include <stdio.h>
#include <malloc.h>
#include <string.h>

#include <affix/sdp.h>
#include <affix/sdpclt.h>
#include "servicedb.h"
#include "attr.h"


ServiceAttributeIdentifier svcRecStateAttrId = AttrServiceRecordState;
ServiceAttributeIdentifier svcAvailAttrId = AttrServiceAvailability;
ServiceAttributeIdentifier svcInfoTTLAttrId = AttrServiceInfoTimeToLive;
ServiceAttributeIdentifier svcDBStateAttrId = AttrServiceDatabaseState;
ServiceAttributeIdentifier svcNameAttrId = AttrServiceName_Primary;
ServiceAttributeIdentifier svcDescAttrId = AttrServiceDescription_Primary;
ServiceAttributeIdentifier svcProvNameAttrId = AttrServiceProviderName_Primary;
ServiceAttributeIdentifier docURLAttrId = AttrDocumentationURL;
ServiceAttributeIdentifier clientExecURLAttrId = AttrClientExecutableURL;
ServiceAttributeIdentifier iconURLAttrId = AttrIconURL;
ServiceAttributeIdentifier svcClassIdAttrId = AttrServiceClassIDList;
ServiceAttributeIdentifier browseGroupAttrId = AttrBrowseGroupList;
ServiceAttributeIdentifier profileDescAttrId = AttrBTProfileDescriptorList;
ServiceAttributeIdentifier protocolDescAttrId = AttrProtocolDescriptorList;
ServiceAttributeIdentifier svcIDAttrId = AttrServiceID;
ServiceAttributeIdentifier groupIDAttrId = AttrGroupID;
ServiceAttributeIdentifier versionNumberListAttrId = AttrVersionNumberList;


int getAccessProtocols(ServiceRecord *svcRec, AccessProtocols **pap)
{
	int status = E_NOT_EXIST;
	SDPData *pSDPData, *currSDPData, *currEntrySDPData;
	int i = 0;
	int count = 0;
	AccessProtocols *ap = NULL;

	*pap = NULL;
	pSDPData = getSDPData(svcRec, protocolDescAttrId);
	if (pSDPData != NULL) {
		ap = (AccessProtocols *) malloc(sizeof(AccessProtocols));
		count = pSDPData->alternatesCount; 
		ap->pSequence = NULL;
		BTDEBUG("AP count : %d\n", count);
		BTDEBUG("AP type : 0%x\n", pSDPData->dtd);
		for (i = 0; i <= count; i++) {
			GSList *protDescSeq = NULL;
			if (i == 0) {
				currSDPData = pSDPData->sdpDataValue.dataSeq; 
			}
			else {
				BTDEBUG("Alt count : %d\n", i);
				currSDPData = pSDPData->alternates[i-1]->sdpDataValue.dataSeq;
			} 
			while(currSDPData != NULL) {
				currEntrySDPData = currSDPData->sdpDataValue.dataSeq;
				if (currEntrySDPData != NULL) {
					ProtocolDescriptor *pDesc = (ProtocolDescriptor *) 
						malloc(sizeof(ProtocolDescriptor));               
					bzero(pDesc, sizeof(ProtocolDescriptor));
					BTDEBUG("Getting protocol id : %x\n", currEntrySDPData->dtd);
					pDesc->protocolId = currEntrySDPData->sdpDataValue.uuid;
					printUUID(&pDesc->protocolId);
					currEntrySDPData= currEntrySDPData->next;
					if (currEntrySDPData != NULL) {
						if (currEntrySDPData->dtd == UnsignedInteger_16) {
							pDesc->portNumber = currEntrySDPData->sdpDataValue.uint16;
							pDesc->portNumberSize = 2;
						}
						else {
							pDesc->portNumber = currEntrySDPData->sdpDataValue.uint8;
							pDesc->portNumberSize = 1;
						}
						BTDEBUG("Port : %d\n", pDesc->portNumber);
						currEntrySDPData= currEntrySDPData->next;
					}
					if (currEntrySDPData != NULL) {
						pDesc->versionNumber = currEntrySDPData->sdpDataValue.uint16;
						BTDEBUG("Vnum : %d\n", pDesc->versionNumber);
					}
					protDescSeq = g_slist_append(protDescSeq, pDesc);
				}
				currSDPData = currSDPData->next;
			}
			ap->pSequence = g_slist_append(ap->pSequence, protDescSeq);
		}
		*pap = ap;
		status = E_OK;
	}
	return status ;
}

int getServiceClassIDList(ServiceRecord *svcRec, GSList **svcClassSeq)
{
	GetUUIDSequenceAttribute(svcClassIdAttrId, svcClassSeq); 
}

int getBrowseGroupList(ServiceRecord *svcRec, GSList **ppGSList)
{
	GetUUIDSequenceAttribute(browseGroupAttrId, ppGSList); 
}


int getProtocolDescList(ServiceRecord *svcRec, GSList **ppGSList)
{
	GetUUIDSequenceAttribute(protocolDescAttrId, ppGSList); 
}


int getServiceID(ServiceRecord *svcRec, UUID *pUUID)
{
	GetBasicAttribute(svcIDAttrId, pUUID, uuid);
}


int getGroupID(ServiceRecord *svcRec, UUID *pUUID)
{
	GetBasicAttribute(groupIDAttrId, pUUID, uuid);
}


unsigned char isGroup(ServiceRecord *svcRec)
{
	unsigned char exists = 0;

	SDPData *pSDPData = getSDPData(svcRec, groupIDAttrId);
	if ((pSDPData != NULL) && 
			(svcRec->serviceRecordHandle != SDPServerServiceRecordHandle)) {
		exists = 1;
	}
	return exists;
}


int getServiceRecordState(ServiceRecord *svcRec, ServiceRecordState *svcRecState)
{
	GetBasicAttribute(svcRecStateAttrId, svcRecState, uint32);
}


int getServiceAvailability(ServiceRecord *svcRec, ServiceAvailability *svcAvail)
{
	GetBasicAttribute(svcAvailAttrId, svcAvail, uint8);
}


int getServiceInfoTimeToLive(ServiceRecord *svcRec, ServiceInfoTimeToLive *svcTTLInfo)
{
	GetBasicAttribute(svcInfoTTLAttrId, svcTTLInfo, uint32);
}


int getLanguageAttributes(ServiceRecord *svcRec, GSList **langSeq)
{
	return 0; 
}


int getProfileDescriptor(ServiceRecord *svcRec, GSList **profDescSeq)
{
	int status = E_NOT_EXIST;
	ProfileDescriptor *profDesc;
	SDPData *pSDPData = NULL;

	*profDescSeq = NULL;
	pSDPData = getSDPData(svcRec, profileDescAttrId);
	if ((pSDPData != NULL) && (pSDPData->sdpDataValue.dataSeq != NULL)) {
		SDPData *pDataSeq = pSDPData->sdpDataValue.dataSeq;
		while ((pDataSeq != NULL) && (pDataSeq->sdpDataValue.dataSeq != NULL)) {
			SDPData *pUUID = NULL;
			SDPData *pVnum = NULL;

			pUUID = pDataSeq->sdpDataValue.dataSeq;
			pVnum = pDataSeq->sdpDataValue.dataSeq->next;
			if ((pUUID != NULL) && (pVnum != NULL)) {
				profDesc = (ProfileDescriptor *)malloc(sizeof(ProfileDescriptor));
				profDesc->profileUUID =  pUUID->sdpDataValue.uuid;
				profDesc->vnum = pUUID->sdpDataValue.uint16;
				printUUID(&profDesc->profileUUID);
				BTDEBUG("Vnum : %d\n", profDesc->vnum);
				*profDescSeq = g_slist_append(*profDescSeq, profDesc);
				status = E_OK;
			}
			pDataSeq = pDataSeq->next;
		}
	}

	return status;
}


int getServiceName(ServiceRecord *svcRec, char **pName)
{
	GetStringAttribute(svcNameAttrId, pName);
}


int getServiceDescription(ServiceRecord *svcRec, char **pDesc)
{
	GetStringAttribute(svcDescAttrId, pDesc);
}


int getServiceProviderName(ServiceRecord *svcRec, char **pSvcProviderName)
{
	GetStringAttribute(svcProvNameAttrId, pSvcProviderName);
}


int getDocumentationURL(ServiceRecord *svcRec, char **pDocURL)
{
	GetStringAttribute(docURLAttrId, pDocURL);
}


int getClientExecutableURL(ServiceRecord *svcRec, char **pClientURL)
{
	GetStringAttribute(clientExecURLAttrId, pClientURL);
}


int getIconURL(ServiceRecord *svcRec, char **pIconURL)
{
	GetStringAttribute(iconURLAttrId, pIconURL);
}


void printServiceRecordById(ServiceRecord *svcRec)
{
	// printServiceRecord(bdAddr, svcRecHandle);
	printf("printServiceRecordById: Print the ServiceRecord\n");
	// printPduForm(&svcRec->pduForm);
	generateServiceRecordPduForm(svcRec);
	printSDPDataAsServiceRecord(svcRec);

}


int getSDPServerServiceDatabaseState(ServiceRecord *svcRec, ServiceDatabaseState *svcDBState)
{
	GetBasicAttribute(svcDBStateAttrId, svcDBState, uint32);
}


int getSDPServerVersionNumber(ServiceRecord *svcRec, GSList **pVersionNumberSeq)
{
	int status = E_NOT_EXIST;
	SDPData *pSDPData, *currSDPData;

	*pVersionNumberSeq = NULL;
	pSDPData = getSDPData(svcRec, versionNumberListAttrId);
	if (pSDPData != NULL) {
		for (currSDPData = pSDPData->sdpDataValue.dataSeq;
				currSDPData != NULL;
				currSDPData = currSDPData->next)
		{ 
			*pVersionNumberSeq = g_slist_append(*pVersionNumberSeq, 
					&currSDPData->sdpDataValue.uint16);
		}
		status = E_OK;
	}
	return status;
}


