/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author:Guruprasad Krishnamurthy <guruprasad.krishnamurthy@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   */

/*
   $Id: sdpserver.c,v 1.6 2002/08/30 13:28:36 kds Exp $

   Contains the implementation of the SDP server itself including
   the main() of the server and it's initialization

	Fixes:     Guruprasad Krishnamurthy <guruprasad.krishnamurthy@nokia.com>
		Imre Deak <ext-imre.deak@nokia.com>
		Dmitry Kasatkin
*/

#include <affix/config.h>

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <getopt.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <affix/bluetooth.h>

#include <affix/sdp.h>
#include <affix/sdpclt.h>
#include <affix/sdpsrv.h>
#include "utils.h"
#include "queue.h"
#include "servicedb.h"
#include "match.h"
#include "lock.h"
#include "pdu.h"
#include "attr.h"
#include "cstate.h"
#include "request.h"

//#define SDP_THREADED

fd_set activeFdSet;
int serverBTSockFd;
int serverTCPSockFd;
int maxFd;
int argind;

/*
 ** Server and client socket addresses
 */
struct sockaddr_affix serverBTAddr;
struct sockaddr_in serverINAddr;
struct sockaddr_affix clientBTAddr;
struct sockaddr_in clientINAddr;

extern void startWorkers(int);

UUID16Bit sdpSrvSvcClassID = SDPServerServiceClassID;
UUID sdpSrvUUID;
GSList *versionNumberSeq = NULL;
ServiceAttributeIdentifier svcDBTimeStampAttrId = AttrServiceDatabaseState;
ServiceAttributeIdentifier svcRecAttrId = AttrServiceRecordHandle;

static UUID16Bit publicBrowseGroupSvcClassID = BrowseGroupDescServiceClassID;

ServiceRecord *sdpSvcRec;
ServiceRecord *publicBrowseGroupRoot;

/*
 ** List of version numbers supported
 ** by the SDP server. Add to this list
 ** when newer versions are supported
 */
const int sdpServerVnumEntries = 1;
SDPServerVersionNumber sdpVnumArray[1/*sdpServerVnumEntries*/] = {
	{1, 0} 
};

const char *pSvcName = "SDP Server";
const char *pProvName = "Nokia Research Center";
const char *pSvcDesc = "Affix Service Discovery server";

const char *pBrowseGroupName = "Public Browse Group Root";
const char *pGroupDesc = "Root of public group hierarchy";

/*
 ** The service DB state is an attribute of the service record
 ** of the SDP server itself. This attribute is guaranteed to
 ** change if any of the contents of the service repository
 ** changes. This function updates the timestamp of value of
 ** the svcDBState attribute
 ** Set the SDP server DB. Simply a timestamp which is the marker
 ** when the DB was modified.
 */
void updateSDPServerDBTimestamp(void)
{
	uint32_t dbts = getTimeInSeconds();
	SDPData *pData = createSDPData(UnsignedInteger_32, (char *)&dbts);
	addToAttributeList(sdpSvcRec, svcDBTimeStampAttrId, pData, 1);
	generateServiceRecordPduForm(sdpSvcRec);
}


int registerPublicBrowseGroupRoot(void)
{
	int status = 0;
	char *groupName;
	char *groupDesc;
	char *groupProvName;
	GSList *browseClassIDList = NULL;
	UUID *pBrowseGroupSvcClassID;
	UUID publicBrowseGroupID;
	SDPData *pSDPData;

	publicBrowseGroupRoot = createServiceRecord();
	publicBrowseGroupRoot->serviceRecordHandle =
		(ServiceRecordHandle)publicBrowseGroupRoot;
	addServiceRecord(publicBrowseGroupRoot);
	pSDPData = createSDPData(UnsignedInteger_32, 
			(char *)&publicBrowseGroupRoot->serviceRecordHandle);
	createSDPAttribute(publicBrowseGroupRoot, svcRecAttrId, pSDPData);

	groupName = (char *)malloc(SequencePduFormSize);
	groupDesc = (char *)malloc(SequencePduFormSize);
	groupProvName = (char *)malloc(SequencePduFormSize);

	strcpy(groupName, pBrowseGroupName);
	strcpy(groupProvName, pProvName);
	strcpy(groupDesc, pGroupDesc);
	status = setInformationalAttributes(publicBrowseGroupRoot, groupName, 
			groupProvName, groupDesc);

	pBrowseGroupSvcClassID = (UUID *)malloc(sizeof(UUID));
	makeUUIDFrom16Bits(pBrowseGroupSvcClassID, publicBrowseGroupSvcClassID);
	browseClassIDList = g_slist_append(browseClassIDList, pBrowseGroupSvcClassID);

	status = setServiceClassID(publicBrowseGroupRoot, browseClassIDList);

	makeUUIDFrom16Bits(&publicBrowseGroupID, PublicBrowseGroup);
	status = setGroupID(publicBrowseGroupRoot, publicBrowseGroupID);

	generateServiceRecordPduForm(publicBrowseGroupRoot);

	return status;
}

/*
 ** The SDP server must present it's own service record to
 ** the service repository. This can be accessed by service
 ** discovery clients. This methods constructs a service record
 ** and stores it in the reposssitory
 */
int registerSDPServerService(void)
{
	int status = 0, i;
	GSList *svcClassIDList = NULL;
	GSList *browseList = NULL;
	AccessProtocols *accessProtocols = NULL;
	UUID *pSdpSrvSvcClassID = NULL;
	UUID *pBrowseGroupId = NULL;
	char *svcName;
	char *svcProvName;
	char *svcDesc;
	SDPData *pSDPData;
	GPtrArray *vnumArray;
	GPtrArray *vnumDTDArray;
	DataTypeDescriptor dtd;
	SDPData *pData;
	ProtocolDescriptor *pd;
	GSList *pSeq = NULL;
	ServiceAttributeIdentifier attrId;

	svcName = (char *)malloc(SequencePduFormSize);
	svcDesc = (char *)malloc(SequencePduFormSize);
	svcProvName = (char *)malloc(SequencePduFormSize);

	sdpSvcRec = createServiceRecord();
	sdpSvcRec->targetPattern = NULL;
	/*
	 ** Force the svcRecHandle to be SDPServerServiceRecordHandle
	 */
	sdpSvcRec->serviceRecordHandle = SDPServerServiceRecordHandle;
	addServiceRecord(sdpSvcRec);
	pSDPData = createSDPData(UnsignedInteger_32, 
			(char *)&sdpSvcRec->serviceRecordHandle);
	createSDPAttribute(sdpSvcRec, svcRecAttrId, pSDPData);

	/*
	 ** Add all attributes to service record. No need to
	 ** commit since we are the server and this record is
	 ** already in the serviceDB
	 */
	strcpy(svcName, pSvcName);
	strcpy(svcProvName, pProvName);
	strcpy(svcDesc, pSvcDesc);

	status = setInformationalAttributes(sdpSvcRec, svcName, 
			svcProvName, svcDesc);

	pSdpSrvSvcClassID = (UUID *)malloc(sizeof(UUID));
	makeUUIDFrom16Bits(pSdpSrvSvcClassID, sdpSrvSvcClassID);
	svcClassIDList = g_slist_append(svcClassIDList, pSdpSrvSvcClassID);

	status = setServiceClassID(sdpSvcRec, svcClassIDList);

	/*
	 ** Set the version numbers supported, these are passed as arguments
	 ** to the server on command line. Now defaults to 1.0
	 ** Build the version number sequence first
	 */
	vnumArray = g_ptr_array_new();
	vnumDTDArray = g_ptr_array_new();
	dtd = UnsignedInteger_16;
	for (i = 0; i < sdpServerVnumEntries; i++) {
		uint16_t *pVNum;

		pVNum = (uint16_t *)malloc(sizeof(uint16_t));
		*pVNum = sdpVnumArray[i].major;
		*pVNum = (*pVNum << 8);
		*pVNum |= sdpVnumArray[i].minor;
		g_ptr_array_add(vnumArray, pVNum);
		g_ptr_array_add(vnumDTDArray, &dtd);
	}
	pData = createSDPDataSequence(vnumDTDArray, vnumArray);
	attrId = AttrVersionNumberList;
	createSDPAttribute(sdpSvcRec, attrId, pData);

	makeUUIDFrom16Bits(&sdpSrvUUID, SDP_UUID);
	status = setServiceID(sdpSvcRec, sdpSrvUUID);

	pd = (ProtocolDescriptor *)malloc(sizeof(ProtocolDescriptor));
	makeUUIDFrom16Bits(&pd->protocolId, (UUID16Bit)L2CAP_UUID);
	pd->portNumber = SDP_PSM;
	pd->versionNumber = 1;

	accessProtocols = (AccessProtocols *)malloc(sizeof(AccessProtocols));
	pSeq = g_slist_append(pSeq, pd);
	accessProtocols->pSequence = NULL;
	accessProtocols->pSequence = g_slist_append(accessProtocols->pSequence,pSeq);

	status = setAccessProtocols(sdpSvcRec, accessProtocols);

	pBrowseGroupId = (UUID *)malloc(sizeof(UUID));
	makeUUIDFrom16Bits(pBrowseGroupId, PublicBrowseGroup);
	browseList = g_slist_append(browseList, pBrowseGroupId);
	status = setBrowseGroupList(sdpSvcRec, browseList);

	updateSDPServerDBTimestamp();

	return 0;
}


/*
 ** SDP server initialization on startup includes creating
 ** the BT PSM and TCP sockets over which service discovery clients
 ** and service registration clients respectively access the
 ** SDP server
 */
int initServer(void)
{
	int status = 0;

	status = sdp_init(server);
	if (status != 0) {
		BTERROR("Could not find local BT device address, terminating");
		return -1;
	}

	/*
	 ** Register the public browse group Root
	 */
	registerPublicBrowseGroupRoot();

	/*
	 ** Register the SDP server as a service
	 */
	registerSDPServerService();

	/*
	 * init locks to access Service Database
	 */
	initLock();

#if defined(SDP_THREADED)
	initQueue();
	/*
	 ** Start 1 worker thread; this should be
	 ** configurable later
	 */
	startWorkers(1);
#endif

	/*
	 ** Create our socket on PSM 1 and
	 ** set activeFdSet
	 */
	serverBTSockFd = socket(PF_AFFIX, SOCK_STREAM, BTPROTO_L2CAP);
	if (serverBTSockFd == -1) {
		BTERROR("BT Socket creation failure:%s", strerror(errno));
		status = -1;
	} else {
		serverBTAddr.family = AF_AFFIX;
		serverBTAddr.port = SDP_PSM;
		serverBTAddr.bda = BDADDR_ANY;
		status = bind(serverBTSockFd, 
				(struct sockaddr *)&serverBTAddr, sizeof(serverBTAddr));
		if (status != -1) {
			listen(serverBTSockFd, 5);
			FD_SET(serverBTSockFd, &activeFdSet);
			maxFd = serverBTSockFd;
		} else {
			BTERROR("BT bind failure:%s", strerror(errno));
			status = -1;
		}
	}

	if (status != -1) {
		serverTCPSockFd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
		if (serverTCPSockFd == -1) {
			BTERROR("TCP socket creation failure:%s", strerror(errno));
			status = -1;
		}
		else {
			serverINAddr.sin_family = AF_INET;
			serverINAddr.sin_port = SDP_TCP_PORT;
			status = bind(serverTCPSockFd, 
					(struct sockaddr *)&serverINAddr, sizeof(serverINAddr));
			if (status != -1) {
				status = listen(serverTCPSockFd, 5);
				FD_SET(serverTCPSockFd, &activeFdSet);
				maxFd = serverTCPSockFd;
			}
			else {
				BTERROR("TCP bind failure:%s", strerror(errno));
				status = -1;
			}
		}
	}

	return status;
}


/*
 ** Accept connections from clients both
 ** local and remote
 */
int acceptBTConnection(void)
{
	socklen_t sockLength = sizeof(clientBTAddr);

	int clientSocket = accept(serverBTSockFd, 
			(struct sockaddr *)&clientBTAddr, &sockLength);
	if (clientSocket == -1) {
		BTERROR("Error accepting Bluetooth client connection:%s", strerror(errno));
	}

	return clientSocket;
}

int acceptTCPConnection(void)
{
	socklen_t sockLength = sizeof(clientINAddr);

	int clientSocket = accept(serverTCPSockFd, 
			(struct sockaddr *)&clientINAddr, &sockLength);
	if (clientSocket == -1) {
		BTERROR("Error accepting TCP client connection:%s", strerror(errno));
	}

	return clientSocket;
}


/*
 ** This is the SDP server main thread
 ** After appropriate initialization, uses the select()
 ** to field inputs from both the BT and TCP sockets.
 **
 ** Calls the processRequest() if single threaded else
 ** extracts the request and populates the request queue 
 ** for servicing by one or more worker threads that are
 ** active
 */

void sdpServerSignalHandler(int sig)
{
	if (sig == SIGUSR2) {
		/* check database and quite if nothing */
		if (emptyServiceDB()) {
			BTDEBUG("Service Database empty\n");
		} else
			return;
	}
	BTINFO("Sig handler : %d", sig);
	resetServiceDB();
	cleanupContinuationStateBuffer();
	close(serverBTSockFd);
	close(serverTCPSockFd);
	BTINFO("Terminating ..\n");
	exit(0);
}

int become_daemon(void)
{
	int 		fd;

	if (getppid() != 1) {
		signal(SIGTTOU, SIG_IGN);
		signal(SIGTTIN, SIG_IGN);
		signal(SIGTSTP, SIG_IGN);
		if (btdaemon != 2 && fork() != 0)
			return 0;	/* exit if parent */
		setsid();
	}
	for (fd = 0; fd < 3; fd++)
		close(fd);
	chdir("/");

	return 1;
}

void usage(void)
{
	fprintf(stderr, "%s %s [-d] [-b]\n"
			"  -d: start as daemon\n  -b: with debug info\n", argv[0], affix_version);
}

int main(int _argc, char **_argv)
{
	/*
	 ** Create our socket for connection
	 ** requests. Start worker threads
	 */
	int status = 0, lind, c;

	struct option	opts[] = {
		{"help", 0, 0, 'h'},
		{"daemon", 0, 0, 'd'},
		{"debug", 0, 0, 'b'},
		{0, 0, 0, 0}
	};

	argc = _argc;
	argv = _argv;

	btdaemon = 0;
	for (;;) {
		c = getopt_long(argc, argv, "+hdbD", opts, &lind);
		if (c == -1)
			break;
		switch (c) {
			case 'h':
				usage();
				return -1;
				break;
			case 'd':
				btdaemon = 1;
				break;
			case 'D':
				btdaemon = 2;	// do not fork
				break;
			case 'b':
				btlog_mask = BTDEBUG_MODULE;
				break;
			case ':':
				printf("Missing parameters for option: %c\n", optopt);
				return 1;
				break;
			case '?':
				printf("Unknown option: %c\n", optopt);
				return 1;
				break;
		}
	}

	argind = optind;

	BTINFO("%s %s started.", argv[0], affix_version);

	status = initServer();
	if (status != 0) {
		BTERROR("Terminating, init error");
		return -1;
	}

	if (btdaemon && !become_daemon())
		return 0;

	signal(SIGINT, sdpServerSignalHandler);
	signal(SIGTERM, sdpServerSignalHandler);
	signal(SIGABRT, sdpServerSignalHandler);
	signal(SIGQUIT, sdpServerSignalHandler);
	signal(SIGUSR2, sdpServerSignalHandler);

	for (;;) {
		int numFds;
		int newFd;
		fd_set checkFds;
		fd_set exceptFds;
#ifdef CONFIG_AFFIX_DEBUG
		static char log_buf[1024];
#endif

		FD_ZERO(&checkFds); FD_ZERO(&exceptFds);
		checkFds = activeFdSet;

		/*
		 ** Check on active fds
		 */
#ifdef CONFIG_AFFIX_DEBUG
		{
			int pos, i;

			BTDEBUG("Waiting for data maxFd : %d", maxFd);
			pos = sprintf(log_buf, "Checking fds :");
			for (i = 0; pos + 6 < (int)sizeof(log_buf) && i < maxFd+1; i++) {
				if (FD_ISSET(i, &checkFds)) {
					pos += sprintf(&log_buf[pos], " %d ", i);
				}
			}
			BTDEBUG("%s", log_buf);
			BTDEBUG("serverBTSockFd : %d  serverTCPSockFd : %d",
					serverBTSockFd, serverTCPSockFd);
		}
#endif
		numFds = select(maxFd+1, &checkFds, NULL, &exceptFds, NULL);
		if (numFds <= 0) {
			BTDEBUG("Select error:%s", strerror(errno));
			goto exit;
		}
#ifdef CONFIG_AFFIX_DEBUG
		{
			int pos, i;

			BTDEBUG("Data in : %d", numFds);
			BTDEBUG("serverBTSockFd : %d  serverTCPSockFd : %d",
					serverBTSockFd, serverTCPSockFd);
			pos = sprintf(log_buf, "Active fd(s) : ");
			for (i = 0; pos + 6 < (int)sizeof(log_buf) && i < maxFd+1; i++)
				if (FD_ISSET(i, &checkFds))
					pos += sprintf(&log_buf[pos], " %d ", i);
			BTDEBUG("%s", log_buf);

			pos = sprintf(log_buf, "Except fd : ");
			for (i = 0; pos + 6 < (int)sizeof(log_buf) && i < maxFd+1; i++)
				if (FD_ISSET(i, &exceptFds))
					pos += sprintf(&log_buf[pos], " %d ", i);
		}
#endif

		if (FD_ISSET(serverBTSockFd, &checkFds)) {
			/*
			 ** A new connection ..
			 */
			newFd = acceptBTConnection();
			BTDEBUG("new bt fd : %d", newFd); 
			if (newFd != -1) {
				if (newFd > maxFd)
					maxFd = newFd;

				FD_SET(newFd, &activeFdSet);
			}
		} else if (FD_ISSET(serverTCPSockFd, &checkFds)) {
			/*
			 ** A new TCP connection
			 */
			newFd = acceptTCPConnection();
			BTDEBUG("new tcp fd : %d", newFd);
			if (newFd != -1) {
				if (newFd > maxFd)
					maxFd = newFd;

				FD_SET(newFd, &activeFdSet);
			}
		} else {
			int checkCount, fd;

			checkCount = 0;
			for (fd = 0; fd <= maxFd; fd++) {
				SDPRequest *sdpReq;
				char *realBuf;
				int bufSize;

				if (fd == serverBTSockFd || fd == serverTCPSockFd)
					continue;
				if (FD_ISSET(fd, &checkFds)) {
					int bytesRead = 0;
					SdpPduHeader sdpHeader;

					BTDEBUG("Reading from : %d", fd);
					bytesRead = recv(fd, (void *)&sdpHeader, 
							sizeof(SdpPduHeader), MSG_PEEK);
					BTDEBUG("Bytes read : %d", bytesRead);
					if ((bytesRead < 0) || (bytesRead == 0)) {
						if (bytesRead < 0) {
							BTERROR("Read error:%s", strerror(errno));
						}
						closeConnection(fd);
						BTDEBUG("Conn closed: %d, read: %d", fd, bytesRead);
						FD_CLR(fd, &activeFdSet);
						if (fd == maxFd)
							maxFd--;
					} else {
						bufSize = sizeof(SdpPduHeader) + 
							ntohs(sdpHeader.paramLength);
						BTDEBUG("Allocating size : %d", bufSize);
						realBuf = (char *)malloc(bufSize);
						bytesRead = recv(fd, realBuf, bufSize, 0);
						BTDEBUG("Bytes read : %d", bytesRead);

						if ((bytesRead < 0) || (bytesRead == 0)) {
							if (bytesRead < 0) {
								BTERROR("Read error:%s", strerror(errno));
							}
							closeConnection(fd);
							BTDEBUG("Conn closed : %d", fd);
							FD_CLR(fd, &activeFdSet);
							if (fd == maxFd)
								maxFd--;
						} else {
							BTDEBUG("Bytes read : %d", bytesRead);
							sdpReq = (SDPRequest *)malloc(sizeof(SDPRequest));

							/*
							 ** Bytes and who it came from
							 */
							sdpReq->requestSocket = fd;
							sdpReq->requestBuffer = realBuf;
						#if defined(SDP_THREADED)
							addEntryToRequestQueue(sdpReq);
						#else
							BTDEBUG("Processing request");
							processRequest(sdpReq);
							BTDEBUG("Out of process request");
						#endif
						}
					}
					checkCount++;
				}
				if (checkCount == numFds)
					break;
			}
		}
	}
exit:
	resetServiceDB();
	return status;
}


