/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author:Guruprasad Krishnamurthy <guruprasad.krishnamurthy@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   */

/*
   $Id: sdpsrv.c,v 1.3 2002/08/01 17:52:08 kds Exp $

   Contains implementation of SDP service registration APIs

	Fixes:     Guruprasad Krishnamurthy <guruprasad.krishnamurthy@nokia.com>

		Manel Guerrero				: 
		Dmitry Kasatkin				: SDPServerHandle changed to complex type
*/

#include <affix/config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <affix/sdp.h>
#include <affix/sdpsrv.h>
#include <affix/sdpclt.h>
#include "pdu.h"
#include "servicedb.h"
#include "utils.h"
#include "attr.h"


static ServiceAttributeIdentifier svcRecHandleAttrId = AttrServiceRecordHandle;

/*
 ** Returns the internal state of the service record
 ** pointed to by svcRecHandle
 */
ServiceRecordInternalState getServiceRecordInternalState(ServiceRecord *svcRec)
{
	if (svcRec != NULL) {
		return svcRec->internalState;
	} else
		return doesNotExist;
}

/*
 **
 ** Registers a newly created ServiceRecord with the SDP server
 ** specified by the SDPServerHandle.
 **
 ** It is incorrect to call this method on a ServiceRecord that
 ** has been already registered with the server.
 **
 ** Returns a non-null value (a pointer) to a service
 ** record if successful, else -1
 */
int __registerServiceRecord(SDPServerHandle srvHandle, ServiceRecord *svcRec)
{
	char	*pdata;
	int	status = 0;
	char	*requestBuffer;
	char	*responseBuffer;
	int	requestSize;
	int	responseSize;

	SdpPduHeader	*pduRequestHeader;
	SdpPduHeader	*pduResponseHeader;
	SDPData		*pData;

	ServiceRecordHandle	svcRecHandle;


	if (svcRec->internalState != createdRegisterReady)
		return E_FAILURE;
	
	requestBuffer = (char *)malloc(RequestBufferSize);
	if (!requestBuffer)
		return E_FAILURE;
	responseBuffer = (char *)malloc(ResponseBufferSize);
	if (!responseBuffer) {
		free(requestBuffer);
		return E_FAILURE;
	}

	pduRequestHeader = (SdpPduHeader *)requestBuffer;
	pduRequestHeader->pduId = SDP_ServiceRegistrationRequest;
	pduRequestHeader->transactionId = htons(generateTransactionId());

	pdata = (char *)(requestBuffer + sizeof(SdpPduHeader));
	requestSize = sizeof(SdpPduHeader);                       

	generateServiceRecordPduForm(svcRec);
	memcpy(pdata, svcRec->pduForm.data, svcRec->pduForm.dataSize);
	requestSize += svcRec->pduForm.dataSize;

	pduRequestHeader->paramLength = htons(requestSize - sizeof(SdpPduHeader));

	status = sendRequestWaitForResponse(srvHandle, requestBuffer,
			responseBuffer, requestSize, &responseSize);

	if (status == E_OK) {
		pduResponseHeader = (SdpPduHeader *)responseBuffer;
		pdata = responseBuffer + sizeof(SdpPduHeader);                 
		if ((pduResponseHeader->pduId == SDP_ServiceRegistrationResponse) &&
				(status == E_OK)) {
			svcRecHandle = ntohl(__get_u32(pdata));
			svcRec->serviceRecordHandle = svcRecHandle;
			pData = createSDPData(UnsignedInteger_32, 
					(char *)&svcRecHandle);
			addToAttributeList(svcRec, svcRecHandleAttrId, pData, 1);
			svcRec->internalState = registered;
		} else
			status = -1;
	}
	free(requestBuffer);
	free(responseBuffer);
	return status;
}

int registerServiceRecord(ServiceRecord *svcRec)
{
	SDPServerHandle srvHandle;
	int		status;


	srvHandle = SDPOpenLocalServerConnection();
	if (srvHandle < 0)
		return srvHandle;

	status = __registerServiceRecord(srvHandle, svcRec);

	SDPCloseServerConnection(srvHandle);
	return status;
}

/*
 ** Deletes a service record pointed to by
 ** svcRecHandle. Note that you still need
 ** to call commitServiceRecord() to effect
 ** an actual delete in the SDP server DB
 **
 ** Returns 0 if successful or -1 if there
 ** is a failure.
 **
 */
int __deleteServiceRecord(SDPServerHandle srvHandle, ServiceRecord *svcRec)
{
	char	*pdata;
	int	status = 0;
	char	*requestBuffer;
	char	*responseBuffer;
	int	requestSize = 0;
	int	responseSize = 0;

	SdpPduHeader	*pduRequestHeader;
	SdpPduHeader	*pduResponseHeader;

	ServiceRecordHandle	svcRecHandle = 0;

	svcRecHandle = svcRec->serviceRecordHandle;
	if ((svcRecHandle == SDPServerServiceRecordHandle) ||
			((svcRec->internalState != registered) &&
			 (svcRec->internalState != created) && /* FIXME POPO TAROM */
			 (svcRec->internalState != modifiedUpdateReady)))
		return E_INVALID_ARGS;

	requestBuffer = (char *)malloc(RequestBufferSize);
	if (!requestBuffer)
		return E_FAILURE;
	responseBuffer = (char *)malloc(ResponseBufferSize);
	if (!responseBuffer) {
		free(requestBuffer);
		return E_FAILURE;
	}

	pduRequestHeader = (SdpPduHeader *)requestBuffer;
	pduRequestHeader->pduId = SDP_ServiceRemovalRequest;
	pduRequestHeader->transactionId = htons(generateTransactionId());

	pdata = (char *)(requestBuffer + sizeof(SdpPduHeader));
	requestSize = sizeof(SdpPduHeader);                       
	__put_u32(pdata, htonl(svcRecHandle));
	requestSize += sizeof(ServiceRecordHandle);

	pduRequestHeader->paramLength = htons(requestSize - sizeof(SdpPduHeader));
	status = sendRequestWaitForResponse(srvHandle, requestBuffer,
			responseBuffer, requestSize, &responseSize);
	if (status == E_OK) {
		pduResponseHeader = (SdpPduHeader *)responseBuffer;
		pdata = responseBuffer + sizeof(SdpPduHeader);                 
		status = __get_u16(pdata);
		if ((pduResponseHeader->pduId == SDP_ServiceRemovalResponse) &&
				(status == E_OK)) {
			BTDEBUG("Removing local copy\n");
			freeServiceRecord(svcRec);
			svcRec->internalState = doesNotExist;
		} else
			status = -1;
	}
	free(requestBuffer);
	free(responseBuffer);
	return status;
}

int deleteServiceRecord(ServiceRecord *svcRec)
{
	SDPServerHandle srvHandle;
	int		status;


	srvHandle = SDPOpenLocalServerConnection();
	if (srvHandle < 0)
		return srvHandle;

	status = __deleteServiceRecord(srvHandle, svcRec);

	SDPCloseServerConnection(srvHandle);
	return status;
}


int __deleteServiceRecordByHandle(SDPServerHandle srvHandle, ServiceRecordHandle svcRecHandle)
{
	char	*pdata;
	int	status = 0;
	char	*requestBuffer;
	char	*responseBuffer;
	int	requestSize = 0;
	int	responseSize = 0;

	SdpPduHeader	*pduRequestHeader;
	SdpPduHeader	*pduResponseHeader;


	requestBuffer = (char *)malloc(RequestBufferSize);
	if (!requestBuffer)
		return E_FAILURE;
	responseBuffer = (char *)malloc(ResponseBufferSize);
	if (!responseBuffer) {
		free(requestBuffer);
		return E_FAILURE;
	}

	pduRequestHeader = (SdpPduHeader *)requestBuffer;
	pduRequestHeader->pduId = SDP_ServiceRemovalRequest;
	pduRequestHeader->transactionId = htons(generateTransactionId());

	pdata = (char *)(requestBuffer + sizeof(SdpPduHeader));
	requestSize = sizeof(SdpPduHeader);                       
	__put_u32(pdata, htonl(svcRecHandle));
	requestSize += sizeof(ServiceRecordHandle);

	pduRequestHeader->paramLength = 
		htons(requestSize - sizeof(SdpPduHeader));
	status = sendRequestWaitForResponse(srvHandle, requestBuffer,
			responseBuffer, requestSize, &responseSize);
	if (status == E_OK) {
		pduResponseHeader = (SdpPduHeader *)responseBuffer;
		pdata = responseBuffer + sizeof(SdpPduHeader);                 
		status = __get_u16(pdata);
		if ((pduResponseHeader->pduId == SDP_ServiceRemovalResponse) &&
				(status == E_OK)) {
			BTDEBUG("Removing local copy\n");
		} else
			status = -1;
	}
	free(requestBuffer);
	free(responseBuffer);
	return status;
}

int deleteServiceRecordByHandle(ServiceRecordHandle svcRecHandle)
{
	SDPServerHandle srvHandle;
	int		status;

	srvHandle = SDPOpenLocalServerConnection();
	if (srvHandle < 0)
		return srvHandle;

	status = __deleteServiceRecordByHandle(srvHandle, svcRecHandle);

	SDPCloseServerConnection(srvHandle);
	return status;
}



/*
 ** Use this function to modify an already created
 ** service record
 */
int __updateServiceRecord(SDPServerHandle srvHandle, ServiceRecord *svcRec)
{
	char *pdata;
	int status = 0;
	char *requestBuffer;
	char *responseBuffer;
	int requestSize;
	int responseSize;
	SdpPduHeader *pduRequestHeader;
	SdpPduHeader *pduResponseHeader;
	ServiceRecordHandle svcRecHandle;

	svcRecHandle = svcRec->serviceRecordHandle;
	if ((svcRecHandle == SDPServerServiceRecordHandle) ||
			((svcRec->internalState != modifiedUpdateReady)))
		return E_INVALID_ARGS;

	requestBuffer = (char *)malloc(RequestBufferSize);
	if (!requestBuffer)
		return E_FAILURE;
	responseBuffer = (char *)malloc(ResponseBufferSize);
	if (!responseBuffer) {
		free(requestBuffer);
		return E_FAILURE;
	}
	pduRequestHeader = (SdpPduHeader *)requestBuffer;
	pduRequestHeader->pduId = SDP_ServiceUpdateRequest;
	pduRequestHeader->transactionId = htons(generateTransactionId());

	pdata = (char *)(requestBuffer + sizeof(SdpPduHeader));
	requestSize = sizeof(SdpPduHeader);                       

	__put_u32(pdata, htonl(svcRecHandle));
	requestSize += sizeof(ServiceRecordHandle);
	pdata += sizeof(ServiceRecordHandle);

	generateServiceRecordPduForm(svcRec);
	memcpy(pdata, svcRec->pduForm.data, svcRec->pduForm.dataSize);
	requestSize += svcRec->pduForm.dataSize;

	pduRequestHeader->paramLength = htons(requestSize - sizeof(SdpPduHeader));
	status = sendRequestWaitForResponse(srvHandle, requestBuffer,
			responseBuffer, requestSize, &responseSize);
	BTDEBUG("Send req status : %d\n", status);
	if (status == E_OK) {
		pduResponseHeader = (SdpPduHeader *)responseBuffer;
		pdata = responseBuffer + sizeof(SdpPduHeader);                 
		status = __get_u16(pdata);
		if ((pduResponseHeader->pduId == SDP_ServiceUpdateResponse) &&
				(status == E_OK)) {
			svcRec->internalState = registered;
		} else
			status = -1;
	}
	free(requestBuffer);
	free(responseBuffer);
	return status;
}

int updateServiceRecord(ServiceRecord *svcRec)
{
	SDPServerHandle srvHandle;
	int		status;


	srvHandle = SDPOpenLocalServerConnection();
	if (srvHandle < 0)
		return srvHandle;

	status = __updateServiceRecord(srvHandle, svcRec);

	SDPCloseServerConnection(srvHandle);
	return status;
}

