/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author:Guruprasad Krishnamurthy <guruprasad.krishnamurthy@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   */

/*
   $Id: servicedb.c,v 1.4 2002/08/29 12:01:15 kds Exp $

   Contains the implementation of the service repository.
   Has methods to create and clean the repository, besides
   methods to add/find/modify/delete individual entries

	Fixes:  	Guruprasad Krishnamurthy <guruprasad.krishnamurthy@nokia.com>
	
			Dmitry Kasatkin		: some fixes
*/

#include <affix/config.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/ioctl.h>
#include <affix/bluetooth.h>
#include <fcntl.h>
#include <affix/btcore.h>
#include <string.h>

#include "servicedb.h"
#include "glib.h"
#include "pthread.h"

extern BD_ADDR localBTDeviceAddr;

/*
 ** Internal data structure holding service
 ** records. Need not be seen by any other module
 */
static GHashTable	*deviceTable;
static GSList		*serviceList;


/*
 ** Ordering function called when inserting a service record.
 ** The service repository is a linked list in sorted order
 ** and the service record handle is the sort key
 */
static gint orderingFunction(gconstpointer rec1, gconstpointer rec2)
{
	gint order = -1;

	const ServiceRecord *svcRec1 = (const ServiceRecord *)rec1;
	const ServiceRecord *svcRec2 = (const ServiceRecord *)rec2;

	if ((svcRec1 == NULL) || (svcRec2 == NULL)) {
		BTERROR("NULL SVCREC LIST FATAL\n");
	}
	else
		order = svcRec1->serviceRecordHandle -
			svcRec2->serviceRecordHandle;

	return order;
}


/*
 ** Initialize the service repository
 */
void initServiceDB(void)
{
	deviceTable = g_hash_table_new(g_str_hash, g_str_equal);
	serviceList = NULL;
}


void deviceTableList(gpointer key, gpointer value, gpointer data)
{
	BTDEBUG("Key : %s value : 0x%x", (char *)key, (uint32_t)value);
}


void svcListCleanup(gpointer key, gpointer value, gpointer data)
{
	GSList *svcList;

	svcList = (GSList *)value;
	if (svcList != NULL) {
		/*
		 ** First cleanup the ServiceRecord(s) we
		 ** may have cached ..
		 */
		int i = 0;
		for (;;) {
			ServiceRecord *svcRec = NULL;

			svcRec = (ServiceRecord *)g_slist_nth_data(svcList, i);
			if (svcRec != NULL) {
				BTDEBUG("Freeing : 0x%x", svcRec->serviceRecordHandle);
				freeServiceRecord(svcRec);
				i++;
			}
			else
				break;
		}

		/*
		 ** Next we cleanup the list itself
		 */
		g_slist_free(svcList);
	}
	if (key)
		free(key);
}


/*
 ** Reset the service repository
 ** Implies deleting all it's contents
 */
void resetServiceDB(void)
{
	if (deviceTable != NULL) {
		g_hash_table_foreach(deviceTable, svcListCleanup, NULL);
		g_hash_table_destroy(deviceTable);
	}
	svcListCleanup(NULL, serviceList, NULL);
}


/*
 ** Add a service record to the repository
 */
int addServiceRecord(ServiceRecord *svcRecToAdd)
{
	int status = -1;
	char devAddrStr[32];
	GSList *svcList = NULL;
	gchar *ptr = devAddrStr;
	gboolean serviceListExists = FALSE;

	if (svcRecToAdd == NULL) {
		BTERROR("TRYING TO ADD NULL SVCREC");
		return -1;
	}

	memset(ptr, 0, 32);
	bda2str(devAddrStr, &localBTDeviceAddr);
	svcList = (GSList *)g_hash_table_lookup(deviceTable, devAddrStr);
	if (svcList != NULL)
		serviceListExists = TRUE;

#ifdef CONFIG_AFFIX_DEBUG
	BTDEBUG("Dev addr : %s", devAddrStr);
	BTDEBUG("Adding svc rec : 0x%lx", (long)svcRecToAdd);
	BTDEBUG("with handle : 0x%lx", (long)svcRecToAdd->serviceRecordHandle);
	g_hash_table_foreach(deviceTable, deviceTableList, NULL);
	if (svcList == NULL) {
		BTDEBUG("No svc list for : %s", devAddrStr);
	} else {
		BTDEBUG("Entries in svcList : %d", g_slist_length(svcList));
	}
#endif

	svcList = g_slist_insert_sorted(svcList, svcRecToAdd, orderingFunction);

	BTDEBUG("Entries in svcList : %d", g_slist_length(svcList));

	if (svcList != NULL) {
		char *key;

		if (serviceListExists == TRUE)
			key = devAddrStr;
		else {
			key = (char *)malloc(32);
			memset(key, 0, 32);
			strcpy(key, devAddrStr);
		}
		g_hash_table_insert(deviceTable, key, svcList);
		status = 0;
	}

#ifdef CONFIG_AFFIX_DEBUG
	BTDEBUG("Entries in devicetable : %d", g_hash_table_size(deviceTable));
	g_hash_table_foreach(deviceTable, deviceTableList, NULL);
#endif

	return status;
}


GSList *locateServiceRecord(ServiceRecordHandle svcRecHandle, GSList **listStart)
{
	ServiceRecord tmpRec;
	// ServiceRecord *realRec = NULL;
	GSList *svcList = NULL;
	GSList *tmpList = NULL;
	char devAddrStr[32];

	bda2str(devAddrStr, &localBTDeviceAddr);
	svcList = (GSList *)g_hash_table_lookup(deviceTable, devAddrStr);

	if (svcList != NULL) {
		*listStart = svcList;
		tmpRec.serviceRecordHandle = svcRecHandle;
		tmpList = g_slist_find_custom(svcList, &tmpRec, orderingFunction);
#ifdef CONFIG_AFFIX_DEBUG
		if (tmpList == NULL)
			BTDEBUG("Could not find svcRec for : 0x%x on device : %s\n", svcRecHandle, devAddrStr);
	} else {
		BTDEBUG("Device : %s does not have a service table\n", devAddrStr);
#endif
	}

	return tmpList;
}

/*
 ** Given a service record handle, find the record associated
 ** with it.
 */
ServiceRecord * findServiceRecord(ServiceRecordHandle svcRecHandle)
{
	ServiceRecord *realRec = NULL;
	GSList *tmpList = NULL;
	GSList *listStart = NULL;

	tmpList = locateServiceRecord(svcRecHandle, &listStart);

	if (tmpList != NULL) {
		realRec = (ServiceRecord *)tmpList->data; 
	}
#ifdef CONFIG_AFFIX_DEBUG
	if (realRec == NULL) {
		BTDEBUG("Couldn't find record for : 0x%x\n", svcRecHandle);
	}
#endif

	return realRec;
}


/*
 ** Given a service record handle, remove the service
 ** record from the repository
 */
int removeServiceRecord(ServiceRecordHandle svcRecHandle)
{
	int status = -1;
	GSList *tmpList = NULL;
	GSList *svcList = NULL;
	ServiceRecord *svcRecToRemove = NULL;
	char devAddrStr[32];
	GSList *listStart = NULL;

	tmpList = locateServiceRecord(svcRecHandle, &listStart);
	if (tmpList != NULL) {
		svcRecToRemove = (ServiceRecord *)tmpList->data;
		if (svcRecToRemove != NULL) {
			bda2str(devAddrStr, &localBTDeviceAddr);
			BTDEBUG("Removing : 0x%x from : %s\n", (uint32_t)svcRecToRemove, devAddrStr);
			svcList = (GSList *)g_slist_remove(listStart, svcRecToRemove);
			g_hash_table_insert(deviceTable, devAddrStr, svcList);
			status = 0;
		}
	}

	if (status == -1)
		BTERROR("Remove : Couldn't find record for : 0x%x\n", svcRecHandle);

	return status;
}


/*
 ** Return a pointer to the linked list containing
 ** the records in sorted order
 */
GSList *getServiceRecordList(BD_ADDR bdAddr)
{
	GSList *svcList = NULL;
	char devAddrStr[32];

	bda2str(devAddrStr, &bdAddr);
	svcList = (GSList *)g_hash_table_lookup(deviceTable, devAddrStr);
#ifdef CONFIG_AFFIX_DEBUG
	if (svcList == NULL)
		BTDEBUG("NULL svc list for : %s\n", devAddrStr);
#endif

	return svcList;
}


/*
 * check if there are records in database
 */

int emptyServiceDB(void)
{
	GSList	*svcRecList;

	svcRecList = getServiceRecordList(localBTDeviceAddr);

	return g_slist_length(svcRecList);
}

/*
 * Remove services by RequestSocket
 *
 * will it work?? if I remove going through list
 */
int removeServiceRecordByFd(int fd)
{
	GSList	*svcRecList;
	int	i, status;

start_again:
	svcRecList = getServiceRecordList(localBTDeviceAddr);
	for (i = 0; i < (int)g_slist_length(svcRecList); i++) {
		ServiceRecord *svcRec;

		svcRec = (ServiceRecord *)g_slist_nth_data(svcRecList, i);
		if (svcRec->ownerSocket == fd) {
			status = removeServiceRecord(svcRec->serviceRecordHandle);
			freeServiceRecord(svcRec);
			goto start_again;
		}
	}
	return 0;	
}

