/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author:Guruprasad Krishnamurthy <guruprasad.krishnamurthy@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   */

/*
   $Id: utils.c,v 1.3 2002/07/19 15:18:34 kds Exp $

   Contains implementation of miscellaneous utility functions
   used by SDP clients and server

	Fixes:		Guruprasad Krishnamurthy <guruprasad.krishnamurthy@nokia.com>

			Dmitry Kasatkin		: MTU handling. support for TCP socket 
							
*/

#include <affix/config.h>

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <glib.h>
#include <sys/syslog.h>
#include <errno.h>

#include <affix/sdp.h>
#include <affix/sdpclt.h>
#include "servicedb.h"
#include "utils.h"
#include "cstate.h"
#include "attr.h"

const int handleInfoCount = 10;
int handleInfoCurrentCount = 0;

BD_ADDR localBTDeviceAddr;
OperationMode opMode;


/*
 ** A simple function which returns the time of day in
 ** seconds. Used for updating the service db state
 ** attribute of the service record of the SDP server
 */
long getTimeInSeconds(void)
{
	/*
	 ** To handle failure in gettimeofday, so an old
	 ** value is returned and service does not fail
	 */
	static struct timeval tm;

	gettimeofday(&tm, NULL);

	return tm.tv_sec;
}


/*
 ** Generate unique transaction identifiers
 */
uint16_t generateTransactionId(void)
{
	uint16_t j;

	j=1+(int) (100.0*rand()/(RAND_MAX+1.0));

	return j;
}


/*
 ** Close an existing connection to the SDP server on a said
 ** socket descriptor. 
 */
void closeConnection(int socketFd)
{
	int status;

	if (socketFd != -1) {
		status = close(socketFd);
		if (opMode == server)
			cleanupCachedResponses(socketFd);
	}
}                         


/*
 ** Debug init
 */
void initDebug(void)
{
	/*
	 ** Initiliaze the debug infrastructure
	 */
	// char progStr[32];
	// char pidStr[16];

	// strcpy(progStr, "CONFIG_AFFIX_DEBUG_");
	// sprintf(pidStr, "%d", (int)getpid());
	// strcat(progStr, pidStr);
	// printf("Debug str : %s\n", progStr);

	// openlog(progStr, LOG_PERROR, LOG_LOCAL2);
}



/*
 ** Initialize the SDP machinery
 */
int sdp_init(OperationMode mode)
{
	int status = 0;

	opMode = mode;
	getLocalBDAddr(&localBTDeviceAddr);

	/*
	 ** All we do for now is create a service DB
	 */
	initDebug();
	initServiceDB();
	if (mode == server)
		initContinuationStateCache();

	return status;  
}


/*
 ** This must be called to shut the system down
 */
void sdp_shutdown(void)
{
}


/*
 ** Extract the sequence type, it's length
 ** and return a pointer to the beginning of
 ** sequence
 */
char *extractSequenceType(
		char *dstBuffer,
		DataTypeDescriptor *dataType,
		int *sequenceSize,
		int *bytesScanned
		)
{
	char *pdata;
	DataTypeDescriptor localDataType;

	pdata = dstBuffer;
	localDataType = __get_u8(pdata);
	pdata += sizeof(DataTypeDescriptor);
	*bytesScanned += sizeof(DataTypeDescriptor);
	*dataType = localDataType;

	switch (localDataType) {
		case SequenceSize8 :
		case SequenceAlternatesSize8 :
			*sequenceSize = __get_u8(pdata);
			pdata += sizeof(DataSizeDescriptor8);
			*bytesScanned += sizeof(DataSizeDescriptor8);
			break;

		case SequenceSize16 :
		case SequenceAlternatesSize16 :
			*sequenceSize = ntohs(__get_u16(pdata));
			pdata += sizeof(DataSizeDescriptor16);
			*bytesScanned += sizeof(DataSizeDescriptor16);
			break;

		case SequenceSize32 :
		case SequenceAlternatesSize32 :
			*sequenceSize = ntohl(__get_u32(pdata));
			pdata += sizeof(DataSizeDescriptor32);
			*bytesScanned += sizeof(DataSizeDescriptor32);
			break;

		default :
			BTERROR("Unknown sequence type, aborting\n");
			pdata = NULL;
			break;
	}

	return pdata;
}

/*
 ** Generate the attribute sequence pdu form
 ** from GSList elements. Return length of attr seq
 */
int  generateDataSequencePduForm(char *dstBuffer, GSList *seq,DataTypeDescriptor dataType)
{
	SDPData *pDataSeq;
	GPtrArray *typeArray;
	GPtrArray *valueArray;
	PDUForm pduForm;
	int seqLength = 0, i;

	//
	// Fill up the value and the dtd arrays
	//
	BTDEBUG("gen sequence\n");
	memset(&pduForm, 0, sizeof(PDUForm));
	pduForm.data = (char *)malloc(UUIDSequenceSize);
	pduForm.bufferSize = UUIDSequenceSize;

	seqLength = g_slist_length(seq);
	BTDEBUG("Seq length : %d\n", seqLength);
	typeArray = g_ptr_array_new();
	valueArray = g_ptr_array_new();
	for (i = 0; i < seqLength; i++) {
		gpointer data = g_slist_nth_data(seq, i);
		g_ptr_array_add(typeArray, &dataType);
		if ((dataType == UUID_16) || 
				(dataType == UUID_32) || 
				(dataType == UUID_128))
			data = &((UUID *)data)->value;
		g_ptr_array_add(valueArray, data);
	}

	seqLength = 0;
	pDataSeq = createSDPDataSequence(typeArray, valueArray);
	BTDEBUG("Data Seq : 0x%x\n", pDataSeq);
	seqLength = generatePDUForm(&pduForm, pDataSeq);
	BTDEBUG("Copying : %d\n", pduForm.dataSize);
	memcpy(dstBuffer, pduForm.data, pduForm.dataSize);

	g_ptr_array_free(typeArray, FALSE);
	g_ptr_array_free(valueArray, FALSE);
	free(pDataSeq); free(pduForm.data);

	return seqLength;
}

int generateSearchSequencePduForm(char *dstBuffer, GSList *seq)
{
	UUID *pUUID;
	DataTypeDescriptor dataType;

	pUUID = (UUID *)g_slist_nth_data(seq, 0);
	if (pUUID->uuidType == bits_16)
		dataType = UUID_16;
	else
		if (pUUID->uuidType == bits_32)
			dataType = UUID_32;
		else
			dataType = UUID_128;

	return generateDataSequencePduForm(dstBuffer, seq, dataType);
}

int generateAttrIDSequencePduForm(char *dstBuffer, GSList *seq, DataTypeDescriptor dataType)
{
	return generateDataSequencePduForm(dstBuffer, seq, dataType);
}


/*
 ** Extract a sequence of service record handles from a PDU buffer
 ** and add the entries to a GSList. Note that the service record
 ** handles are not in "data element sequence" form, but just like
 ** an array of service handles
 */
int __extractServiceRecordHandles(
		char *pduBuffer,
		GSList **svcReqSeq,
		int currentHandleCount,
		int *bytesScanned
		)
{
	int status = -1;
	char *pdata;
	GSList *pSeq;
	ServiceRecordHandle *pSvcRec;
	int numberOfSvcRec = 0;

	pdata = pduBuffer;

	numberOfSvcRec = 0; pSeq = *svcReqSeq;
	while (1) {
		pSvcRec = (ServiceRecordHandle *)malloc(sizeof(ServiceRecordHandle));
		*pSvcRec = ntohl(__get_u32(pdata));
		pSeq = g_slist_append(pSeq, pSvcRec);
		pdata += sizeof(ServiceRecordHandle);
		*bytesScanned += sizeof(ServiceRecordHandle);
		numberOfSvcRec++;
		if (numberOfSvcRec == currentHandleCount) {
			status = 0;
			break;
		}
	}

	*svcReqSeq = pSeq;

	return status;
}

int extractServiceRecordHandles(
		char *pduBuffer,
		GSList **svcReqSeq,
		int currentHandleCount,
		int *bytesScanned
		)
{
	int status = -1;
	char *pdata;
	GSList *pSeq;
	ServiceRecordHandle *pSvcRec;
	int numberOfSvcRec = 0;

	pdata = pduBuffer;

	numberOfSvcRec = 0; pSeq = NULL;
	while (1) {
		pSvcRec = (ServiceRecordHandle *)malloc(sizeof(ServiceRecordHandle));
		*pSvcRec = ntohl(__get_u32(pdata));
		pSeq = g_slist_append(pSeq, pSvcRec);
		pdata += sizeof(ServiceRecordHandle);
		*bytesScanned += sizeof(ServiceRecordHandle);
		numberOfSvcRec++;
		if (numberOfSvcRec == currentHandleCount) {
			status = 0;
			break;
		}
	}

	*svcReqSeq = pSeq;

	return status;
}


/*
 ** Generic data element sequence extractor. Builds
 ** a GSList whose elements are those found in the 
 ** sequence. The data type of elements found in the
 ** sequence (now populated in GSList) is returned
 ** in the reference pDataType
 */
int extractDataElementSequence(
		char *pduBuffer,
		GSList **svcReqSeq,
		DataTypeDescriptor *pDataType,
		DataTypeDescriptor expectedType
		)
{
	int status, bytesScanned;
	char *pdata;
	DataTypeDescriptor seqType;
	int dataSize = 0;

	bytesScanned = 0; status = 0; 

	pdata = extractSequenceType(pduBuffer, &seqType, &dataSize, &bytesScanned);

	BTDEBUG("Seq type : %d", seqType);
	if ((pdata == NULL) ||
			((seqType != SequenceSize8) && (seqType != SequenceSize16))) {
		BTERROR("Unknown seq type");
		status = -1;
	}

	if (status != -1) {
		short numberOfElements;
		int seqLength;
		GSList *pSeq;
		DataTypeDescriptor dataType;

		BTDEBUG("Data size : %d", dataSize);
		numberOfElements = 0; seqLength = 0; pSeq = NULL;
		while (1) {
			char *pElem = NULL;
			int localSeqLength = 0;

			dataType = __get_u8(pdata);
			BTDEBUG("Data type : %d", dataType);

			switch (dataType) {
				case UnsignedInteger_16 :
					pdata += sizeof(DataTypeDescriptor);
					seqLength += sizeof(DataTypeDescriptor);
					pElem = (char *)malloc(sizeof(uint16_t));
					*(uint16_t *)pElem = ntohs(__get_u16(pdata));
					pdata += sizeof(uint16_t);
					seqLength += sizeof(uint16_t);
					break;

				case UnsignedInteger_32 :
					pdata += sizeof(DataTypeDescriptor);
					seqLength += sizeof(DataTypeDescriptor);
					pElem = (char *)malloc(sizeof(uint32_t));
					*(uint32_t *)pElem = ntohl(__get_u32(pdata));
					pdata += sizeof(uint32_t);
					seqLength += sizeof(uint32_t);
					break;

				case UUID_16 :
				case UUID_32 :
				case UUID_128 :
					pElem = (char *)malloc(sizeof(UUID));
					status = extractUUID(pdata, (UUID *)pElem, &localSeqLength);
					if (status == 0) {
						seqLength += localSeqLength;
						pdata += localSeqLength;
					}
					break;
			}

			if (status == 0) {
				pSeq = g_slist_append(pSeq, pElem);
				numberOfElements++;
				BTDEBUG("No of elements : %d", numberOfElements);
				if (seqLength == dataSize)
					break;
				else
					if (seqLength > dataSize) {
						status = -1;
						break;
					}
			}
			else
				free(pElem);
		}

		if (status != -1) {
			*svcReqSeq = pSeq;
			bytesScanned += seqLength;
			*pDataType = dataType;
		}
	}

	if (status == 0)
		return bytesScanned;
	else
		return -1;
}



/*
 ** A generic send request, wait for response method. Uses
 ** select system call to block for response or timeout
 */
int sendRequestWaitForResponse(
		SDPServerHandle srvHandle,
		char *requestBuffer,
		char *responseBuffer,
		int requestSize,
		int *responseSize
		)
{
	int status = 0;
	int bytesSent = 0;
	int bytesRecvd = 0;
	SdpPduHeader *pduRequestHeader;
	SdpPduHeader *pduResponseHeader;
	fd_set readFds;
	struct timeval timeout;

	pduRequestHeader = (SdpPduHeader *)requestBuffer;
	pduResponseHeader = (SdpPduHeader *)responseBuffer;

	/*
	 ** Send the request, wait for response and if
	 ** not error, set the appropriate values
	 ** and return
	 */
	status = send(srvHandle, requestBuffer, requestSize, 0);
	if (status == -1) {
		BTERROR("Error sending data:%s", strerror(errno));
		goto exit;
	}
	bytesSent = status;
	if (bytesSent != requestSize) {
		BTDEBUG("Attempt to send : %d sent : %d\n", requestSize, bytesSent);
		status = E_FAILURE;
		goto exit;
	}
#ifdef CONFIG_AFFIX_DEBUG
	BTDEBUG("Bytes sent to handle : %d is : %d\n", srvHandle, bytesSent);        
	BTDUMP(requestBuffer, bytesSent);
	_BTDEBUG("\n");
#endif
	
	BTDEBUG("Waiting for response\n");
	timeout.tv_sec = WaitTimeForResponse; 
	timeout.tv_usec = 0;

	FD_ZERO(&readFds);
	FD_SET(srvHandle, &readFds);
	status = select(srvHandle+1, &readFds, NULL, NULL, &timeout);
	if (status < 0)
		goto exit;
	
	if (status == 0) {
		status = E_TIMEOUT;
		BTERROR("Client timed out\n");
		goto exit;
	}

	BTDEBUG("Trying to read\n");
	status = recv(srvHandle, responseBuffer, ResponseBufferSize, 0);
	if (status <= 0)
		goto exit;

	bytesRecvd = status;
#ifdef CONFIG_AFFIX_DEBUG
	BTDEBUG("Read : %d\n", bytesRecvd);
	BTDUMP(responseBuffer, bytesRecvd);
	_BTDEBUG("\n");
#endif
	/*
	 ** Process the response
	 */
	pduResponseHeader = (SdpPduHeader *)responseBuffer;
	if (pduRequestHeader->transactionId ==
			pduResponseHeader->transactionId) {
		*responseSize = bytesRecvd;
		status = 0;
	} else {
		status = E_FAILURE;
	}
exit:
	return status;
}

union sockaddr_sdp {
	struct sockaddr_affix	bt;
	struct sockaddr_in	in;
};

int getMTUFromFd(int socketFd)
{
	union sockaddr_sdp	caddr;
	socklen_t		len = sizeof(caddr);
	int			MTU;

	getsockname(socketFd, (struct sockaddr*)&caddr, &len);
	if (caddr.bt.family == PF_AFFIX)
		MTU = l2cap_getmtu(socketFd);
	else {
		len = sizeof(int);
		getsockopt(socketFd, SOL_SOCKET, SO_SNDBUF, &MTU, &len);
		BTDEBUG("TCP MTU: %d\n", MTU);
	}

	return MTU;
}

BD_ADDR getBdAddrFromFd(int socketFd)
{
	union sockaddr_sdp	caddr;
	socklen_t		len = sizeof(caddr);

	getsockname(socketFd, (struct sockaddr*)&caddr, &len);
	if (caddr.bt.family == PF_AFFIX)
		return caddr.bt.bda;

	return localBTDeviceAddr;
}

/*
 ** System logging utility
 */
void logMessage(char *str)
{
	// if (str != NULL)
	// syslog(LOG_LOCAL2+LOG_DEBUG, str);
}
